package de.fhg.iais.outdoor.simba.core;

import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Logger;

public abstract class EnvironmentModel extends ParameterElement implements Cloneable {
	
	
	private Vector mAddQueue = new Vector();
	private Vector mRemoveQueue = new Vector();
	private Vector mDependencySortedSimObjects = new Vector();
	protected Vector mLayerSortedSimObjects = new Vector();
	private boolean mInitialized;
	private String mName; 
	
	
	public EnvironmentModel() {
		this("Default");
	}
	
	
	/**
	 * Standard constructor.
	 */
	public EnvironmentModel(String name) {
		super(name, "/Environment/" + name + "/");
		mInitialized = false;
		mName = name;
	}
	
	/**
	 * Returns the name of this EnvironmentModel.
	 * @return
	 */
	public String getName() {
		return mName;
	}
	
	/**
	 * This method is called in the beginning of each try
	 * in which an environment clone is used for the simulation execution.
	 * This method is used to reset the parameters of all SimObjects of the environment
	 * to an initial state. In this method the randomized parameters can be used 
	 * to modify the initial parameters of the objects.
	 * 
	 * Important: Each clone has to reset the object paramters to exactly the same
	 * state each time this method is called. Between multiple invocations of this
	 * method at the same environment clone no (randomized) differences are allowed. 
	 */
	public abstract void resetEnvironmentObjects();
	
	/**
	 * This method is called after a variation object was cloned from the
	 * master Environment. The method is called exactly once and only
	 * for the environment clones.
	 *
	 */
	public abstract void updateRandomizationParameters();
	
	
	/**
	 * This method is called only for the MasterEnvironment and only once.
	 * All other variations are clones that reuse the objects created in this
	 * method.
	 */
	public abstract void createEnvironment();
	
	
	
	/**
	 * Initializes a master environment, which means that the createEnvironment 
	 * method is called exactly once. Additional calls to this method do not have
	 * any effect.
	 */
	public boolean initEnvironment() {
		if(!mInitialized) {
			updateContent();
			createEnvironment();
			mInitialized = true;
			if(hasObjectsInQueue()) {
				updateContent();
			}
			else {
				resolveDependencies();
			}
		}
		return true;
	}
	
	public boolean hasObjectsInQueue() {
		if(!mAddQueue.isEmpty() || !mRemoveQueue.isEmpty()) {
			return true;
		}
		return false;
	}
	
	
	/**
	 * Creates the given number of environment variation clones 
	 * based on this Environment object and returns all these
	 * EnvironmentModels in a Vector.
	 * 
	 * @param numberOfEnvironments the number of EnvironmentModels to create.
	 * @return a Vector with the newly created EnvironmentModels.
	 */
	public synchronized Vector createEnvironments(int numberOfEnvironments) {
		boolean hasObjectsInQueue = hasObjectsInQueue();
		
		if(!mInitialized) {
			initEnvironment();
		}
		else {
			if(hasObjectsInQueue) {
				updateContent();
			}
			else {
				resolveDependencies();
			}
		}
		
		
		Vector environments = new Vector();
		for(int i = 0; i < numberOfEnvironments; ++i) {
			EnvironmentModel model = createEnvironmentVariation();
			if(model != null) {
				environments.add(model);
			}
		}
		return environments;
	}
	
	/**
	 * Returns true if this EnvironmentModel was already initialized.
	 * The initialization is only done for the master environment.
	 * Variation clones of an initialized master environment are automatically
	 * marked as initialized.
	 * 
	 * @return true if this EnvironmentModel was already initialized.
	 */
	public boolean isInitialized() {
		return mInitialized;
	}
	
	/**
	 * This method should be overriden by uses to create variants of this
	 * EnvironmentModel with the same SimObjects, but with different 
	 * initial parameters.
	 * 
	 * @return a variation of this EnvironmentModel. 
	 */
	public  synchronized EnvironmentModel createEnvironmentVariation() {
		if(!mInitialized) {
			initEnvironment();
		}
		else {
			updateContent();
		}
		
		EnvironmentModel model = null;
		try {
			model = (EnvironmentModel) this.clone();
			
			//Choose new randomization parameters used to set up the Environment.
			model.updateRandomizationParameters();
			
		} catch (CloneNotSupportedException e) {
			Logger.getLogger("Simulator")
				.warning("EnvironmentModel.createEnvironmentVariation: Could not clone the EnvironmentModel.");
			e.printStackTrace();
		} 
		return model;
	}


	
	/**
	 * Resets this EnvironmentModel to its initial state. Therefore all registered
	 * SimObjects are reset and AFTERWARDS the abstract method resetEnvironmentObjects()
	 * is called to allow the EnvironmentModel to modify the starting positions
	 * of all SimObjects after they have been reset to their own initial positions.
	 */
	public synchronized void reset() {
		updateContent();
		
		for(Iterator i = mDependencySortedSimObjects.iterator(); i.hasNext();) {
			((SimObject) i.next()).reset();
		}
		resetEnvironmentObjects();
	}
	
	/**
	 * Updates the environment, then the actuators and finally the sensors.
	 */
	public synchronized void executeSimulationStep() {
		updateContent();
		for(Iterator i = mDependencySortedSimObjects.iterator(); i.hasNext();) {
			SimObject object = (SimObject) i.next();
			if(object.isActive()) {
				object.update();
			}
		}
		updateContent();
	}
	
	
	/**
	 * Adds a normal SimObject.
	 * 
	 * @param object the object to add.
	 * @return true if successful. If this object is already contained,
	 * then false is returned.
	 */
	public synchronized boolean addSimObject(SimObject object) {
		if(object == null || mDependencySortedSimObjects.contains(object) || mAddQueue.contains(object)) {
			return false;
		}
		mAddQueue.add(object);
		object.addChildren(this);
		return true;
	}
	
	
	/**
	 * Adds the SimObjects from the add-buffer to the EnvironmentModel.
	 * Removes the SimObjects from the remove-buffer from the EnvironmentModel.
	 * Sorts all SimObjects by layers and dependencies.
	 */
	public synchronized void updateContent() {
		for(Iterator i = mRemoveQueue.iterator(); i.hasNext();) {
			//TODO remove objects
		}
		if(mAddQueue.isEmpty()) {
			return;
		}
		for(int i = mAddQueue.size() - 1; i >= 0; --i) {
			SimObject obj = (SimObject) mAddQueue.elementAt(i);
			if(mDependencySortedSimObjects.contains(obj)) {
				continue;
			}
			if(obj instanceof Sensor) { //heuristic to speed up dependency resolving.
				int j = 0;
				for(; j < mDependencySortedSimObjects.size(); ++j) {
					if(!(mDependencySortedSimObjects.elementAt(j) instanceof Sensor)) {
						continue;
					}
					break;
				}
				mDependencySortedSimObjects.add(j, obj);
			}
			else {
				mDependencySortedSimObjects.add(0, obj);
			}
			addObject(obj);
		}
		mAddQueue.clear();
		resolveDependencies(mDependencySortedSimObjects);
	}
	
	/**
	 * Returns a copy of the vector containing all registered SimObjects 
	 * (except of sensors and actuators).
	 * 
	 * @return all registered SimObjects.
	 */
	public Vector getSimObjects() {
		return new Vector(mDependencySortedSimObjects);
	}
	
	/**
	 * Adds the SimObject to a vector where all SimObjects are sorted by their layers.
	 * 
	 * @param object the object to add to the sorted vector.
	 */
	private void addObject(SimObject object) {
		if(object == null || mLayerSortedSimObjects.contains(object)) {
			return;
		}
		for(int i = 0; i < mLayerSortedSimObjects.size(); ++i) {
			SimObject obj = (SimObject) mLayerSortedSimObjects.elementAt(i);
			if(obj.getLayer() > object.getLayer()) {
				mLayerSortedSimObjects.add(i, object);
				return;
			}
		}
		mLayerSortedSimObjects.add(object);
	}
	
	/**
	 * Returns a Vector that contains all objects of the EnvironmentModel
	 * sorted in ascending order of their layers.
	 * @return a vector with all SimObjects, sorted by their layers.
	 */
	public synchronized Vector getSimObjectsSortedByLayers() {
		//TODO check if this can be done faster (or if it matters to be slow) (often called, small impact on performance.
		updateContent();
		//check if the vector is still sorted.
		int layer = Integer.MAX_VALUE;
		boolean sorted = true;
		for(Iterator i = mLayerSortedSimObjects.iterator(); i.hasNext();) {
			SimObject obj = (SimObject) i.next();
			if(obj.getLayer() >= layer) {
				layer = obj.getLayer();
			}
			else {
				sorted = false;
				break;
			}
		}
		//Resort the vector.
		if(!sorted) {
			Vector objects = mLayerSortedSimObjects;
			mLayerSortedSimObjects = new Vector();
			for(Iterator i = objects.iterator(); i.hasNext();) {
				addObject((SimObject) i.next());
			}
		}
		return new Vector(mLayerSortedSimObjects);
	}
	
	
	/**
	 * Resolves the dependencies of the SimObjects.
	 */
	public void resolveDependencies() {
		resolveDependencies(mDependencySortedSimObjects);
	}
	
	
	/**
	 * Resolves the dependencies of SimObjects within a vector. After calling this 
	 * method the vector, given as a parameter of this method, is sorted by the
	 * dependencies of the SimObject. Dependent SimObjects occure later than the objects
	 * they depend on. Therefore a sequential update of all SimObjects automatically 
	 * consideres the dependency of the objects.
	 * 
	 * Note: during the specification of dependencies make sure that there are no 
	 * dependency cycles. If cycles are present, then the system might not work as
	 * expected. Dependency cycles do not lead to infinite loops and can be detected. 
	 * The warning currently does NOT contain the information about the problematic objects.
	 * 
	 * @param objects the vector to sort according to the object dependencies.
	 */
	private synchronized void resolveDependencies(Vector objects) {
		
		//worstCase number of changes only for cycle detection.
		int worstCase = objects.size() * objects.size();
		int changes = 0;
		
		boolean allOk = false;
		while(!allOk) {
			allOk = true;
			for(int i = 0; i < objects.size(); ++i) {
				boolean thisObjectOk = true;
				SimObject obj = (SimObject) objects.elementAt(i);
				Vector dependencies = obj.getDependencies();
				for(int j = i + 1; j < objects.size(); ++j) {
					SimObject dep = (SimObject) objects.elementAt(j);
					if(dependencies.contains(dep)) {
						//resort and check again.
						objects.insertElementAt(obj, j + 1);
						objects.removeElementAt(i);
						thisObjectOk = false;
						changes++;
						if(changes >= worstCase) {
							//TODO identify problematic objects.
							Logger.getLogger("Simulator")
								.warning("There was a cycle detected in the object dependencies!");
							return;
						}
						break;
					}
				}
				if(!thisObjectOk) {
					allOk = false;
					break;
				}
			}
		}
		
	}

	/**
	 * Clones this object. This method overrides the standard clone method
	 * to allow a deep copy of the internal Vectors. The elements in the Vectors
	 * are not deeply copied, so after a clone() all internal Vectors contain
	 * the same (identical) objects.
	 */
	protected synchronized Object clone() throws CloneNotSupportedException {
		EnvironmentModel model = (EnvironmentModel) super.clone();
		
		model.mAddQueue = new Vector();
		for(int i = 0; i < mAddQueue.size(); ++i) {
			model.mAddQueue.add(mAddQueue.elementAt(i));
		}
		model.mRemoveQueue = new Vector();
		for(int i = 0; i < mRemoveQueue.size(); ++i) {
			model.mRemoveQueue.add(mRemoveQueue.elementAt(i));
		}
		model.mDependencySortedSimObjects = new Vector();
		for(int i = 0; i < mDependencySortedSimObjects.size(); ++i) {
			model.mDependencySortedSimObjects.add(mDependencySortedSimObjects.elementAt(i));
		}
		model.mLayerSortedSimObjects = new Vector();
		for(int i = 0; i < mLayerSortedSimObjects.size(); ++i) {
			model.mLayerSortedSimObjects.add(mLayerSortedSimObjects.elementAt(i));
		}
		
		return model;
	}
	
	public synchronized void invalidateCalculationBuffers() {
		for(Iterator i = mLayerSortedSimObjects.iterator(); i.hasNext();) {
			((SimObject) i.next()).invalidateCalculationBuffers();
		}
	}
	
	/**
	 * Adds tags for each randomization parameter of the form:
	 * <param name="ParameterName" value="value" />
	 * The value might also be an array.
	 * Use methods getParameterEntry to get the correct format.
	 * 
	 * Adding all randomized variables to the log file allows the user
	 * to replicate the entire experimentation setup.
	 * @return
	 */
	public String createRandomizationParameterLog() {
		return "";
	}
	
	
	

}
