package de.fhg.iais.outdoor.simba.core;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.evolution.EvolutionLogger;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.StringValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.value.ValueReplacedListener;

public class Scheduler implements SystemObject, ValueReplacedListener {
	
//	TODO TestCases!
	
	private EnvironmentModel mCurrentEnvironment;
	private Vector mEnvironmentModels = new Vector();
	
	private IntValue mNumberOfSteps;
	private IntValue mCurrentStep;
	private IntValue mNumberOfTries;
	private BooleanValue mTerminateCurrentTry;
	
	private IntValue mCurrentTry;
	private IntValue mCurrentGeneration;
	private IntValue mCurrentIndividual;
	
	private Vector mGenerationListeners = new Vector();
	private Vector mTryListeners = new Vector();
	private Vector mStepListeners = new Vector();
	private Vector mIndividualListeners = new Vector();
	
	private boolean mStepRunning;
	private boolean mIndividualRunning;
	private boolean mTryRunning;
	private boolean mGenerationRunning;
	
	private EnvironmentModel mMasterEnvironment;
	
	private EvolutionLogger mEvolutionLogger;
	private StringValue mLogFileName;
	private DateFormat mDateFormatter;
	private int mStepsTillGarbageCollectionCounter;
	private IntValue mStepsTillGarbageCollectionValue;
	
	private Vector mListenerBufferVector = new Vector(); //used as buffer during notifications.
	
	private Vector mTerminateCurrentTryValues = new Vector();
	
	
	
	
	/**
	 * Constructor.
	 *
	 */
	public Scheduler(Simulator sim) {
		
		
		ValueManager vm = sim.getValueManager();
		mNumberOfSteps = vm.getIntValue("/Simulation/NumberOfSteps");
		if(mNumberOfSteps == null) {
			mNumberOfSteps = new IntValue();
			mNumberOfSteps.addValueReplacedListener(this);
			vm.addValue("/Simulation/NumberOfSteps", mNumberOfSteps);
		}
		mCurrentStep = vm.getIntValue("/Simulation/CurrentStep");
		if(mCurrentStep == null) {
			mCurrentStep = new IntValue();
			mCurrentStep.addValueReplacedListener(this);
			vm.addValue("/Simulation/CurrentStep", mCurrentStep);
		}
		mNumberOfTries = vm.getIntValue("/Simulation/NumberOfTries");
		if(mNumberOfTries == null) {
			mNumberOfTries = new IntValue();
			mNumberOfTries.addValueReplacedListener(this);
			vm.addValue("/Simulation/NumberOfTries", mNumberOfTries);
		}
		mCurrentTry = vm.getIntValue("/Simulation/CurrentTry");
		if(mCurrentTry == null) {
			mCurrentTry = new IntValue();
			mCurrentTry.addValueReplacedListener(this);
			vm.addValue("/Simulation/CurrentTry", mCurrentTry);
		}
		mCurrentGeneration = vm.getIntValue("/Simulation/CurrentGeneration");
		if(mCurrentGeneration == null) {
			mCurrentGeneration = new IntValue();
			mCurrentGeneration.addValueReplacedListener(this);
			vm.addValue("/Simulation/CurrentGeneration", mCurrentGeneration);
		}
		mCurrentIndividual = vm.getIntValue("/Simulation/CurrentIndividual");
		if(mCurrentIndividual == null) {
			mCurrentIndividual = new IntValue();
			mCurrentIndividual.addValueReplacedListener(this);
			vm.addValue("/Simulation/CurrentIndividual", mCurrentIndividual);
		}
		mTerminateCurrentTry = vm.getBooleanValue("/Simulation/TerminateCurrentTry");
		if(mTerminateCurrentTry == null) {
			mTerminateCurrentTry = new BooleanValue(false);
			mTerminateCurrentTry.addValueReplacedListener(this);
			vm.addValue("/Simulation/TerminateCurrentTry", mTerminateCurrentTry);
		}
		SimpleDateFormat df = new SimpleDateFormat("'EvoLog-'yyyy-MM-dd_HH-mm");
		mLogFileName = vm.getStringValue("/Simulation/LogFileName");
		if(mLogFileName == null) {
			mLogFileName = new StringValue(df.format(new Date(System.currentTimeMillis())) + ".elg");
			mLogFileName.addValueReplacedListener(this);
			vm.addValue("/Simulation/LogFileName", mLogFileName);
		}
		
		mStepsTillGarbageCollectionValue = new IntValue(100000);
		mStepsTillGarbageCollectionValue.addValueReplacedListener(this);
		vm.addValue("/Simulation/GarbageCollectionInterval", mStepsTillGarbageCollectionValue);
		
		mStepRunning = false;
		mIndividualRunning = false;
		mTryRunning = false;
		mGenerationRunning = false;
		
		mEvolutionLogger = new EvolutionLogger();
		mDateFormatter = DateFormat.getInstance();

		//add this object to the initialization and cleanUp process.
		sim.addSystemObject(this);
	}
	
	
	/**
	 * Sets the environmentModel that is used in the beginning of each generation
	 * to generate variations for each try.
	 * 
	 * @param environment the master EnvironmentModel.
	 */
	public void setMasterEnvironment(EnvironmentModel environment) {
		mMasterEnvironment = environment;
		if(mMasterEnvironment != null && !mMasterEnvironment.isInitialized()) {
			mMasterEnvironment.initEnvironment();
		}
	}
	
	/**
	 * Returns the master EnvironmentModel.
	 * 
	 * @return the master EnvironmentModel.
	 */
	public EnvironmentModel getMasterEnvironment() {
		return mMasterEnvironment;
	}
	
	
	/**
	 * Creates a number of variations of the master EnvironmentModel. These 
	 * variations should vary so that the simulations start with different 
	 * initial conditions.
	 * 
	 * @param numberOfTries the number of variations to create.
	 */
	public void createEnvironments(int numberOfTries) {
		if(mMasterEnvironment != null) {
			removeAllEnvironmentModels();
			Vector environments = mMasterEnvironment.createEnvironments(numberOfTries);
			for(int i = 0; i < environments.size(); ++i) { 
				EnvironmentModel env = (EnvironmentModel) environments.elementAt(i);
				addEnvironmentModel(env);
				if(i == 0) {
					mCurrentEnvironment = env;
				}
			}
		}
	}
	
	
	public boolean init() {
		return true;
	}
	
	public boolean cleanUp() {
		mEvolutionLogger.closeEvolutionLogger();
		return true;
	}
	
	
	/**
	 * Sets the currently used EnvironmentModel.
	 * @param model the model to use when execution simulation steps.
	 */
	public void setCurrentEnvironment(EnvironmentModel model) {
		mCurrentEnvironment = model;
		if(mCurrentEnvironment != null && mEnvironmentModels.isEmpty()) {
			mEnvironmentModels.add(model);
		}
	}
	
	/**
	 * Returns the currently used EnvironmentModel.
	 * @return the currently used EnvironmentModel.
	 */
	public EnvironmentModel getCurrentEnvironment() {
		return mCurrentEnvironment;
	}
	
	/**
	 * Adds an EnvironmentModel to the available models.
	 * Objects can be contained several times.
	 * 
	 * @param model the model to add.
	 * @return true if successful.
	 */
	public boolean addEnvironmentModel(EnvironmentModel model) {
		if(model == null) {
			return false;
		}
		mEnvironmentModels.addElement(model);
		return true;
	}
	
	/**
	 * Removes all available EnvironmentModels.
	 * Afterwards no EnvironmentModel is available.
	 *
	 */
	public void removeAllEnvironmentModels() {
		mEnvironmentModels.clear();
		mCurrentEnvironment = null;
	}
	
	/**
	 * Prepares the next generation. To do so all environment models for all
	 * tries in this generation are created. 
	 * If a try is running, then this try is completed and the listeners are notified.
	 * If an individual is running, then this individual is completed and the listeners are notified.
	 * If there is already a generation running, then this generation is completed and the are listeners notified.
	 * 
	 * After the environments are created for all tries in this generation, 
	 * all GenerationListeners are norified that the next generation was started.
	 * 
	 * @param generationId the id of the generation to start.
	 * @param numberOfTries the number of tries each individual in this generation will have.
	 */
	public void prepareNextGeneration(int generationId, int numberOfTries, int numberOfStepsPerTry) {
		
		if(mEvolutionLogger.isActive() && !mEvolutionLogger.getFileName().equals(mLogFileName.get())) {
			mEvolutionLogger.closeEvolutionLogger();
		}
		if(!mEvolutionLogger.isActive()) {
			mEvolutionLogger.startEvolutionLogger(mLogFileName.get());
		}
		
		notifyTryListeners(mCurrentTry.get(), false);
		notifyIndividualListeners(mCurrentIndividual.get(), false);
		notifyGenerationListeners(mCurrentGeneration.get(), false);
		
		if(generationId != mCurrentGeneration.get()) {
			mCurrentGeneration.set(generationId);
		}
		if(numberOfTries != mNumberOfTries.get()) {
			mNumberOfTries.set(numberOfTries);
		}
		if(numberOfStepsPerTry != mNumberOfSteps.get()) {
			mNumberOfSteps.set(numberOfStepsPerTry);
		}
		createEnvironments(numberOfTries);

		notifyGenerationListeners(generationId, true);
		
	}
	
	/**
	 * Prepares the next individual.
	 * This method should be followed by a call to prepareNextTry to initialize the first try
	 * for this individual. If not done then the current environment is not changed and
	 * no TryListeners are notified.
	 * 
	 * @param individualId
	 */
	public void prepareNextIndividual(int individualId) {
		
		
		notifyTryListeners(mCurrentTry.get(), false);
		notifyIndividualListeners(mCurrentIndividual.get(), false);
		
		mCurrentIndividual.set(individualId);
		
		if(mCurrentEnvironment != null) {
			mCurrentEnvironment.reset();
		}
		
		notifyIndividualListeners(mCurrentIndividual.get(), true);
	}
	
	/**
	 * Prepares the next try of an individuum.
	 * For this the next EnvironmentModel is chosen from the list of available EnvironmentModels
	 * and set as current EnvironmentModel.
	 * If there is no EnvironmentModel, then the current model is set to null, 
	 * which means that no execution takes place during execution steps.
	 * The chosen EnvironmentModel object is the one with index (tryId % getEnvironemtModels.size()).
	 *  
	 * @param tryId the current try.
	 */
	public void prepareNextTry(int tryId) {
		
		
		
		notifyTryListeners(mCurrentTry.get(), false);
		
		Simulator.getInstance().getFitnessManager().prepareNextTry(tryId);
		mCurrentTry.set(tryId);
		
		if(mTerminateCurrentTry.get()) {
			mTerminateCurrentTry.set(false);
		}
		
		//select next EnvironmentModel from the set of available EnvironmentModels
		if(!mEnvironmentModels.isEmpty()) {
			setCurrentEnvironment((EnvironmentModel) 
					mEnvironmentModels.elementAt(tryId % mEnvironmentModels.size()));
		}
		else {
			setCurrentEnvironment(null);
		}
		if(mCurrentEnvironment != null) {
			mCurrentEnvironment.reset();
		}
		
		//Simulator.getInstance().getFitnessManager().resetFitnessFunctions();
		
		notifyTryListeners(mCurrentTry.get(), true);
	}
	
	
	
	/**
	 * Executes the next simualtion step with the current EnvironmentModel.
	 * First the input of all controllers is updated. 
	 * Then the global events are executed.
	 * After executing the simulation step of the current EnvironmentModel,
	 * the output of all controllers is updated.
	 *
	 */
	public void executeSimulationStep(int stepId) {

		mStepsTillGarbageCollectionCounter++;
		mCurrentStep.set(stepId);
		
		notifyStepListeners(mCurrentStep.get(), true);
		
		if(mCurrentEnvironment != null) {
			mCurrentEnvironment.executeSimulationStep();
		}
		
		//calculate the fitnessfunctions if available.
		Simulator.getInstance().getFitnessManager().executeFitnessFunctions();
		
		notifyStepListeners(mCurrentStep.get(), false);
		
		//check if this try can be terminated.
		if(shouldCurrentTryBeTerminated() && !mTerminateCurrentTry.get()) {
			mTerminateCurrentTry.set(true);
		}
		
		int delay = Simulator.getInstance().getSimulationDelay();
		try {
			if(delay <= 0) {
				Thread.yield();
			}
			else {
				Thread.sleep(delay);
			}
			while(Simulator.getInstance().isPaused()) {
				Thread.sleep(100);
			}
		}
		catch(InterruptedException e) {
		}	
	}
	
	/**
	 * Makes sure that all tries, individuals and generations currently running
	 * are completed. This method should be called before the System is shut down.
	 */
	public void terminateSimulationRun() {
		notifyTryListeners(mCurrentTry.get(), false);
		notifyIndividualListeners(mCurrentIndividual.get(), false);
		notifyGenerationListeners(mCurrentGeneration.get(), false);
	}
	
	/**
	 * Emulates the execution of an evolution run.
	 * This is useful to start the simulator for tests during setup.
	 * 
	 * @param generations
	 * @param individuals
	 * @param tries
	 * @param steps
	 */
	public void emulateEvolutionRun(int generations, int individuals, int tries, int steps) {
		mNumberOfSteps.set(steps);
		mNumberOfTries.set(tries);
		
		boolean continueForever = false;
		if(generations == -1) {
			generations = 1;
			continueForever = true;
		}
		//TODO extend to execute an entire evolution. Rename then.
		for(int gen = 0; gen < generations; ++gen) {
			if(continueForever) {
				gen--;
			}
			prepareNextGeneration(gen, tries, steps);
			for(int ind = 0; ind < individuals; ++ind) {
				prepareNextIndividual(ind);
				for(int t = 0; t < mNumberOfTries.get(); ++t) {
					prepareNextTry(t);
					if(steps == -1) {
						//Run infinitely many steps
						int step = 0;
						while(true) {
							executeSimulationStep(step++);
						}
					}
					else {
						for(int step = 0; step < mNumberOfSteps.get(); ++step) {
							executeSimulationStep(step);
							if(mTerminateCurrentTry.get()) {
								break;
							}
						}
					}
				}
			}
		}
	}
	
	/**
	 * Returns the number of steps per try that are used for all tries within the current
	 * generation.
	 * 
	 * @return the maximal number of steps per try.
	 */
	public int getNumberOfSteps() {
		return mNumberOfSteps.get();
	}
	
	public IntValue getNumberOfStepsValue() {
		return mNumberOfSteps;
	}
	
	/**
	 * Returns the current step of the ongoing try.
	 * @return the current step.
	 */
	public int getCurrentStep() {
		return mCurrentStep.get();
	}
	
	/**
	 * Returns the number of tries that are executed for each individual within the current
	 * generation.
	 * @return the number of tries.
	 */
	public int getNumberOfTries() {
		return mNumberOfTries.get();
	}
	
	public IntValue getNumberOfTriesValue() {
		return mNumberOfTries;
	}
	
	/**
	 * Returns the current try of the currently executed individual.
	 * @return the current try.
	 */
	public int getCurrentTry() {
		return mCurrentTry.get();
	}
	
	public StringValue getLogFileName() {
		return mLogFileName;
	}
	
	/**
	 * Returns the id of the generation that is currently executed.
	 * @return the current generation.
	 */
	public int getCurrentGeneration() {
		return mCurrentGeneration.get();
	}
	
	public boolean isGenerationRunning() {
		return mGenerationRunning;
	}
	
	/**
	 * Returns the number of the currently evaluated individual.
	 * @return the id of the current individual.
	 */
	public int getCurrentIndividual() {
		return mCurrentIndividual.get();
	}
	
	/**
	 * Adds a GenerationListener.
	 * 
	 * @param listener the listener to add.
	 * @return true if successful, otherwise false. Returns also false if the listener is already registered.
	 */
	public synchronized boolean addGenerationListener(GenerationListener listener) {
		if(listener == null || mGenerationListeners.contains(listener)) {
			return false;
		}
		mGenerationListeners.add(listener);
		return true;
	}
	
	/**
	 * Removes a GenerationListener.
	 * @param listener the listener to remove.
	 * @return true if successful, if the listener was not contained, false is returned.
	 */
	public synchronized boolean removeGenerationListener(GenerationListener listener) {
		if(listener == null || !mGenerationListeners.contains(listener)) {
			return false;
		}
		mGenerationListeners.remove(listener);
		return true;
	}
	
	/**
	 * Notifies all registered NextGenerationListeners that a new generation was started.
	 * 
	 * @param generationId the id of the current generation.
	 */
	public synchronized  void notifyGenerationListeners(int generationId, boolean generationStarted) {
		if(generationStarted && !mGenerationRunning) {
			mGenerationRunning = true;
			mEvolutionLogger.logXML("<generation id=\"" + mCurrentGeneration.get() + "\">", "</generation>");
			mEvolutionLogger.logComment("Generation started at: " + mDateFormatter.format(new Date(System.currentTimeMillis())));
			
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mGenerationListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((GenerationListener) i.next()).generationStarted(generationId);
			}
		}
		else if(!generationStarted && mGenerationRunning) {
			mGenerationRunning = false;
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mGenerationListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((GenerationListener) i.next()).generationCompleted(generationId);
			}
			mEvolutionLogger.logComment("Generation completed at: " + mDateFormatter.format(new Date(System.currentTimeMillis())));
			mEvolutionLogger.closeTag();
		}
	}
	
	/**
	 * Adds a TryListener.
	 * 
	 * @param listener the listener to add.
	 * @return true if successful, otherwise false. Returns also false if the listener is already registered.
	 */
	public synchronized boolean addTryListener(TryListener listener) {
		if(listener == null || mTryListeners.contains(listener)) {
			return false;
		}
		mTryListeners.add(listener);
		return true;
	}
	
	/**
	 * Removes a TryListener.
	 * @param listener the listener to remove.
	 * @return true if successful, if the listener was not contained, false is returned.
	 */
	public synchronized boolean removeTryListener(TryListener listener) {
		if(listener == null || !mTryListeners.contains(listener)) {
			return false;
		}
		mTryListeners.remove(listener);
		return true;
	}
	
	/**
	 * Notifies all registered NextGenerationListeners that a new generation was started.
	 * 
	 * @param tryId the id of the current try.
	 */
	public synchronized void notifyTryListeners(int tryId, boolean tryStarted) {
		if(tryStarted && !mTryRunning) {
			mTryRunning = true;
			
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mTryListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((TryListener) i.next()).tryStarted(tryId);
			}
		}
		else if(!tryStarted && mTryRunning) {
			mTryRunning = false;
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mTryListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((TryListener) i.next()).tryCompleted(tryId);
			}
		}
	}
	
	/**
	 * Adds a StepListener.
	 * 
	 * @param listener the listener to add.
	 * @return true if successful, otherwise false. Returns also false if the listener is already registered.
	 */
	public synchronized boolean addStepListener(StepListener listener) {
		if(listener == null || mStepListeners.contains(listener)) {
			return false;
		}
		mStepListeners.add(listener);
		return true;
	}
	
	/**
	 * Removes a StepListener.
	 * @param listener the listener to remove.
	 * @return true if successful, if the listener was not contained, false is returned.
	 */
	public synchronized boolean removeStepListener(StepListener listener) {
		if(listener == null || !mStepListeners.contains(listener)) {
			return false;
		}
		mStepListeners.remove(listener);
		return true;
	}
	
	/**
	 * Notifies all registered StepListeners that a new generation was started.
	 * 
	 * @param stepId the id of the current step.
	 */
	public synchronized void notifyStepListeners(int stepId, boolean stepStarted) {
		if(stepStarted && !mStepRunning) {
			mStepRunning = true;
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mStepListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((StepListener) i.next()).stepStarted(stepId);
			}
		}
		else if(!stepStarted && mStepRunning) {
			mStepRunning = false;
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mStepListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((StepListener) i.next()).stepCompleted(stepId);
			}
		}
	}
	
	/**
	 * Adds a IndividualListener.
	 * 
	 * @param listener the listener to add.
	 * @return true if successful, otherwise false. Returns also false if the listener is already registered.
	 */
	public synchronized boolean addIndividualListener(IndividualListener listener) {
		if(listener == null || mIndividualListeners.contains(listener)) {
			return false;
		}
		mIndividualListeners.add(listener);
		return true;
	}
	
	/**
	 * Removes a IndividualListener.
	 * @param listener the listener to remove.
	 * @return true if successful, if the listener was not contained, false is returned.
	 */
	public synchronized boolean removeIndividualListener(IndividualListener listener) {
		if(listener == null || !mIndividualListeners.contains(listener)) {
			return false;
		}
		mIndividualListeners.remove(listener);
		return true;
	}
	
	/**
	 * Notifies all registered IndividualListeners that a new generation was started.
	 * 
	 * @param IndividualId the id of the current Individual.
	 */
	public synchronized void notifyIndividualListeners(int individualId, boolean individualStarted) {
		if(individualStarted && !mIndividualRunning) {
			mIndividualRunning = true;
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mIndividualListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((IndividualListener) i.next()).individualStarted(individualId);
			}
		}
		else if(!individualStarted && mIndividualRunning) {
			mIndividualRunning = false;
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mIndividualListeners);
			for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
				((IndividualListener) i.next()).individualCompleted(individualId);
			}
			if(mStepsTillGarbageCollectionCounter > mStepsTillGarbageCollectionValue.get()) {
				System.gc();
				mStepsTillGarbageCollectionCounter = 0;
			}
		}
	}
	
	/**
	 * Returns a Vector containing all currently available EnvironmentModels.
	 * @return all available EnvironmentModels.
	 */
	public Vector getEnvironmentModels() {
		return new Vector(mEnvironmentModels);
	}
	
	
	/**
	 * Adds a BooleanValue that is checkt after each executed step. If such a Value object
	 * return true, then the current try is marked to be terminated. Whether the try
	 * really is terminated is up the the controller of the evolution.
	 * 
	 * The standard implementation of the Scheduler continues with the next 
	 * try if at least one of the registered BooleanValues are true.
	 * 
	 * It is also save to set the global termination value to true manually.
	 * This global value can be found in the ValueManager registered with the name
	 * "/Simulation/TerminateCurrentTry"
	 * 
	 * @param terminationValue the BooleanValue to add to the termiantion indicators.
	 * @return true is successful, false if the Value is already contained.
	 */
	public synchronized boolean addTerminateTryValue(BooleanValue terminationValue) {
		if(terminationValue == null || mTerminateCurrentTryValues.contains(terminationValue)) {
			return false;
		}
		mTerminateCurrentTryValues.add(terminationValue);
		return true;
	}
	
	
	/**
	 * Removes a previously added try termination indicator.
	 * 
	 * @param value the Value object to remove.
	 * @return true if successful, false if not contained.
	 */
	public synchronized boolean removeTerminateTryValue(BooleanValue value) {
		if(value == null || !mTerminateCurrentTryValues.contains(value)) {
			return false;
		}
		mTerminateCurrentTryValues.remove(value);
		return true;
	}
	
	
	/**
	 * Returns true if the internal termination Value or one of the external BooleanValues,
	 * added with method addTermianteTryValue, are set to true.
	 * 
	 * @return true if the current try should be terminated, otherwise false.
	 */
	public synchronized boolean shouldCurrentTryBeTerminated() {
		if(mTerminateCurrentTry.get()) {
			return true;
		}
		for(Iterator i = mTerminateCurrentTryValues.iterator(); i.hasNext();) {
			if(((BooleanValue) i.next()).get()) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Sets the flag to termiante the current try to true.
	 * The next time step when the termination flag is checkt, the try will be terminated.
	 */
	public void terminateCurrentTry() {
		mTerminateCurrentTry.set(true);
	}
	
	/**
	 * Returns the EvolutionLogger that can be used to log parameters and results
	 * of the Evolution.
	 * @return
	 */
	public EvolutionLogger getEvolutionLogger() {
		return mEvolutionLogger;
	}

	

	public void valueReplaced(Value oldValue, Value newValue) {
		mNumberOfSteps = (IntValue) ValueManager.synchronizeValue(mNumberOfSteps, oldValue, newValue, this);
		mCurrentStep = (IntValue) ValueManager.synchronizeValue(mCurrentStep, oldValue, newValue, this);
		mNumberOfTries = (IntValue) ValueManager.synchronizeValue(mNumberOfTries, oldValue, newValue, this);
		mTerminateCurrentTry = (BooleanValue) ValueManager.synchronizeValue(mTerminateCurrentTry, oldValue, newValue, this);
		
		mCurrentTry = (IntValue) ValueManager.synchronizeValue(mCurrentTry, oldValue, newValue, this);
		mCurrentGeneration = (IntValue) ValueManager.synchronizeValue(mCurrentGeneration, oldValue, newValue, this);
		mCurrentIndividual = (IntValue) ValueManager.synchronizeValue(mCurrentIndividual, oldValue, newValue, this);
		mLogFileName = (StringValue)  ValueManager.synchronizeValue(mLogFileName, oldValue, newValue, this);
		mStepsTillGarbageCollectionValue = (IntValue) ValueManager.synchronizeValue(mStepsTillGarbageCollectionValue, oldValue, newValue, this);
	}
	
	
	

}
