package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.awt.Polygon;
import java.awt.geom.Point2D;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class Beacon extends Box implements LightSource {

	private int[] mBlinkPattern = new int[0];
	private int mBlinkIndex;
	private int mCounter;
	private double mMaxChangePerCycle;
	private DoubleValue mBrightness;
	private CircularPolygon mLightArea;
	private boolean mSwitchedOn;
	private double mInitBrightness;
	

	public Beacon(String name, double x, double y, double angle, 
			double lightRadius, int corners, double lightAngle, 
			double maxChangePerCycle, double initBrightness) 
	{
		super(name, x, y, angle, 10.0, 10.0);
		
		setShapeColor(Color.red);
		
		mLightArea = new CircularPolygon(getName() + "/LightArea", 0.0, 0.0, 0.0, lightRadius, corners, lightAngle);
		mLightArea.setShapeColor(Color.YELLOW);
		mLightArea.setShapeTransparency(0.5);
		
		mInitBrightness = initBrightness;
		mBlinkIndex = 0;
		mCounter = 0;
		mBrightness = new DoubleValue(initBrightness);
		mMaxChangePerCycle = maxChangePerCycle;
		mLightArea.setReferenceObject(this);
		mSwitchedOn = (initBrightness > 0.0) ? true : false;
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + getName() + "/Brightness", mBrightness);
		mBrightness.addValueReplacedListener(this);
	}
	
	

	public void setActive(boolean active) {
		super.setActive(active);
	}



	/**
	* Calculates the current brightness of the Beacon.
	*/
	public boolean update() {
		boolean ok = super.update();
		
		if(mBlinkPattern != null && mBlinkPattern.length > 0) {
			if(mCounter <= 0) {
				mBlinkIndex++;
				if(mBlinkIndex >= mBlinkPattern.length) {
					mBlinkIndex = 0;
				}
				mCounter = mBlinkPattern[mBlinkIndex];
				//switch activity.
				mSwitchedOn = !mSwitchedOn;
				if(mCounter <= 0.0) {
					return true;
				}
			}
			mCounter--;
		}
		double brightness = mBrightness.get();
		if(mSwitchedOn) {
			if(brightness == 0.0) {
				mLightArea.setActive(true);
			}
			brightness += mMaxChangePerCycle;
			if(brightness > 1.0) {
				brightness = 1.0;
			}
			mBrightness.set(brightness);
		}
		else if(!mSwitchedOn) {
			brightness -= mMaxChangePerCycle;
			if(brightness <= 0.0) {
				brightness = 0.0;
				mLightArea.setActive(false);
			}
			mBrightness.set(brightness);
		}
		mLightArea.setShapeTransparency(brightness / 2.0);
		return ok;
	}

	//TODO
	public void reset() {
		super.reset();
		mBrightness.set(mInitBrightness);
		if(mBrightness.get() > 0.0) {
			mSwitchedOn = true;
		}
		else {
			mSwitchedOn = false;
		}
		if(isActive()) {
			mLightArea.setActive(mSwitchedOn);
		}
		else {
			mLightArea.setActive(false);
		}
		mLightArea.setShapeTransparency(mBrightness.get() / 2.0);
		mBlinkIndex = 0;
		mCounter = 0;
	}

	/**
	* This method is called whenever this SimObject was added to an EnvironmentModel.
	* Subclasses of SimObject can used this method to register sub components at the
	* EnvironmentModel.
	* 
	* @param model the EnvironmentModel where the child SimObjects should be registered.
	*/
	public void addChildren(EnvironmentModel model) {
		super.addChildren(model);
		model.addSimObject(mLightArea);
	}


	public void setBlinkPattern(int[] blinkPattern) {
		mBlinkPattern = blinkPattern;
		if(mBlinkPattern.length > 0) {
			mBlinkIndex = 0;
			mCounter = mBlinkPattern[mBlinkIndex];
		}
	}

	/**
	* Returns a copy of the BlinkPattern.
	* @return a copy of the BlinkPattern.
	*/
	public int[] getBlinkPattern() {
		int[] pattern = new int[mBlinkPattern.length];
		for(int i = 0; i < mBlinkPattern.length; ++i) {
			pattern[i] = mBlinkPattern[i];
		}
		return pattern;
	}

	/**
	* Returns the child component that indicates the bulb of the light source.
	* @return the bulb child component.
	*/
	public CircularPolygon getLightArea() {
		return mLightArea;
	}
	
	/**
	 * Sets the current brightness of the light area of this LightSource.
	 */
	public void setInitialBrightness(double brightness) {
		mInitBrightness = Math.max(Math.min(1.0, brightness), 0.0);
		
		mBrightness.set(mInitBrightness);
		if(mBrightness.get() > 0.0) {
			mSwitchedOn = true;
		}
		else {
			mSwitchedOn = false;
		}
		if(isActive()) {
			mLightArea.setActive(mSwitchedOn);
		}
		else {
			mLightArea.setActive(false);
		}
		mLightArea.setShapeTransparency(mBrightness.get() / 2.0);
		
	}

	/**
	* Returns the brightness of the LightSource at the given point.
	* @param globalPosition the position to check for brightness in global coordinates.
	* @return the brightness at the given global position.
	*/
	public double getBrightness(Point2D.Double globalPosition) {
		if(!mLightArea.isActive() || !isActive()) {
			return 0.0;
		}
		Polygon globalShape = mLightArea.getGlobalShape();
		if(globalShape.contains(globalPosition)) {
			return mBrightness.get();
		}
		return 0.0;
	}

	/**
	* Called whenever an observed Value object is replaced by another one in the 
	* ValueManager. The object can then change its observatoin to the new Value object
	* and cancel the observation of the old Value.
	*/
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mBrightness = (DoubleValue) synchronizeValue(mBrightness, oldValue, newValue);
	}

}
