package de.fhg.iais.outdoor.simba.devices;

import java.awt.Polygon;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class Box extends SimObject {

	
	protected DoubleValue mWidth;
	protected DoubleValue mHeight;
	
	
	public Box(String name, double x, double y, double angle, double width, double height) {
		this(name, x, y, angle, width, height, true);
	}
	
	/**
	 * Constructor.
	 * 
	 * @param name the name of the Box
	 * @param x the initial x coordinate
	 * @param y the initial y coorindate
	 * @param angle the initial angle
	 * @param width the width of the box
	 * @param height the height of the box.
	 */
	public Box(String name, double x, double y, double angle, double width, double height,
			boolean registerAtValueManager) 
	{
		super(name, x, y, angle, new Polygon(), registerAtValueManager);
		mWidth = new DoubleValue(width);
		mHeight = new DoubleValue(height);
		
		mWidth.addValueChangedListener(this);
		mHeight.addValueChangedListener(this);
		
		if(registerAtValueManager) {
			ValueManager vm = Simulator.getInstance().getValueManager();
			vm.addValue("/SimObjects/" + mName + "/Width", mWidth);
			vm.addValue("/SimObjects/" + mName + "/Height", mHeight);
		}
		
		updateShape();
	}
	
	
	

	public boolean update() {
		boolean ok = super.update();
		return ok;
	}
	
	public void reset() {
		super.reset();
	}
	
	/**
	 * Returns the width of this objet.
	 * 
	 * @return the width.
	 */
	public double getWidth() {
		return mWidth.get();
	}
	
	/**
	 * Returns the height of this object.
	 * 
	 * @return the height.
	 */
	public double getHeight() {
		return mHeight.get();
	}
	
	/**
	 * Sets width and height of the box. 
	 * @param width
	 * @param height
	 */
	public void setSize(double width, double height) {
		mWidth.set(width);
		mHeight.set(height);
		updateShape();
	}
	
	/**
	 * Sets the width of this SimObject.
	 * @param width the new widht.
	 */
	public void setWidth(double width) {
		mWidth.set(width);
		updateShape();
	}
	
	/**
	 * Sets the height of this SimObject.
	 * @param height the new height.
	 */
	public void setHeight(double height) {
		mHeight.set(height);
		updateShape();
	}
	
	/**
	 * Updates the shape Polygon with the current width and height.
	 */
	protected void updateShape() {
		double width = mWidth.get();
		double height = mHeight.get();
		
		mShape = new Polygon();
		mShape.addPoint((int) Math.round(-1 * width/2), (int) Math.round(height/2));
		mShape.addPoint((int) Math.round(width/2), (int) Math.round(height/2));
		mShape.addPoint((int) Math.round(width/2), (int) Math.round(-1 * height/2));
		mShape.addPoint((int) Math.round(-1 * width/2), (int) Math.round(-1 * height/2));
	}
	
	public void valueChanged(Value value) {
		super.valueChanged(value);
		if(value == null) {
			return;
		}
		if(value == mWidth || value == mHeight) {
			updateShape();
		}
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mWidth = (DoubleValue) ValueManager.synchronizeValue(mWidth, oldValue, newValue, this);
		mHeight = (DoubleValue) ValueManager.synchronizeValue(mHeight, oldValue, newValue, this);
	}

	/**
	 * Creates and returns a line that is based on the given parameters.
	 * The mid point of the is followd a distance of width / 2 in the direction of the angle.
	 * Then the line perpendicular to this vector is calculated. Its length is specified by parameter
	 * height. 
	 * The lines can be calculated with angle, angle +45, angle + 180, angle + 225 to get
	 * all four possible lines around a SimObject center point. At angles +45 and +225, width and 
	 * height have to be exchanged to get the correct lines.
	 * Examples:
	 * createLine(Point(0, 0), 2.0, 4.0, 0.0) => Line: (1.0, 2.0, 1.0, -2.0)
	 * createLine(Point(0, 0), 2.0, 4.0, 180.0) => Line: (-1.0, -2.0, -1.0, 2.0)
	 * createLine(Point(0, 0), 4.0, 2.0, 90.0) => Line: (-1.0, 2.0, 1.0, 2.0) //width and height exchanged
	 * createLine(Point(0, 0), 4.0, 2.0, 270.0) => Line: (1.0, -2.0, -1.0, -2.0) //width and height exchanged
	 * 
	 * @param mid a point describing the center point of a SimObject.
	 * @param width the width of the center object
	 * @param height the height of the center object.
	 * @param angle the angle of line spacing direction (line direction will be perpendicular to this angle)
	 * @return a line object with a distance 'width' to the 'mid' point, perpendicular to 'angle' with a size 'height'.
	 */
	public Line2D.Double createLine(Point2D.Double mid, double width, double height, double angle) {
		
		Point2D.Double lineMid = new Point2D.Double(
				mid.x + (Math.cos(angle * Math.PI / 180.0) * width / 2.0),
				mid.y + (Math.sin(angle * Math.PI / 180.0) * width / 2.0));
		
		
		return new Line2D.Double(
				lineMid.x + (Math.cos((angle + 90.0) * Math.PI / 180.0) * height / 2.0),
				lineMid.y + (Math.sin((angle + 90.0) * Math.PI / 180.0) * height / 2.0),
				lineMid.x + (Math.cos((angle - 90.0) * Math.PI / 180.0) * height / 2.0),
				lineMid.y + (Math.sin((angle - 90.0) * Math.PI / 180.0) * height / 2.0));
	}
	
	
	
}
