package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.GeometricMath;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.SimObjectSet;


public class DistanceScanner extends DistanceSensor {


	protected double mCurrentDistance;

	/**
	 * Constructor.
	 * @param x
	 * @param y
	 * @param angle
	 * @param minScanDistance
	 * @param maxScanDistance
	 * @param minValue
	 * @param maxValue
	 * @param posDirection
	 */
	public DistanceScanner(String name, double x, double y, double angle, 
			double minScanDistance, double maxScanDistance) 
	{
		super(name, x, y, angle, minScanDistance, maxScanDistance);
		
	}
	
	public DistanceScanner(String name, double x, double y, double angle, 
			double minScanDistance, double maxScanDistance, boolean treatObstaclesAsMassive) 
	{
		super(name, x, y, angle, minScanDistance, maxScanDistance, treatObstaclesAsMassive);
		
	}
	

	/**
	 * Calculates the distance to the closest obstacle.
	 * 
	 */
	public boolean update() {
		boolean ok = super.update();
		if(!isActive()) {
			if(getSensorReading() != mMaxScanDistance.get()) {
				mDistance.set(mMaxScanDistance.get());
			}
			return ok;
		}
	
		mCurrentDistance = mMaxScanDistance.get();
		Point2D.Double ownPosition = getGlobalPosition();
		
		Line2D.Double scannerLine = getGlobalScanLine();
		
		SimObjectSet obstacles = mObstacles;
		int setIndex = 0;
		do {
			for(Iterator i = obstacles.getIterator(); i.hasNext();) {
				SimObject obs = (SimObject) i.next();
				if(!obs.isActive() || mIgnorableObjects.contains(obs)) {
					continue;
				}
				//if the sensor is contained in one of the obstacle objects, then
				//the scanner should deliver the minimal measureable value.
				//This is only true if the objects are treated as beeing massive.
				if(mTreatObstaclesAsMassive && obs.getGlobalShape().contains(getGlobalPosition())) {
					mDistance.set(Math.max(Math.min(addNoise(mMinScanDistance.get()), mMaxScanDistance.get()), mMinScanDistance.get()));
					return ok;
				}
				
				//Check if the shape of the object has a chance of beeing intersecting.
				if(!checkIfObjectsMightIntersect(obs.getGlobalShape(), scannerLine)) {
					continue;
				}
				
				//Check distances for all lines of the polygon shape of the SimObject.
				Vector lines = obs.getShapeAsGlobalLineSegments();
				
				for(Iterator it = lines.iterator(); it.hasNext();) {
					Line2D.Double line = (Line2D.Double) it.next();
					checkDistance(line, scannerLine, ownPosition);
				}
			}
			if(setIndex < mObstacleLists.size()) {
				obstacles = (SimObjectSet) mObstacleLists.elementAt(setIndex);
				setIndex++;
			}
			else {
				obstacles =  null;
			}
		}
		while(obstacles != null);
		
		double dist = Math.max(Math.min(addNoise(mCurrentDistance), mMaxScanDistance.get()), mMinScanDistance.get());

		mDistance.set(Math.max(0.0, dist)); //do not allow distances less than zero.

		
		return ok;
	}
	
	
	
	public void paintSelf(Graphics2D g) {
		Color c = g.getColor();
		double distance = mDistance.get();
		if(distance >= mMinScanDistance.get() && distance < mMaxScanDistance.get()) {
			if(mObstacleDetectedColor == null) {
				g.setColor(mShapeColor);
			}
			else {
				g.setColor(mObstacleDetectedColor);
			}
		}
		else {
			g.setColor(mShapeColor);
		}
		g.drawLine(0, 0, (int) Math.round(distance), 0);
		g.setColor(c);
	}
	
	
	
	
	
	/**
	 * Returns a line that corresponds to the scanner (angle, length) in global coordinates.
	 * 
	 * @return the scanner line.
	 */
	public Line2D.Double getGlobalScanLine() {
		
		Point2D.Double pos = getGlobalPosition();
		double ang = getGlobalAngle();
		
		Line2D.Double scanner = new Line2D.Double(
				pos.x, 
				pos.y, 
				pos.x + (Math.cos(ang * Math.PI / 180.0) * mMaxScanDistance.get()), 
				pos.y + (Math.sin(ang * Math.PI / 180.0) * mMaxScanDistance.get()));
		
		return scanner;
	}
	
	
	
	
	/**
	 * Checks if the scanner intersects the bounding box of the given shape. 
	 * This allows to check in advance whether the scanner has a chance to 
	 * intersect with one of the shape lines. If this method returns false, then
	 * the scanner does not need to check all lines of a potentially complicated
	 * shape. 
	 * 
	 * @param shape the shape to check 
	 * @return
	 */
	protected boolean checkIfObjectsMightIntersect(Polygon shape, Line2D.Double scanner) {
		
		if(scanner.intersects(shape.getBounds())) {
			return true;
		}
		return false;
	}
	
	
	/**
	 * Checks whether the line object has a closer intersection point than measured
	 * up to now. If so then the new value is the new closest measure.
	 * 
	 * @param line the line to check for an intersection with the scanner line.
	 */
	private void checkDistance(Line2D.Double line, Line2D.Double scanner, Point2D.Double ownPosition) {
		Point2D.Double intersection = new Point2D.Double();
		
		//check for all four possible lines around a standard SimObject (Box). 
		if(GeometricMath.getIntersectionPoint(line, scanner, intersection)) {
			double dist = ownPosition.distance(intersection);
			if(dist < mCurrentDistance) {
				if(dist < mMinScanDistance.get()) {
					mCurrentDistance = mMinScanDistance.get();
				}
				else {
					mCurrentDistance = dist;
				}
			}
		}
	}
	
	
		
	
	
}
