package de.fhg.iais.outdoor.simba.devices;

import java.awt.Polygon;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.value.ValueReplacedListener;

public class OmniDrive extends SimObject implements ValueReplacedListener  {

	private DoubleValue mVelocityX;
	private DoubleValue mVelocityY;
	private DoubleValue mAngularVelocity;
	private double mMinVelocity;
	private double mMaxVelocity;
	private double mMinAngularVelocity;
	private double mMaxAngularVelocity;
	
	
	public OmniDrive(String name, double x, double y, double angle) {
		this(name, x, y, angle, -5.0, 5.0, -5.0, 5.0);
	}
	
	public OmniDrive(String name, double x, double y, double angle, 
			double minVelocity, double maxVelocity, 
			double minAngularVelocity, double maxAngularVelocity)
	{
		super(name, x, y, angle, new Polygon());
		mVelocityX = new DoubleValue(0.0);
		mVelocityY = new DoubleValue(0.0);
		mAngularVelocity = new DoubleValue(0.0);
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + name + "/VelocityX", mVelocityX);
		vm.addValue("/SimObjects/" + name + "/VelocityY", mVelocityY);
		vm.addValue("/SimObjects/" + name + "/AngularVelocity", mAngularVelocity);
		
		mVelocityX.addValueReplacedListener(this);
		mVelocityY.addValueReplacedListener(this);
		mAngularVelocity.addValueReplacedListener(this);
		
		mMinVelocity = minVelocity;
		mMaxVelocity = maxVelocity;
		mMinAngularVelocity = minAngularVelocity;
		mMaxAngularVelocity = maxAngularVelocity;
	}
	
	
	/**
	 * Moves this OmniDrive object to the new position that is determined
	 * by the two perpendicular velocities x and y and the angular velocity.
	 * The translation of the linear velocities is applied before the 
	 * angular velocity.
	 * 
	 */
	public boolean update() {
		boolean ok = super.update();
		
		AffineTransform transformation = new AffineTransform();
		transformation.rotate(Math.toRadians(mAngle.get()));
		
		Point2D.Double newPos = new Point2D.Double(mVelocityX.get(), mVelocityY.get());
		Point2D.Double translatedPos = new Point2D.Double();
		transformation.transform(newPos, translatedPos);
		
		Point2D.Double position = mPosition.get();
		translatedPos.x += position.x;
		translatedPos.y += position.y;
		
		setPosition(translatedPos);
		mAngle.set(mAngle.get() + mAngularVelocity.get());
		
		return ok;
	}
	
	/**
	 * Sets the velocity in relative x direction.
	 * @param vel the velocity in x direction.
	 */
	public void setVelocityX(double vel) {
		if(vel > mMaxVelocity) {
			vel = mMaxVelocity;
		}
		else if(vel < mMinVelocity) {
			vel = mMinVelocity;
		}
		mVelocityX.set(vel);
	}
	
	/**
	 * Returns the current velocity in x direction.
	 * @return the velocity in x direction.
	 */
	public double getVelocityX() {
		return mVelocityX.get();
	}
	
	/**
	 * Sets the velocity in relative y direction.
	 * @param vel the velocity in y direction.
	 */
	public void setVelocityY(double vel) {
		if(vel > mMaxVelocity) {
			vel = mMaxVelocity;
		}
		else if(vel < mMinVelocity) {
			vel = mMinVelocity;
		}
		mVelocityY.set(vel);
	}
	
	/**
	 * Returns the current velocity in y direction.
	 * @return the velocity in y direction.
	 */
	public double getVelocityY() {
		return mVelocityY.get();
	}
	
	/**
	 * Sets the angular velocity.
	 * @param angVel the angular velocity.
	 */
	public void setAngularVelocity(double angVel) {
		if(angVel > mMaxAngularVelocity) {
			angVel = mMaxAngularVelocity;
		}
		else if(angVel < mMinAngularVelocity) {
			angVel = mMinAngularVelocity;
		}
		mAngularVelocity.set(angVel);
	}
	
	/**
	 * Returns the current angular velocity.
	 * @return the angular velocity.
	 */
	public double getAngularVelocity() {
		return mAngularVelocity.get();
	}
	
	
	
	public double getMinVelocity() {
		return mMinVelocity;
	}
	
	public double getMaxVelocity() {
		return mMaxVelocity;
	}
	
	public double getMinAngularVeloctiy() {
		return mMinAngularVelocity;
	}
	
	public double getMaxAngularVelocity() {
		return mMaxAngularVelocity;
	}
	
	/**
	 * Called whenever an observed Value object is replaced by another one in the 
	 * ValueManager. The object can then change its observatoin to the new Value object
	 * and cancel the observation of the old Value.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mVelocityX = (DoubleValue) synchronizeValue(mVelocityX, oldValue, newValue);
		mVelocityY = (DoubleValue) synchronizeValue(mVelocityY, oldValue, newValue);
		mAngularVelocity = (DoubleValue) synchronizeValue(mAngularVelocity, oldValue, newValue);
	}
	
}
