package de.fhg.iais.outdoor.simba.devices;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.LinkedList;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class PathDrawer extends SimObject {
	
	private LinkedList mTrajectory = new LinkedList();
	private IntValue mInterval;
	private BooleanValue mRecordTrajectory;
	private IntValue mMaxTrajectoryLength;
	private IntValue mZoneInterval;
	private Color mSecondColor;
	private int mIntervalCounter;
	private int mPointCounter;
	
	public PathDrawer(String name, double x, double y, int interval, int zoneInterval, int maxTrajectoryLength) {
		super(name, x, y, 0.0, new Polygon());
		
		
		mInterval = new IntValue(interval);
		mRecordTrajectory = new BooleanValue(true);
		mMaxTrajectoryLength = new IntValue(maxTrajectoryLength);
		mZoneInterval = new IntValue(zoneInterval);
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + getName() + "/Interval", mInterval);
		mInterval.addValueReplacedListener(this);
		vm.addValue("/SimObjects/" + getName() + "/RecordTrajectory", mRecordTrajectory);
		mRecordTrajectory.addValueReplacedListener(this);
		vm.addValue("/SimObjects/" + getName() + "/MaxTrajectoryLength", mMaxTrajectoryLength);
		mMaxTrajectoryLength.addValueReplacedListener(this);
		vm.addValue("/SimObjects/" + getName() + "/ZoneInterval", mZoneInterval);
		mZoneInterval.addValueReplacedListener(this);
		
		mSecondColor = getShapeColor();
		
		
		
	}
	
	public void setSecondColor(Color color) {
		mSecondColor = color;
	}

	public void paintSelf(Graphics2D g) {
		Color c = g.getColor();
		Composite composite = g.getComposite();
		g.setColor(mShapeColor);
		g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float) mTransparency.get()));
		
		try {
			AffineTransform transform = getAffineTransform().createInverse();
			
			Point2D.Double lastPoint = new Point2D.Double(0.0, 0.0);
			Point2D.Double point = new Point2D.Double();
			
			synchronized(mTrajectory) {
				if(mTrajectory.size() > 0) {
					int pointNumber = mPointCounter - mTrajectory.size();
					boolean firstPoint = true;
					if(((pointNumber / mZoneInterval.get()) % 2) == 0) {
						g.setColor(getShapeColor());
					}
					else {
						g.setColor(mSecondColor);
					}
					
					for(Iterator i = mTrajectory.iterator(); i.hasNext();) {
						pointNumber++;
						if(pointNumber % mZoneInterval.get() == 0) {
							if(g.getColor() == getShapeColor()) {
								g.setColor(mSecondColor);
							}
							else {
								g.setColor(getShapeColor());
							}
						}
						Point2D.Double nextPoint = (Point2D.Double) i.next();
						transform.transform(nextPoint, point);
						if(!firstPoint) {
							g.drawLine((int) lastPoint.x, (int) lastPoint.y, (int) point.x, (int) point.y);
						}
						firstPoint = false;
						lastPoint = new Point2D.Double(point.x, point.y);
					}
					g.drawLine((int) lastPoint.x, (int) lastPoint.y, 0, 0);
				}
			}
		} catch (NoninvertibleTransformException e) {
			e.printStackTrace();
		}
		
		g.setComposite(composite);
		g.setColor(c);
	}

	public void reset() {
		super.reset();
		synchronized(mTrajectory) {
			mTrajectory.clear();
		}
		mIntervalCounter = 0;
		mPointCounter = 0;
	}

	public boolean update() {
		boolean ok = super.update();
		mIntervalCounter++;
		
		if(mRecordTrajectory.get() && (mIntervalCounter % mInterval.get() == 0)) {
			Point2D.Double pos = getGlobalPosition(); 
			synchronized(mTrajectory) {
				mTrajectory.addLast(new Point2D.Double(pos.x, pos.y));
				mPointCounter++;
				while(mTrajectory.size() > mMaxTrajectoryLength.get()) {
					mTrajectory.removeFirst();
				}
			}
		}
		
		return ok;
	}

	public void valueChanged(Value value) {
		super.valueChanged(value);
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mInterval = (IntValue) ValueManager.synchronizeValue(mInterval, oldValue, newValue, this);
		mRecordTrajectory = (BooleanValue) ValueManager.synchronizeValue(mRecordTrajectory, oldValue, newValue, this);
		mMaxTrajectoryLength = (IntValue) ValueManager.synchronizeValue(mMaxTrajectoryLength, oldValue, newValue, this);
		mZoneInterval = (IntValue) ValueManager.synchronizeValue(mZoneInterval, oldValue, newValue, this);
	}
	
	
	
}
