package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.Sensor;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.SimObjectSet;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class SimpleBumpDetector extends Sensor {
	
	protected BooleanValue mBumpDetected;
	protected NormalizedDoubleValue mBumpDetectedDouble;
	protected SimObjectSet mObstacles = new SimObjectSet();
	protected Vector mObstacleLists = new Vector();
	
	public SimpleBumpDetector(String name, double x, double y, double angle) {
		this(name, x, y, angle, new Polygon());
	}
	
	public SimpleBumpDetector(String name, double x, double y, double angle, Polygon shape) {
		super(name, x, y, angle, shape);
		
		if(shape.npoints == 0) {
			shape.addPoint(0, 0);
			setShape(shape);
		}
		
		mBumpDetected = new BooleanValue(false);
		mBumpDetectedDouble = new NormalizedDoubleValue(0.0, 0.0, 1.0, -1.0, 1.0);
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + mName + "/BumpDetected", mBumpDetected);
		vm.addValue("/SimObjects/" + mName + "/BumpDetectedDouble", mBumpDetectedDouble);
		
		mBumpDetected.addValueReplacedListener(this);
		mBumpDetectedDouble.addValueReplacedListener(this);
		
		mShapeColor = Color.RED;
	}
	
	/**
	 * Calculates the distance to the closest obstacle.
	 * 
	 */
	public boolean update() {
		boolean ok = super.update();
		
		Polygon ownShape = getGlobalShape();
		
		SimObjectSet obstacles = mObstacles;
		int setIndex = 0;
		do {
			for(Iterator i = obstacles.getIterator(); i.hasNext();) {
				SimObject obj = (SimObject) i.next();
				
				if(!obj.isActive()) {
					continue;
				}
				Polygon shape = obj.getGlobalShape();
				
				for(int j = 0; j < ownShape.npoints; ++j) {
					if(shape.contains(ownShape.xpoints[j], ownShape.ypoints[j])) {
						mBumpDetected.set(true);
						mBumpDetectedDouble.set(1.0);
						return ok;
					}
				}
			}
			if(setIndex < mObstacleLists.size()) {
				obstacles = (SimObjectSet) mObstacleLists.elementAt(setIndex);
				setIndex++;
			}
			else {
				obstacles =  null;
			}
		}
		while(obstacles != null);
		
		mBumpDetected.set(false);
		mBumpDetectedDouble.set(0.0);
		return ok;
	}
	
	/**
	 * Returns the last measured value of this sensor.
	 * @return
	 */
	public double getSensorReading() {
		return mBumpDetectedDouble.get();
	}
	
	/**
	 * Paints this SimObject on a graphics context in local coordinates.
	 */
	public void paintSelf(Graphics2D g) {
		if(mBumpDetected.get()) {
			Color c = g.getColor();
			g.setColor(mShapeColor);
			if(mShape.npoints > 1) {
				g.drawPolygon(mShape);
			}
			else {
				g.drawOval(-4, -4, 8, 8);
			}
			g.setColor(c);
		}
		
	}
	
	/**
	 * Adds an obstacle this bump detector considers during its update.
	 * 
	 * @param obstacle an obstacle that can be detected by this scanner.
	 * @return true if the obstacle object could be added.
	 */
	public boolean addObstacle(SimObject obstacle) {
		if(mObstacles.add(obstacle)) {
			addDependency(obstacle);
			return true;
		}
		return false;
	}
	
	/**
	 * Adds an external set of SimObjects that are considered as obstacles.
	 * 
	 * @param obstacles the SimObjectSet to add.
	 * @return
	 */
	public boolean addObstacleSet(SimObjectSet obstacles) {
		if(obstacles == null || mObstacleLists.contains(obstacles)) {
			return false;
		}
		mObstacleLists.add(obstacles);
		return true;
	}
	
	/**
	 * Removes a previously added SimObjectSet.
	 * @param obstacles
	 * @return
	 */
	public boolean removeObstacleList(SimObjectSet obstacles) {
		if(obstacles == null || !mObstacleLists.contains(obstacles)) {
			return false;
		}
		mObstacleLists.remove(obstacles);
		return true;
	}
	
	/**
	 * Returns a copy of the obstacle list of this bump detector.
	 * 
	 * @return a copy of the obstacle list.
	 */
	public Vector getObstacles() {
		return mObstacles.getElements();
	}
	
	public BooleanValue getSensorValueObject() {
		return mBumpDetected;
	}
	
	public NormalizedDoubleValue getSensorDoubleValue() {
		return mBumpDetectedDouble;
	}
	
	/**
	 * Called whenever an observed Value object is replaced by another one in the 
	 * ValueManager. The object can then change its observatoin to the new Value object
	 * and cancel the observation of the old Value.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mBumpDetected = (BooleanValue) synchronizeValue(mBumpDetected, oldValue, newValue);
		mBumpDetectedDouble = (NormalizedDoubleValue) synchronizeValue(mBumpDetectedDouble, oldValue, newValue);
	}

}
