package de.fhg.iais.outdoor.simba.evolution;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class EvolutionIndividual {

	private static int mIndividualIds = 0;
	
	private Object mGenom;
	private int mId;
	private int mAge;
	private Vector mFitnessValues = new Vector();
	private Vector mParents = new Vector();
	private Hashtable mProperties = new Hashtable();
	private int mNumberOfChildren;
	
	public EvolutionIndividual(Object genom) {
		mId = mIndividualIds++;
		
		if(isValidGenom(genom)) {
			mGenom = genom;
		}
		
		mAge = 0;
		mNumberOfChildren = 0;
	}


	/**
	 * Returns true if the object is a valid genom.
	 * This method implementation always returns true.
	 * Subclasses may override this method to make sure that 
	 * only a specific type of objects can be used in this Individual.
	 * 
	 * @param genom the object to check.
	 * @return true if the object is a valid genom structure.
	 */
	public boolean isValidGenom(Object genom) {
		return true;
	}
	
	
	/**
	 * Sets the genom object for this individual.
	 * @param genom
	 * @return
	 */
	public boolean setGenom(Object genom) {
		if(genom == null || !isValidGenom(genom)) {
			return false;
		}
		mGenom = genom;
		return true;
	}
	
	/**
	 * Returns the internal genom representation of the individual.
	 * @return the genom representation. (may be null).
	 */
	public Object getGenom() {
		return mGenom;
	}
	
	/**
	 * Returns the id of this individual.
	 * @return the id.
	 */
	public int getId() {
		return mId;
	}

	/**
	 * Adds a fitness value. All fitness values are append to a vector, so the order
	 * of adding is reflected in the order of appearance in the vector.
	 * @param value the fitness value to add.
	 */
	public void addFitnessValue(double value) {
		mFitnessValues.add(new Double(value));
	}
	
	/**
	 * Returns the vector with the FitnessValues.
	 * This method returns no copy (performance), so modifications of the vector affect this object!
	 * @return the vector with all FitnessFunctions.
	 */
	public Vector getFitnessValues() {
		return mFitnessValues;
	}
	
	/**
	 * Removes all fitness values.
	 */
	public void clearFitnessValues() {
		mFitnessValues.clear();
	}
	
	
	/**
	 * Sets the age of the EvolutionIndividual, indicating the number of generations
	 * this individual survived.
	 * 
	 * @param age the current age of the individual.
	 */
	public void setAge(int age) {
		mAge = age;
	}
	
	/**
	 * Returns the current age of the individual (number of generations survived).
	 * 
	 * @return the age.
	 */
	public int getAge() {
		return mAge;
	}
	
	
	/**
	 * Adds a parent individual. 
	 * 
	 * @param individual the parent to add.
	 * @return true if successful, false if null or already contained.
	 */
	public boolean addParent(EvolutionIndividual individual) {
		if(individual == null || mParents.contains(individual)) {
			return false;
		}
		mParents.add(individual);
		return true;
	}
	
	/**
	 * Removes a parent individual.
	 * 
	 * @param individual the parent to remove.
	 * @return true if successful, false if null or not contained.
	 */
	public boolean removeParent(EvolutionIndividual individual) {
		if(individual == null || !mParents.contains(individual)) {
			return false;
		}
		mParents.remove(individual);
		return true;
	}
	
	/**
	 * Removes all parents.
	 */
	public void removeParents() {
		mParents.clear();
	}
	
	/**
	 * Returns a the internal Vector with all currently registered parents.
	 * Warning: the returned Vector is NOT a copy, so modifications will affect the 
	 * object internals (this decision was made due to performance reasons).
	 * 
	 * @return
	 */
	public Vector getParents() {
		return mParents;
	}
	
	public void setNumberOfChildren(int children) {
		mNumberOfChildren = children;
	}
	
	public int getNumberOfChildren() {
		return mNumberOfChildren;
	}
	
	/**
	 * Sets the property identified by propertyName.
	 * @param propertyName
	 * @param property
	 * @return true if successful.
	 */
	public boolean setProperty(String propertyName, Object property) {
		mProperties.put(propertyName, property);
		return true;
	}
	
	/**
	 * Returns the property with the given name. If no such property is available, 
	 * then null is returned.
	 * @param propertyName
	 * @return
	 */
	public Object getProperty(String propertyName) {
		return mProperties.get(propertyName);
	}
	
	/**
	 * Adds all properties of a property HashTable to the internal 
	 * property table.
	 * @param properties
	 * @return
	 */
	public boolean addProperties(Hashtable properties) {
		if(properties == null) {
			return false;
		}
		Enumeration e = properties.keys();
		for(; e.hasMoreElements();) {
			String name = (String) e.nextElement();
			mProperties.put(name, properties.get(name));
		}
		return true;
	}
	
	public Hashtable getProperties() {
		return new Hashtable(mProperties);
	}
	
	
}
