package de.fhg.iais.outdoor.simba.evolution;


import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.Scheduler;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.core.SystemObject;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;


/**
 * The EvolutionManager manages and executes the evolution process. 
 * To customize the evolution process it is possible to add and remove
 * the EvolutionOperators, the EvaluationMethod and the SelectionMethods.
 * 
 * It is also possible to use custom genoms in the EvolutionIndividuals. 
 * If other genoms than NeuralNets are used, then the EvaluationMethod and
 * the EvolutionOperators also must be able to work with these genoms.
 * 
 * @author Christian Rempis
 */
public class EvolutionManager implements ValueChangedListener, SystemObject {
	
	
	private Vector mEvolutionOperators = new Vector();
	private Vector mIndividuals = new Vector();
	private int mCurrentGeneration;
	private Vector mEvolutionListeners = new Vector();
	private Vector mSelectionMethods = new Vector();
	private Vector mSelectionMethodOffspringPercentages = new Vector();
	private EvaluationMethod mEvaluationMethod;
	
	
	private IntValue mMeanPopulationSize;
	private IntValue mCurrentPopulationSize;
	private BooleanValue mRunEvolution;
	private BooleanValue mEvolutionIsRunning;
	private IntValue mPreserveParents;
	private BooleanValue mStaseMode;
	private boolean mStaseModeWasActive;
	private BooleanValue mEvolutionEnabled;
	private BooleanValue mEvaluateOnlyValue;
	
	private boolean mIsChangingPercentageValue;
	private boolean mRestartGeneration;
	
	private EvolutionTesterInterface mEvolutionTester;
	
	
	
	public EvolutionManager(ValueManager valueManager) {
		
		mCurrentGeneration = 0;
		
		mMeanPopulationSize = new IntValue(100);
		mMeanPopulationSize.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Control/PopulationSize", mMeanPopulationSize);
		
		mCurrentPopulationSize = new IntValue(0);
		mCurrentPopulationSize.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Status/CurrentPopulationSize", mCurrentPopulationSize);
		
		mRunEvolution = new BooleanValue(false);
		mRunEvolution.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Control/Run", mRunEvolution);
		
		mEvolutionIsRunning = new BooleanValue(false);
		mEvolutionIsRunning.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Status/Running", mEvolutionIsRunning);
		
		mPreserveParents = new IntValue(2);
		mPreserveParents.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Control/PreserveParents", mPreserveParents);
		
		mStaseMode = new BooleanValue(false);
		mStaseMode.addValueChangedListener(this);
		valueManager.addValue("/Evolution/Control/StaseMode", mStaseMode);
		
		mEvolutionEnabled = new BooleanValue(false);
		mEvolutionEnabled.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Status/EvolutionEnabled", mEvolutionEnabled);
		
		mEvaluateOnlyValue = new BooleanValue(false);
		mEvaluateOnlyValue.addValueReplacedListener(this);
		valueManager.addValue("/Evolution/Control/EvaluateOnly", mEvaluateOnlyValue);
		
		
		//This boolean is used to avoid infinite loops when the 
		//updateSelectionMethodOffspringPercentage() method is called.
		mIsChangingPercentageValue = false;
		
		setEvaluationMethod(new NeuralNetEvaluationMethod());
	}
	
	
	/**
	 * Starts and runs the evolution process. 
	 * If this method is called, then the manager waits till the RunEvolution Value 
	 * is set to true. As long as this Value is true, the evolution will be executed. 
	 *
	 */
	public void runEvolution() {
		
		mEvolutionEnabled.set(true);

		//he evolution can be disabled. 
		//Method runEvolution has to be called again to enable the evoulution again.
		while(mEvolutionEnabled.get()) { 
	
			while(!mRunEvolution.get()) {
				if(mEvolutionIsRunning.get() == true) {
					mEvolutionIsRunning.set(false);
				}
				if(!mEvolutionEnabled.get()) { //disable the evolution
					return;
				}
				//For test cases only.
				if(mEvolutionTester != null) {
					mEvolutionTester.runEvolutionPaused();
				}
				
				try {
					Thread.sleep(100);
				}
				catch(InterruptedException e) {
				}
			}
			int meanPopulationSize = mMeanPopulationSize.get();
			int preserveParents = mPreserveParents.get();
			
			Scheduler scheduler = Simulator.getInstance().getScheduler();
			
			synchronized(mIndividuals) {
			
				if(mEvolutionIsRunning.get() == false) {
					mEvolutionIsRunning.set(true);
				}
				
				if(mEvaluationMethod == null) {
					Logger.getLogger("Simulator")
						.warning("Could not find an evaluation method. Evolution stopped.");
					mRunEvolution.set(false);
					continue;
				}
				
				//If there is no initial networks (or more), then create a minimal network based on the agent interface.
				if(mIndividuals.isEmpty()) {
					NeuralNet net = NeuralNet.createMinimalNet(Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface());
					if(net != null) {
						EvolutionIndividual ind = new EvolutionIndividual(net);
						Vector generation = new Vector();
						generation.add(ind);
						replaceGeneration(generation);
					}
				}
				
				//for unit test purposes only.
				if(mEvolutionTester != null) {
					mEvolutionTester.preparationsCompleted();
				}
	
				//Skip generation of next individuals if in evaluateOnlyMode.
				boolean evaluateOnly = mEvaluateOnlyValue.get();
				if(!evaluateOnly) {
					//Select parents and create children stubs
					Vector children = new Vector();
					Vector parents = new Vector();
					for(Iterator select = mSelectionMethods.iterator(); select.hasNext();) {
						SelectionMethod selectionMethod = (SelectionMethod) select.next();
						
						//Create next generation: Use network selection as parents.
						//TODO extend for multiple fitness functions.
						Vector localParents = selectionMethod.selectParents(mIndividuals);
						
						//Reset number of children for each parent.
						for(Iterator i = localParents.iterator(); i.hasNext();) {
							((EvolutionIndividual) i.next()).setNumberOfChildren(0);
						}
						
						int numberOfChildren = (int) Math.max(0.0, (Math.round((selectionMethod.getOffspringPercentageValue().get() * meanPopulationSize / 100.0)) - localParents.size()));
						children.addAll(selectionMethod.createChildrenStubs(localParents, numberOfChildren));
		
						if(scheduler.isGenerationRunning()) {
							//Log selection method parameters and best networks.
							EvolutionLogger logger = Simulator.getInstance().getScheduler().getEvolutionLogger();
							logger.logXML("<selectionMethod name=\"" + selectionMethod.getName() + "\" fitnessIndex=\"" + selectionMethod.getFitnessFunctionIndexValue().get() 
											+ "\" individualCount=\"" + meanPopulationSize + "\">", "</selectionMethod>");
							selectionMethod.logGenoms(new Vector(localParents));
							logger.closeTag();
						}
						
						//add new parents to the parent container, hereby avoiding dulicates
						for(int k = 0; k < localParents.size(); ++k) {
							Object parent = localParents.elementAt(k);
							if(!parents.contains(parent)) {
								parents.add(parent);
							}
						}
						
					}
					notifyParentsSelected(parents);
		
					mIndividuals.clear();
					
					
					LinkedList offspringList = new LinkedList();
					offspringList.addAll(children);
					
					
					//Preserve the specified number of parents from beeing changed (-1 means all parents).
					for(int i = 0; i < parents.size(); ++i) {
						EvolutionIndividual parent = (EvolutionIndividual) parents.elementAt(i);
						if(preserveParents == -1 || i < preserveParents) {
							//put parent unchanged to the next generation
							parent.setAge(parent.getAge() + 1);
							mIndividuals.add(parent);
							parent.clearFitnessValues();
						}
						else {
							 //put the parents in front of the offspring list (that will be modified)
							parent.addParent(parent);
							offspringList.add(parent);
						}
					}
					
					
					//modify the offspring.
					synchronized(mEvolutionOperators) {
						
						sortOperators(); //by layer, so the layer can be changed during the evolution.
					
						while(!offspringList.isEmpty()) {
							EvolutionIndividual ind = (EvolutionIndividual) offspringList.poll();
							ind.setAge(0);
							boolean processCompleted = false;
							boolean operatorsApplicable = true;
							for(Iterator i = mEvolutionOperators.iterator(); i.hasNext();) {
								((EvolutionOperator) i.next()).prepareForNextIndividual();
							}
							
							while(!processCompleted && operatorsApplicable) {
								operatorsApplicable = false;
								processCompleted = true;
		
								for(Iterator i = mEvolutionOperators.iterator(); i.hasNext();) {
									EvolutionOperator op = (EvolutionOperator) i.next();
									
									if(op.canStillBeApplied()) {
										operatorsApplicable = true;
										if(!op.executeOperation(ind)) {
											processCompleted = false;
										}
									}
								}
							}
							if(processCompleted && operatorsApplicable) {
								mIndividuals.add(ind);
								ind.removeParents();
								ind.clearFitnessValues();
							}
							else {
								Simulator.getInstance().reportStatus("Could not find a suitable mutation for individual " + ind.getId());
							}
						}
					}
				}
				if(scheduler.isGenerationRunning()) {	
					//Close the current generation and start the next one.
					
					scheduler.notifyGenerationListeners(mCurrentGeneration, false);
					if(!evaluateOnly) {
						mCurrentGeneration++;
					}
				}
				
				scheduler.prepareNextGeneration(mCurrentGeneration, scheduler.getNumberOfTries(), scheduler.getNumberOfSteps());
				
				
				
				//Report the current generation size
				mCurrentPopulationSize.set(mIndividuals.size());
				
				notifyIndividualsCreated(mIndividuals);
				
			}
		
			mRestartGeneration = false;
			do {
				if(mRestartGeneration) {
					scheduler.getEvolutionLogger().logComment("Current generation was restarted manually.");
					scheduler.notifyGenerationListeners(mCurrentGeneration, false);
					scheduler.prepareNextGeneration(mCurrentGeneration, scheduler.getNumberOfTries(), scheduler.getNumberOfSteps());
				}
				Vector generation = new Vector(mIndividuals);
				
				mRestartGeneration = false;
				
				//Evaluate each individual
				mEvaluationMethod.prepareNextGeneration(mCurrentGeneration);
				int individualCounter = 0;
				for(; individualCounter < generation.size(); ++individualCounter) {
				//for(Iterator i = generation.iterator(); i.hasNext();) {
					//EvolutionIndividual ind = (EvolutionIndividual) i.next();
					EvolutionIndividual ind = (EvolutionIndividual) generation.elementAt(individualCounter);
					
					//evaluate and consider stase mode.
					boolean reevaluate = false;
					do {
						reevaluate = false;
						if(mStaseMode.get()) {
							reevaluate = true;
						}
						mStaseModeWasActive = false;
						
						mEvaluationMethod.evaluateIndividual(ind, individualCounter);
						if(mStaseMode.get()) {
							mStaseModeWasActive = true;
						}
						if(mStaseModeWasActive) {
							reevaluate = true;
						}

						if(mRestartGeneration) {
							//Start the evaluation from scratch, using the same individuals.
							break;
						}
					} while(reevaluate);
					//individualCounter++;
					if(mRestartGeneration) {
						break;
					}
				}
				
			} while(mRestartGeneration);
			
			notifyIndividualsEvaluated(mIndividuals);
			//statistics data calculation is done in generation completed. 

		}
		
		
		if(mEvolutionIsRunning.get() == true) {
			mEvolutionIsRunning.set(false);
		}
		mEvolutionEnabled.set(false);
		
	}
	
	
	/**
	 * Makes sure that all operators are sorted according to their layer,
	 * even if the layers are changed during the evolution.
	 * //TODO Solve this with ValueChangedListeners instead of a call per generation. 
	 */
	protected void sortOperators() {
		synchronized(mEvolutionOperators) {
			Vector operations = new Vector(); //TODO check memory and system performance.
			for(Iterator i = mEvolutionOperators.iterator(); i.hasNext();) {
				EvolutionOperator ops = (EvolutionOperator) i.next();
				int j = 0;
				for(; j < operations.size(); ++j) {
					if(ops.getLayerValue().get() <= ((EvolutionOperator) operations.elementAt(j)).getLayerValue().get()) {
						operations.add(j, ops);
						break;
					}
				}
				if(j == operations.size()) {
					operations.add(ops);
				}
			}
			mEvolutionOperators = operations;
		}
	}
	
	
	
	public boolean cleanUp() {
		return true;
	}

	/**
	 * Initializes the EvolutionManager.
	 * If no selection method was set, then the roulette wheel selection is default.
	 * If no evaluation method was set, then the neural net evaluation method is default.
	 */
	public boolean init() {
		return true;
	}

	/**
	 * Resets the EvolutionManager and all operators and methods.
	 */
	public synchronized void reset() {
		mCurrentGeneration = 0;
		synchronized(mIndividuals) {
			mIndividuals.clear();
		}
		for(Iterator i = mEvolutionOperators.iterator(); i.hasNext();) {
			((EvolutionOperator) i.next()).reset();
		}
	}
	
	
	
	/**
	 * Returns the Value object that can be used to set the mean population size.
	 * This value affects only entire generations, even if the Value is changed
	 * while a generation is evaluated.
	 * 
	 * @return the Value that represents the mean generation size.
	 */
	public IntValue getMeanPopulationSizeValue() {
		return mMeanPopulationSize;
	}
	
	/**
	 * Returns the Value object that can be used to interrupt and proceed 
	 * the evolution (only between generations).
	 * 
	 * @return the value object that indicates whether the evolution should run.
	 */
	public BooleanValue getRunEvolutionValue() {
		return mRunEvolution;
	}
	
	
	/**
	 * Returns the BooleanValue object that indicates whether the evolution
	 * is currently running.
	 * @return the Value indicating whether the evolution is running.
	 */
	public BooleanValue getEvolutionRunningValue() {
		return mEvolutionIsRunning;
	}
	
	/**
	 * Returns an IntValue object that determines the number of parents per
	 * SelectionMethod that are moved to the next generation without 
	 * modifications.
	 * 
	 * @return the IntValue with the number of parents to perserve.
	 */
	public IntValue getPreserveParentsValue() {
		return mPreserveParents;
	}
	
	/**
	 * Returns the Value object that contains the number of individuals in the 
	 * current generation. This value should not be changed.
	 * @return
	 */
	public IntValue getCurrentPopulationSizeValue() {
		return mCurrentPopulationSize;
	}
	
	
	/**
	 * Returns the value object that is used to control the stase mode.
	 * In stase mode the evolution stopps and repeatedly executes the current
	 * individual in an endless loop as long as stase is active. 
	 * If stase is released then the current individual is restarted and evaluated 
	 * normally. Changes done during stase remain and become part of the generation.
	 * @return
	 */
	public BooleanValue getStaseModeValue() {
		return mStaseMode;
	}
	
	/**
	 * Returns the Value object that indicates whether the evolution is
	 * enabled. This is the case if the runEvolution() method was called.
	 * 
	 * @return the BooleanValue indicating whether evolution is enabled.
	 */
	public BooleanValue getEvolutionEnabledValue() {
		return mEvolutionEnabled;
	}
	
	/**
	 * Returns the Value object that cna be used to control the evaluate only mode.
	 * If this value is true, then the generations are evaluated without 
	 * modifications. No new generations will be created.
	 * @return
	 */
	public BooleanValue getEvaluateOnlyModeValue() {
		return mEvaluateOnlyValue;
	}
	
	
	/**
	 * Replaces the current generation with a new set of individuals.
	 * This method can be used during the initialization of the GA to
	 * set the initial individuals of to load a generation during evolution
	 * from a file.
	 * The generation is replaced only if all individuals are instances 
	 * of the current master EvolutionIndividual.
	 * 
	 * @param individuals a vector with individuals.
	 * @return true is successful, otherwise false.
	 */
	public boolean replaceGeneration(Vector individuals) {
		//TODO buffered -> wait till current generation is done.
		if(individuals == null) {
			return false;
		}
		synchronized(mIndividuals) {
			mIndividuals.clear();
			mIndividuals.addAll(individuals);
		}
		return true;
	}
	
	
	/**
	 * Returns a Vector with the current individuals. 
	 */
	public Vector getGeneration() {
		synchronized(mIndividuals) {
			return new Vector(mIndividuals);
		}
	}
	
	

	/**
	 * Sets the current EvaluationMethod. This object avaluates the genoms
	 * of the EvolutionIndividuals. Therefore a suitable method should be set
	 * for each different kind of genom.
	 * 
	 * @param method the EvaluationMethod to use.
	 */
	public void setEvaluationMethod(EvaluationMethod method) {
		mEvaluationMethod = method;
	}
	
	
	/**
	 * Returns the currently used EvaluationMethod.
	 * 
	 * @return the EvaluationMethod.
	 */
	public EvaluationMethod getEvaluationMethod() {
		return mEvaluationMethod;
	}
	
	
	/**
	 * Adds an evolution operator to the EvolutionManager.
	 * The order in which the operators are added to the manager is similar
	 * to the order in which the operators are called during evolution.
	 * @param operator the operator to add.
	 * @return true if successful, if already contained false is returned.
	 */
	public synchronized boolean addEvolutionOperator(EvolutionOperator operator, int layer) {
		if(operator == null || mEvolutionOperators.contains(operator)) {
			return false;
		}
		operator.getLayerValue().set(layer);
		mEvolutionOperators.add(operator);
		return true;
	}
	
	
	/**
	 * Removes an evolution operator.
	 * @param operator the operator to remove.
	 * @return true if successful, false if not contained.
	 */
	public synchronized boolean removeEvolutionOperator(EvolutionOperator operator) {
		if(operator == null || !mEvolutionOperators.contains(operator)) {
			return false;
		}
		mEvolutionOperators.remove(operator);
		return true;
	}
	
	
	/**
	 * Returns a Vector with all EvolutionOperators that are currently available.
	 * 
	 * @return a Vector containing all EvolutionOperators.
	 */
	public synchronized Vector getEvolutionOperators() {
		return new Vector(mEvolutionOperators);
	}
	
	
	/**
	 * Makes sure that the offspring percentages of all selection methods sum up 
	 * to exactly 100 percent. If they do not, then the percentages are changed 
	 * to fit the 100 percent. 
	 * 
	 * If the protectedValue object is not null, then the selection method that owns 
	 * this value (getOffspringPercentageValue) is not changed during the percentage
	 * adaptation. Only if not possible otherwise, this value still can be changed.
	 * 
	 * @param protectedValue the offspringPercentageValue that is preferrably not changed.
	 */
	private void updateSelectionMethodOffspringPercentages(NormalizedDoubleValue protectedValue) {
		if(mIsChangingPercentageValue) {
			return;
		}
		mIsChangingPercentageValue = true;
		
		double percentage = 100.0;
		if(protectedValue != null) {
			percentage -= protectedValue.get();
		}
		for(int i = 0; i < mSelectionMethods.size(); ++i) {
			SelectionMethod selectionMethod = (SelectionMethod) mSelectionMethods.elementAt(i);
			if(protectedValue != null && protectedValue == selectionMethod.getOffspringPercentageValue()) 
			{
				continue;
			}
			double smp = selectionMethod.getOffspringPercentageValue().get();
			if(percentage > smp) {
				percentage -= smp;
			}
			else {
				if(selectionMethod.getOffspringPercentageValue().get() != percentage) {
					selectionMethod.getOffspringPercentageValue().set(percentage);
				}
				percentage = 0.0;
			}
		}
		if(percentage > 0.0 && !mSelectionMethods.isEmpty()) {
			for(int i = mSelectionMethods.size() - 1; i >= 0; --i) {
				SelectionMethod sm = (SelectionMethod) mSelectionMethods.elementAt(i);
				if(sm.getOffspringPercentageValue() != protectedValue || mSelectionMethods.size() == 1) {
					percentage = sm.getOffspringPercentageValue().get() + percentage;
					sm.getOffspringPercentageValue().set(percentage);
					break;
				}
			}
			
		}
		mIsChangingPercentageValue = false;
	}
	
	/**
	 * Adds an evolution operator to the EvolutionManager.
	 * The order in which the operators are added to the manager is similar
	 * to the order in which the operators are called during evolution.
	 * @param operator the operator to add.
	 * @return true if successful, if already contained false is returned.
	 */
	public synchronized boolean addSelectionMethod(SelectionMethod operator) {
		if(operator == null || mSelectionMethods.contains(operator)) {
			return false;
		}
		mSelectionMethods.add(operator);
		Value v = operator.getOffspringPercentageValue();
		v.addValueChangedListener(this);
		mSelectionMethodOffspringPercentages.add(v);
		updateSelectionMethodOffspringPercentages(operator.getOffspringPercentageValue());
		return true;
	}
	
	/**
	 * Removes an evolution operator.
	 * @param operator the operator to remove.
	 * @return true if successful, false if not contained.
	 */
	public synchronized boolean removeSelectionMethod(SelectionMethod operator) {
		if(operator == null || !mSelectionMethods.contains(operator)) {
			return false;
		}
		mSelectionMethods.remove(operator);
		Value v = operator.getOffspringPercentageValue();
		mSelectionMethodOffspringPercentages.remove(v);
		v.removeValueChangedListener(this);
		updateSelectionMethodOffspringPercentages(null);
		return true;
	}
	
	/**
	 * Returns a Vector that contains all selection methods that are currently present.
	 * @return a Vector containing the selection methods.
	 */
	public synchronized Vector getSelectionMethods() {
		return new Vector(mSelectionMethods);
	}
	
	/**
	 * Removes all selection methods.
	 */
	public synchronized void removeAllSelectionMethods() {
		Vector selectionMethods = new Vector(mSelectionMethods);
		for(Iterator i = selectionMethods.iterator(); i.hasNext();) {
			removeSelectionMethod((SelectionMethod) i.next());
		}
	}
	
	/**
	 * Sets the number of the current generation. This has no effect on the
	 * individuals that are evolved.
	 * 
	 * @param generation the current generation number.
	 */
	public void setCurrentGeneration(int generation) {
		mCurrentGeneration = generation;
	}

	/**
	 * Called if one of the observed value objects changed.
	 */
	public void valueChanged(Value value) {
		if(value == null) {
			return;
		}
		if(value == mStaseMode) { 
			if(mStaseMode.get()) {
				mStaseModeWasActive = true;
			}
		}
		else if(mSelectionMethodOffspringPercentages.contains(value)) {
			updateSelectionMethodOffspringPercentages((NormalizedDoubleValue) value);
		}
		
	}

	/**
	 * Called if one of the observed value objects are replaced by another
	 * Value in the ValueManager.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		mMeanPopulationSize = (IntValue) ValueManager.synchronizeValue(mMeanPopulationSize, oldValue, newValue, this);
		mRunEvolution = (BooleanValue) ValueManager.synchronizeValue(mRunEvolution, oldValue, newValue, this);
		mEvolutionIsRunning = (BooleanValue) ValueManager.synchronizeValue(mEvolutionIsRunning, oldValue, newValue, this);
		mPreserveParents = (IntValue) ValueManager.synchronizeValue(mPreserveParents, oldValue, newValue, this);
		mCurrentPopulationSize = (IntValue) ValueManager.synchronizeValue(mCurrentPopulationSize, oldValue, newValue, this);
		mStaseMode = (BooleanValue) ValueManager.synchronizeValue(mStaseMode, oldValue, newValue, this);
		mEvolutionEnabled = (BooleanValue) ValueManager.synchronizeValue(mEvolutionEnabled, oldValue, newValue, this);
		mEvaluateOnlyValue = (BooleanValue) ValueManager.synchronizeValue(mEvaluateOnlyValue, oldValue, newValue, this);
		
	}
	
	/**
	 * Sets a mark that the current generation should be restarted.
	 * If an evolution is running, then after the next evaluated individual the entire
	 * generation is started from the first individual on. The generation will not
	 * modified and the so far determined fitness values are discarded.
	 */
	public void restartCurrentGeneration() {
		mRestartGeneration = true;
	}
	
	
	/**
	 * Adds an EvolutionListener.
	 * @param listener the listener to add.
	 * @return true if successful.
	 */
	public synchronized boolean addEvolutionListener(EvolutionListener listener) {
		if(listener == null || mEvolutionListeners.contains(listener)) {
			return false;
		}
		mEvolutionListeners.add(listener);
		return true;
	}
	
	/**
	 * Removes an EvolutionListener.
	 * @param listener
	 * @return true if successfully.
	 */
	public synchronized boolean removeEvolutionListener(EvolutionListener listener) {
		if(listener == null || !mEvolutionListeners.contains(listener)) {
			return false;
		}
		mEvolutionListeners.remove(listener);
		return true;
	}
	
	/**
	 * Returns a Vector with all registered EvolutionListeners.
	 * @return
	 */
	public synchronized Vector getEvolutionListeners() {
		return new Vector(mEvolutionListeners);
	}

	
	public synchronized void notifyParentsSelected(Vector parents) {
		for(Iterator i = mEvolutionListeners.iterator(); i.hasNext();) {
			((EvolutionListener) i.next()).notifyParentsSelected(parents, mCurrentGeneration);
		}
	}
	
	public synchronized void notifyIndividualsCreated(Vector individuals) {
		for(Iterator i = mEvolutionListeners.iterator(); i.hasNext();) {
			((EvolutionListener) i.next()).notifyIndividualsCreated(individuals, mCurrentGeneration);
		}
	}
	
	public synchronized void notifyIndividualsEvaluated(Vector individuals) {
		for(Iterator i = mEvolutionListeners.iterator(); i.hasNext();) {
			((EvolutionListener) i.next()).notifyIndividualsEvaluated(individuals, mCurrentGeneration);
		}
	}
	
	//Only for test purposes. Do not use for other stuff.
	public void setEvolutionTester(EvolutionTesterInterface tester) {
		mEvolutionTester = tester;
	}
	
	

}
