package de.fhg.iais.outdoor.simba.evolution;


import java.util.Random;

import de.fhg.iais.outdoor.simba.core.ParameterElement;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.ObjectValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.value.ValueReplacedListener;

public abstract class EvolutionOperator extends ParameterElement implements ValueReplacedListener {
	
	private IntValue mLayer;
	private IntValue mMaxNumberOfApplication;
	protected int mApplicationCounter;
	protected Random mRandom;
	
	public EvolutionOperator(String name, int maxNumberOfApplicationPerIndividual) {
		super(name, "/Evolution/Operators/" + name + "/Config/");
		
		mName = name;
		
		mRandom = new Random();
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		//try to add this operator to the ValueManager. If neccessary, change the name.
		ObjectValue objectValue = new ObjectValue(this);
		int counter = 1;
		while(!vm.addValue("/Evolution/Operators/" + mName, objectValue)) {
			mName = name + counter;
			counter++;
		}
		name = mName;
		
		mLayer = new IntValue(0);
		mMaxNumberOfApplication = new IntValue(maxNumberOfApplicationPerIndividual);
		
		
		addParameter(mLayer, "Layer");
		addParameter(mMaxNumberOfApplication, "MaxNumberOfApplications");
		
		mLayer.addValueReplacedListener(this);
		mMaxNumberOfApplication.addValueReplacedListener(this);
	}
	
	/**
	 * Returns true, if the network is valid according to this operator.
	 * A network is passed through all operators as long as at least one
	 * operator returns false of if a maxmial number of tries has been executed.
	 * @param net
	 * @return
	 */
	public abstract boolean executeOperation(EvolutionIndividual individual);
	
	public IntValue getLayerValue() {
		return mLayer;
	}
	
	
	
	public IntValue getMaxNumberOfApplicationsValue() {
		return mMaxNumberOfApplication;
	}
	
	public void prepareForNextIndividual() {
		mApplicationCounter = 0;
	}
	
	public void reset() {
		
	}
	
	/**
	 * Returns true if this operator can still be applied on the same individual.
	 * Operators can restrict the number of applications by setting the maximal
	 * number of applications. 
	 * Restricting the number of applications can be used to force an operation to 
	 * be used only once for an individual, even if this individual is not valid
	 * according to other operators. This is especially useful for very destructive operators.
	 * 
	 * If no operator can be applied any more and the genom still is not valid, then
	 * the individual is removed from the population. So restricting the number of
	 * applications can also reduce the computational overhead used during the generation 
	 * creation.
	 * @return
	 */
	public boolean canStillBeApplied() {
		mApplicationCounter++;
		if(!mActive.get()) {
			return false;
		}
		if(mApplicationCounter <= mMaxNumberOfApplication.get()) {
			return true;
		}
		return false;
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mLayer = (IntValue) ValueManager.synchronizeValue(mLayer, oldValue, newValue, this);
		mMaxNumberOfApplication = (IntValue) ValueManager.synchronizeValue(mMaxNumberOfApplication, oldValue, newValue, this);
		mActive = (BooleanValue) ValueManager.synchronizeValue(mActive, oldValue, newValue, this);
	}
	
	/**
	 * Sets the random number generator for this operator. This can be used to 
	 * set a deterministic subclass for test purposes or to set a specific seed.
	 * @param random
	 */
	public void setRandom(Random random) {
		mRandom = random;
	}
	
	public Random getRandom() {
		return mRandom;
	}



	public String createParameterLog() {
		
		StringBuffer entry = new StringBuffer();
		String parameterLog = super.createParameterLog();
		
		entry.append("<operator name=\"").append(getName());		
		if(parameterLog.equals("")) {
			entry.append("\"/>\n");
		}
		else {
			entry.append("\">\n");
			entry.append(parameterLog);
			entry.append("</operator>");
		}
		
		return entry.toString();
	}

	
	

}
