package de.fhg.iais.outdoor.simba.fitnessfunctions;

import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.fitness.FitnessFunction;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.SimObjectValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class PathMemorizerFitness2 extends FitnessFunction {
	
	private SimObject mRightBound;
	private SimObject mLeftBound;
	private SimObject mRobot;
	private BooleanValue mMovingLeft;
	private double mLastXPosition;
	private double mAccumulator;
	private DoubleValue mLeftThreshold;
	private DoubleValue mRightThreshold;
	private DoubleValue mGainToPower;
	private DoubleValue mGainPerDist;
	private BooleanValue mUseLeftGainForRightMovement;
	private BooleanValue mUseTurnFactor;
	private DoubleValue mLastTurningPoint;
	private DoubleValue mRatio;
	private DoubleValue mTurnFactor;
	
	
	public PathMemorizerFitness2() {
		super("PathMemorizerFitness2");
		
		mLeftThreshold = new DoubleValue(30.0);
		mRightThreshold = new DoubleValue(30.0);
		mGainToPower = new DoubleValue(1.0);
		mUseLeftGainForRightMovement = new BooleanValue(false);
		mUseTurnFactor = new BooleanValue(false);
		
		addParameter(mLeftThreshold, "LeftThreshold");
		addParameter(mRightThreshold, "RightThreshold");
		addParameter(mGainToPower, "GainToPower");
		addParameter(mUseLeftGainForRightMovement, "UseLeftGainForRightMovement");
		addParameter(mUseTurnFactor, "UseTurnFactor");
		
		mGainPerDist = new DoubleValue(1.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/Gain", mGainPerDist);
		
		mMovingLeft = new BooleanValue(true);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/MovingLeft", mMovingLeft);
		
		mLastTurningPoint = new DoubleValue(0.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/LastTurningPoint", mLastTurningPoint);
		
		mRatio = new DoubleValue(1.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/Ratio", mRatio);
		
		mTurnFactor = new DoubleValue(1.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/TurnFactor", mTurnFactor);
		
		
	}
	
	

	public boolean init() {
		boolean ok =  super.init();
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		
		SimObjectValue v = vm.getSimObjectValue("/SimObjects/StartingArea");
		if(v != null) {
			mRightBound = v.get();
		}
		v = vm.getSimObjectValue("/SimObjects/TargetArea");
		if(v != null) {
			mLeftBound = v.get();
		}
		v = vm.getSimObjectValue("/SimObjects/Robot");
		if(v != null) {
			mRobot = v.get();
		}
		if(mRightBound == null || mLeftBound == null || mRobot == null) {
			Logger.getLogger("Simulator").warning("Could not find the required SimObjects");
			return false;
		}
		
		return ok;
	}



	public double calculateFitness() {

		if(mLastXPosition == Double.MAX_VALUE) {
			mLastXPosition = mRobot.getGlobalPosition().x;
		}
		
		double diff = 0;
		if(mMovingLeft.get()) {
			diff = mLastXPosition - mRobot.getGlobalPosition().x;
			//check if the robot is still moving left
			if(diff < 0.0 && 
				mRobot.getGlobalPosition().x <= mLeftBound.getGlobalPosition().x + mLeftThreshold.get()) 
			{
				mMovingLeft.set(false);
				if(!mUseLeftGainForRightMovement.get()) {
					mGainPerDist.set(0.5);
				}
			}
		}
		else {
			diff = mRobot.getGlobalPosition().x - mLastXPosition;
			if(diff < 0.0
					&& mRobot.getGlobalPosition().x >= mRightBound.getGlobalPosition().x - mRightThreshold.get()
					&& mRobot.getGlobalPosition().x <= mRightBound.getGlobalPosition().x + mRightThreshold.get()) 
			{
				double distToRightArea = Math.abs(mRobot.getGlobalPosition().x - mRightBound.getGlobalPosition().x);
				double threshold = mRightThreshold.get();
				if(threshold == 0) {
					threshold = 1;
				}
				mGainPerDist.set((threshold - distToRightArea) / threshold);
				mMovingLeft.set(true);
			}
//			else if(mRobot.getGlobalPosition().x > mRightBound.getGlobalPosition().x + mRightThreshold.get()) {
//				mMovingLeft.set(false);
//				mGainPerDist.set(0.0);
//			}
		}
		if(diff != 0.0) {
			double d = Math.signum(diff) * Math.pow((Math.abs(diff) * mGainPerDist.get()), mGainToPower.get());
			mAccumulator += d;
		}
		
		mLastXPosition = mRobot.getGlobalPosition().x;
		
		return Math.max(0.0, mAccumulator);
	}

	public void reset() {
		mMovingLeft.set(true);
		mLastXPosition = Double.MAX_VALUE;
		mAccumulator = 0.0;
		mGainPerDist.set(0.5);
		mLastTurningPoint.set(mRobot.getPosition().x);
		mTurnFactor.set(1.0);
		mRatio.set(1.0);
	}

}
