package de.fhg.iais.outdoor.simba.neuralnet;

import java.util.Iterator;
import java.util.Vector;

public class NeuralNet implements Cloneable {
	
	private Vector mNeurons = new Vector();
	private TransferFunction mDefaultTransferFunction;
	
	public NeuralNet() {
		mDefaultTransferFunction = new TransferFunctionTanh();
	}
	
	public static NeuralNet createMinimalNet(NeuralNetAgentInterface netInterface) {
		if(netInterface == null) {
			return null;
		}
		NeuralNet net = new NeuralNet();
		Vector inputNeurons = netInterface.getInputNeurons();
		for(Iterator i = inputNeurons.iterator(); i.hasNext();) {
			NeuronWrapper nw = (NeuronWrapper) i.next();
			Neuron n = new Neuron(nw.getName(), null, Neuron.INPUT);
			n.setActivationValue(nw);
			net.addNeuron(n);
		}
		Vector outputNeurons = netInterface.getOutputNeurons();
		for(Iterator i = outputNeurons.iterator(); i.hasNext();) {
			NeuronWrapper nw = (NeuronWrapper) i.next();
			Neuron n = new Neuron(nw.getName(), null, Neuron.OUTPUT);
			n.setActivationValue(nw);
			net.addNeuron(n);
		}
		return net;
	}
	
	
	
	public synchronized void updateNetwork() {
		for(Iterator i = mNeurons.iterator(); i.hasNext();) {
			((Neuron) i.next()).storeLastActivity();
		}
		for(Iterator i = mNeurons.iterator(); i.hasNext();) {
			((Neuron) i.next()).updateActivation();
		}
	}
	
	public synchronized void reset() {
		for(Iterator i = mNeurons.iterator(); i.hasNext();) {
			((Neuron) i.next()).reset();
		}
	}
	
	public synchronized boolean addNeuron(Neuron neuron) {
		if(neuron == null || mNeurons.contains(neuron)) {
			return false;
		}
		mNeurons.add(neuron);
		if(neuron.getTransferFunction() == null) {
			neuron.setTransferFunction(mDefaultTransferFunction);
		}
		return true;
	}
	
	public synchronized boolean removeNeuron(Neuron neuron) {
		if(neuron == null || !mNeurons.contains(neuron)) {
			return false;
		}
		mNeurons.remove(neuron);
		return true;
	}
	
	public synchronized Vector getNeurons() {
		return new Vector(mNeurons);
	}
	
	public void setDefaultTransferFunction(TransferFunction transferFunction) {
		if(transferFunction != null) {
			mDefaultTransferFunction = transferFunction;
		}
	}
	
	public TransferFunction getDefaultTransferFunction() {
		return mDefaultTransferFunction;
	}
	
	public synchronized Vector getSynapses() {
		Vector synapses = new Vector();
		for(Iterator i = mNeurons.iterator(); i.hasNext();) {
			synapses.addAll(((Neuron) i.next()).getSynapses());
		}
		return synapses;
	}
	
	public synchronized void validateNetworkConnections() {
		for(Iterator i = mNeurons.iterator(); i.hasNext();) {
			Neuron n = (Neuron) i.next();
			Vector synapses = n.getSynapses();
			for(Iterator j = synapses.iterator(); j.hasNext();) {
				Synapse s = (Synapse) j.next();
				boolean valid = false;
				for(Iterator k = mNeurons.iterator(); k.hasNext();) {
					if(s.getSourceNeuron() == k.next()) {
						valid = true;
						break;
					}
				}
				if(!valid) {
					n.removeSynapse(s);
				}
			}
		}
	}

	public Object clone() throws CloneNotSupportedException {
		synchronized(this) {
			NeuralNet obj = (NeuralNet) super.clone();
			obj.mNeurons = new Vector();
			for(Iterator i = mNeurons.iterator(); i.hasNext();) {
				obj.mNeurons.add(((Neuron) i.next()).clone()); 
			}
			
			//set targetNeurons top valid neurons of the own network.
			for(Iterator i = obj.mNeurons.iterator(); i.hasNext();) {
				Neuron n = (Neuron) i.next();
				Vector synapses = n.getSynapses();
				for(Iterator j = synapses.iterator(); j.hasNext();) {
					Synapse s = (Synapse) j.next();
					Neuron source = s.getSourceNeuron();
					if(source != null) {
						int neuronId = source.getId();
						for(Iterator k = obj.mNeurons.iterator(); k.hasNext();) {
							Neuron candidate = (Neuron) k.next();
							if(candidate.getId() == neuronId) {
								s.setSourceNeuron(candidate);
								break;
							}
						}
					}
				}
			}
			
			
			
			obj.mDefaultTransferFunction = mDefaultTransferFunction;
			obj.validateNetworkConnections();
			return obj;
		}
	}
	
	
	public static Vector cloneNeuronGroup(Vector neurons) {
		Vector newNeurons = new Vector();
		
		for(Iterator i = neurons.iterator(); i.hasNext();) {
			try {
				newNeurons.add(((Neuron) i.next()).clone());
			} catch (CloneNotSupportedException e) {
			}
		}
		
		//set targetNeurons top valid neurons of the own network.
		for(Iterator i = newNeurons.iterator(); i.hasNext();) {
			Neuron n = (Neuron) i.next();
			n.setActivationValue(new NeuronWrapper("Activity", 0.0, 0.0, 1.0));
			Vector synapses = new Vector(n.getSynapses());
			for(Iterator j = synapses.iterator(); j.hasNext();) {
				Synapse s = (Synapse) j.next();
				Neuron source = s.getSourceNeuron();
				if(source != null) {
					int neuronId = source.getId();
					boolean foundNeuron = false;
					for(Iterator k = newNeurons.iterator(); k.hasNext();) {
						Neuron candidate = (Neuron) k.next();
						if(candidate.getId() == neuronId) {
							s.setSourceNeuron(candidate);
							foundNeuron = true;
							break;
						}
					}
					if(!foundNeuron) {
						n.removeSynapse(s);
					}
				}
			}
		}
		for(Iterator i = newNeurons.iterator(); i.hasNext();) {
			Neuron n = (Neuron) i.next();
			n.makeIdUnique();
		}
		
		return newNeurons;
	}
	
	public void clearNetwork() {
		mNeurons.clear();
		validateNetworkConnections();
	}

}
