package de.fhg.iais.outdoor.simba.neuralnet;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.StringValue;

public class Neuron implements Cloneable {

	public static final String INPUT = "input";
	public static final String OUTPUT = "output";
	public static final String HIDDEN = "hidden";
	
	public static final String NORMAL = "dynamic";
	public static final String PROTECTED = "protected"; //no incomming synapses.
	public static final String LOCKED = "locked"; //no incoming and outgoing synapses.
	
	public static final Vector mProtectionModes = new Vector();
	
	public static final int mHistorySize = 150;
	
	protected static int mNeuronIds;
	
	private NeuronWrapper mActivation;
	private DoubleValue mBias;
	private double mLastActivity;
	private StringValue mName;
	private TransferFunction mTransferFunction;
	private Vector mSynapses = new Vector();
	private LinkedList mActivityHistory = new LinkedList();
	private String mType;
	private String mProtection;
	private int mId;
	private IntValue mMaxIncomingSynapses;
	private IntValue mMaxOutgoingSynapses;
	private boolean mIsInputNeuron;
	
	/**
	 * Constructor.
	 * @param name
	 */
	public Neuron(String name) {
		this(name, null);
	}
	
	/**
	 * Constructor.
	 * If the transferFunction is null, then it is filled with the default transferFunction
	 * in the NeuralNet class this Neuron is used in.
	 * 
	 * @param name the name of this neuron.
	 * @param transferFunction the transferFunction. 
	 */
	public Neuron(String name, TransferFunction transferFunction) {
		this(name, transferFunction, Neuron.HIDDEN);
		
	}
	
	/**
	 * Constructor.
	 * If the transferFunction is null, then it is filled with the default transferFunction
	 * in the NeuralNet class this Neuron is used in.
	 * 
	 * @param name the name of this neuron.
	 * @param transferFunction the transferFunction. 
	 */
	public Neuron(String name, TransferFunction transferFunction, String type) {
		
		if(mProtectionModes.isEmpty()) {
			//Initially fill static protection mode vector with values.
			mProtectionModes.add(Neuron.NORMAL);
			mProtectionModes.add(Neuron.PROTECTED);
			mProtectionModes.add(Neuron.LOCKED);
		}
		
		mId = mNeuronIds++;
		mType = type;
		mName = new StringValue(name);
		mLastActivity = 0.0;
		mActivation = new NeuronWrapper(name, 0.0, -1.0, 1.0);
		mBias = new DoubleValue(0.0);
		mActivation.setNormalizedValueRange(-1.0, 1.0);
		mTransferFunction = transferFunction;
		mMaxIncomingSynapses = new IntValue(-1);
		mMaxOutgoingSynapses = new IntValue(-1);
		mProtection = Neuron.NORMAL;
		
		if(mType.equals(Neuron.INPUT)) {
			mIsInputNeuron = true;
		}
		else {
			mIsInputNeuron = false;
		}
	}

	
	/**
	 * Resets this Neuron to the state it had after construction.
	 * The Activity and the ActivityHistory are cleared.
	 */
	public void reset() {
		if(mTransferFunction != null && (mTransferFunction.getLowerBound() > 0.0 || mTransferFunction.getUpperBound() < 0.0)) {
			mActivation.setNormalized((mTransferFunction.getLowerBound() + mTransferFunction.getUpperBound()) / 2.0);
		}
		else {
			mActivation.setNormalized(0.0);
		}
		mLastActivity = 0.0;
		synchronized (mActivityHistory) {
			mActivityHistory.clear();
		}
		
	}
	
	
	/**
	 * Stores the curent Activity in a history so that it can be recovered later
	 * and to allow the activation update to work on a previously fixed activity.
	 */
	public void storeLastActivity() {
		mLastActivity = mActivation.getNormalized();
		synchronized(mActivityHistory) {
			mActivityHistory.addLast(new Double(mLastActivity));
			if(mActivityHistory.size() > mHistorySize) {
				mActivityHistory.removeFirst();
			}
		}
	}
	
	/**
	 * Returns the last known activity of this neuron.
	 * @return
	 */
	public double getLastActivity() {
		return mLastActivity;
	}
	
	
	/**
	 * Calculates the current activity of this Neuron.
	 * The new Activation is the sum of all incomming activations  from all
	 * connected Synapse objects and the bias, normalized by the TransferFunction.
	 * 
	 * Throws an IllegalStateException if this method is called without a proper
	 * TransferFunction.
	 */
	public void updateActivation() {
		//update only Hidden and Output neurons. Input neurons are set externally.
		if(!mIsInputNeuron) {
			if(mTransferFunction == null) {
				throw new IllegalStateException("No TransferFunction object found.");
			}
			double currentActivation = mBias.get();
			for(Iterator i = mSynapses.iterator(); i.hasNext();) {
				currentActivation += ((Synapse) i.next()).getActivation();
			}
			mActivation.setNormalized(mTransferFunction.transferActivation(currentActivation));
		}
	}
	
	/**
	 * Sets the TransferFunction.
	 * @param transferFunction
	 */
	public void setTransferFunction(TransferFunction transferFunction) {
		if(transferFunction != null) {
			mTransferFunction = transferFunction;
			mActivation.setNormalizedValueRange(mTransferFunction.getLowerBound(), mTransferFunction.getUpperBound());
		}
	}
	
	/**
	 * Returns the TransferFunction used by this Neuron.
	 * @return the Transferfunction.
	 */
	public TransferFunction getTransferFunction() {
		return mTransferFunction;
	}
	
	/**
	 * Sets the name. The name is used to easy the identification of the neuron for humans.
	 * The name does not need to be unique.
	 * @param name
	 */
	public void setName(String name) {
		mName.set(name);
	}
	
	/**
	 * Returns the name.
	 * @return
	 */
	public String getName() {
		return mName.get();
	}
	
	/**
	 * Returns the StringValue object used to encapulate the name of this neuron.
	 * @return
	 */
	public StringValue getNameValue() {
		return mName;
	}
	
	/**
	 * Returns the current activity of the Neuron.
	 * @return
	 */
	public double getActivation() {
		return mActivation.getNormalized();
	}
	
	/**
	 * Returns the NeuronWrapper Value object that holds the activity of the Neuron.
	 * @return
	 */
	public NeuronWrapper getActivationValue() {
		return mActivation;
	}
	
	/**
	 * Sets the bias of the neuron.
	 * @param bias
	 */
	public void setBias(double bias) {
		mBias.set(bias);
	}
	
	/**
	 * Returns the current bias.
	 * @return
	 */
	public double getBias() {
		return mBias.get();
	}
	
	/**
	 * Returns the DoubleValue object that encapsulates the current bias.
	 * @return
	 */
	public DoubleValue getBiasValue() {
		return mBias;
	}
	
	public IntValue getMaxIncomingSynapsesValue() {
		return mMaxIncomingSynapses;
	}
	
	public IntValue getMaxOutgoingSynapsesValue() {
		return mMaxOutgoingSynapses;
	}
	
	public void setType(String type) {
		mType = type;
		if(mType.equals(Neuron.INPUT)) {
			mIsInputNeuron = true;
		}
		else {
			mIsInputNeuron = false;
		}
	}
	
	public String getType() {
		return mType;
	}
	
	/**
	 * Sets the current mode of the neuron (one of NORMAL, PROTECTED, LOCKED).
	 * @param mode the current mode.
	 */
	public void setProtectionMode(String mode) {
		mProtection = mode;
	}
	
	/**
	 * Returns the mode.
	 * @return the mode.
	 */
	public String getProtectionMode() {
		return mProtection;
	}
	
	/**
	 * Sets the NeuronWrapper Value object that will hold the activity of the Neuron.
	 * This Value object can be borrowed from a sensor or an actuator to share a
	 * Value between a Neuron and the sensor / actuator.
	 * 
	 * @param activationValue
	 * @return
	 */
	public boolean setActivationValue(NeuronWrapper activationValue) {
		if(activationValue != null) {
			mActivation = activationValue;
			if(mTransferFunction != null) {
				mActivation.setNormalizedValueRange(mTransferFunction.getLowerBound(), mTransferFunction.getUpperBound());
			}
			return true;
		}
		return false;
	}
	
	/**
	 * Adds a Synapse object. Synapse objects added to a neuron direct to this neuron, which
	 * means that they increase or decrease the activitiy of this neuron during updates.
	 * @param synapse
	 * @return
	 */
	public boolean addSynapse(Synapse synapse) {
		if(synapse == null || mSynapses.contains(synapse) || synapse.getSourceNeuron() == null) {
			return false;
		}
		//make sure that there are only synapses comming from different neurons.
		for(Iterator i = mSynapses.iterator(); i.hasNext();) {
			if(((Synapse) i.next()).getSourceNeuron() == synapse.getSourceNeuron()) {
				return false;
			}
		}
		mSynapses.add(synapse);
		synapse.setTargetNeuron(this);
		return true;
	}
	
	/**
	 * Removes a Synapse object from the neuron.
	 */
	public boolean removeSynapse(Synapse synapse) {
		if(synapse == null || !mSynapses.contains(synapse)) {
			return false;
		}
		mSynapses.remove(synapse);
		if(synapse.getTargetNeuron() == this) {
			synapse.setTargetNeuron(null);
		}
		return true;
	}
	
	/**
	 * Returns a Vector with all registered Synapse objects.
	 * The returned Vector is a copy of the internal one.
	 * @return
	 */
	public Vector getSynapses() {
		return new Vector(mSynapses);
	}
	
	/**
	 * Returns the LinkedList with the activity history of the neuron.
	 * This LinkedList is the same object that is used internally, so it should
	 * not be modified.
	 * The reason to return the internal variable were performance issues.
	 * 
	 * @return
	 */
	public LinkedList getHistory() {
		return mActivityHistory;
	}
	
	public int getId() {
		return mId;
	}
	
	public void makeIdUnique() {
		mId = mNeuronIds++;
	}

	public Object clone() throws CloneNotSupportedException {
		Neuron obj = (Neuron) super.clone();
		
		obj.mActivation = mActivation; //uses the same NeuronWrapper object!
		obj.mBias = new DoubleValue(mBias.get());
		obj.mLastActivity = mLastActivity;
		obj.mName = new StringValue(mName.get());
		obj.mTransferFunction = mTransferFunction;
		obj.mSynapses = new Vector();
		synchronized(mSynapses) {
			for(Iterator i = mSynapses.iterator(); i.hasNext();) {
				Synapse s = (Synapse) ((Synapse) i.next()).clone();
				obj.mSynapses.add(s);
				s.setTargetNeuron(obj);
			}
		}
		obj.mActivityHistory = new LinkedList();
		obj.mType = mType;
		obj.mProtection = mProtection;
		obj.mId = mId; //ids are shared
		obj.mMaxIncomingSynapses = new IntValue(mMaxIncomingSynapses.get());
		obj.mMaxOutgoingSynapses = new IntValue(mMaxOutgoingSynapses.get());
		
		return obj;
	}
	
	
	
	
}
