package de.fhg.iais.outdoor.simba.simulators;

import java.awt.Color;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Random;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.ColorBeacon;
import de.fhg.iais.outdoor.simba.devices.DistanceScanner;
import de.fhg.iais.outdoor.simba.devices.LightSensor;
import de.fhg.iais.outdoor.simba.devices.OmniDrive;
import de.fhg.iais.outdoor.simba.devices.SimpleBumpDetector;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.SimbaDefaultFrame;

public class Memory extends SimbaDefaultFrame {
	
	public NeuralNetAgentInterface mMainController;
	
	public Memory(String args[]) {
		super("Memory", 700, 700);
		
		setDefaultOffset(-430.0, -100.0);
		setDefaultZoom(1.0);
		setVisualizationInterval(1);
		setSimulationDelay(15);
		
		mMainController = Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface();
		
		//Add Fitness function.
		//setFitnessFunction(new PathMemorizerFitness3(), 10000);
		
		setMasterEnvironment(new MyEnvironment(10));
		Simulator.getInstance().getNeuralNetManager().setNeuralNetInterface(mMainController);

		startSystem(mMainController, args);
	}

	
public class MyEnvironment extends EnvironmentModel {
		
		private Box mLeftBox;
		private Box mRightBox;
		private Box mFloor;
		private OmniDrive mRobot;
		private Box mRobotBody;
		private DistanceScanner mDistanceLeft;
		private DistanceScanner mDistanceRight;
		private SimpleBumpDetector mBumpObstacles;
		private LightSensor mLightSensor;
		private int mMaxNumberOfSpots;
		private IntValue mNumberOfSpots;
		private int mNumberOfSpotsBuffer;
		private double mRobotStartPosition;
		private Vector mLightSignals = new Vector();
		private Hashtable mColorTable = new Hashtable();
		private DoubleValue mBeaconDistance;
		private Color[] mColorSequence;
		private Color[] mColorInitialDistribution;
		

		
		public MyEnvironment(int maxNumberOfSpots) {
			super("PathMemorizer");
			mMaxNumberOfSpots = maxNumberOfSpots;
			
			mNumberOfSpots = new IntValue(2);
			mBeaconDistance = new DoubleValue(100.0);
			
			addParameter(mNumberOfSpots, "NumberOfSpots");
			addParameter(mBeaconDistance, "BeaconDistance");
			
			//initialize color table
			mColorTable.put(Color.YELLOW, new Double(0.15));
			mColorTable.put(Color.RED, new Double(0.3));
			mColorTable.put(Color.GREEN, new Double(0.45));
			mColorTable.put(Color.BLUE, new Double(0.6));
			mColorTable.put(Color.CYAN, new Double(0.75));
			mColorTable.put(Color.MAGENTA, new Double(0.9));
			
			mColorSequence = new Color[mColorTable.size()];
			mColorInitialDistribution = new Color[maxNumberOfSpots];
			
		}
		
		public void resetEnvironmentObjects() {
			double beaconDistance = mBeaconDistance.get();
			
			mRobot.setPosition(mRobotStartPosition, mRobot.getPosition().y);
			mRightBox.setPosition((mNumberOfSpotsBuffer * beaconDistance) + 10.0, 15.0);
			mFloor.setWidth(30.0 + (beaconDistance * mNumberOfSpotsBuffer));
			mFloor.setPosition((mNumberOfSpotsBuffer) * (beaconDistance / 2.0), -3);
			
			int i = 0;
			for(; i < mNumberOfSpotsBuffer; ++i) {
				ColorBeacon beacon = (ColorBeacon) mLightSignals.elementAt(i);
				beacon.setPosition((beaconDistance / 2.0) + (i * beaconDistance), 80.0);
				beacon.setActive(true);
				beacon.getLightArea().setActive(true);
				beacon.getLightArea().setAngle(250.0);
			}
			for(; i < mMaxNumberOfSpots; ++i) {
				ColorBeacon beacon = (ColorBeacon) mLightSignals.elementAt(i);
				beacon.setActive(false);
				beacon.getLightArea().setActive(false);
			}
		}
		
		public void updateRandomizationParameters() {
			double beaconDistance = mBeaconDistance.get();
			
			mNumberOfSpotsBuffer = Math.min(Math.max(2, mNumberOfSpots.get()), mMaxNumberOfSpots);
			Random rand = new Random();
			mRobotStartPosition = (beaconDistance / 2.0) + (rand.nextInt(mNumberOfSpotsBuffer) * beaconDistance);
			
			LinkedList colors = new LinkedList(mColorTable.keySet());
			for(int i = 0; i < mColorSequence.length; ++i) {
				Color color = (Color) colors.get(rand.nextInt(colors.size()));
				mColorSequence[i] = color;
				colors.remove(color);
			}
			
			colors = new LinkedList(mColorTable.keySet());
			for(int i = 0; i < mColorInitialDistribution.length && !colors.isEmpty(); ++i) {
				Color color = (Color) colors.get(rand.nextInt(colors.size()));
				mColorSequence[i] = color;
				colors.remove(color);
			}
			
		}
		
	
		public String createRandomizationParameterLog() {
			String log =  super.createRandomizationParameterLog();
			log += getParamEntry("StartingPosition", Double.toString(mRobotStartPosition));
			return log;
		}
		
		

		protected synchronized Object clone() throws CloneNotSupportedException {
			MyEnvironment env = (MyEnvironment) super.clone();
			env.mColorSequence = new Color[mColorTable.size()];
			env.mColorInitialDistribution = new Color[mMaxNumberOfSpots];
			return env;
		}

		public void createEnvironment() {
	
			ValueManager vm = Simulator.getInstance().getValueManager();
			
			//Set up scenario.
			mLeftBox = new Box("LeftBound", -10.0, 15.0, 0.0, 10.0, 30.0);
			mLeftBox.setShapeColor(Color.DARK_GRAY);
			addSimObject(mLeftBox);
			
			mRightBox = new Box("RightBound", 210.0, 15.0, 0.0, 10.0, 30.0);
			mRightBox.setShapeColor(Color.DARK_GRAY);
			addSimObject(mRightBox);
			
			mFloor = new Box("Floor", 100.0, -3.0, 0.0, 230.0, 6.0);
			mFloor.setShapeColor(Color.DARK_GRAY);
			addSimObject(mFloor);
			

			mRobot = new OmniDrive("Robot", 0.0, 10.0, 0.0, -3.0, 3.0, 0.0, 0.0);
			mRobotBody = new Box("Robot/Body", 0.0, 0.0, 0.0, 20.0, 20.0);
			mRobotBody.setReferenceObject(mRobot);
			mRobotBody.setShapeColor(Color.GREEN);
			addSimObject(mRobot);
			addSimObject(mRobotBody);
			//output neuron for motor.
			{
				String name = "/SimObjects/" + mRobot.getName() + "/VelocityX";
				NeuronWrapper wrapper = new NeuronWrapper(name, mRobot.getVelocityX(), 
						mRobot.getMinVelocity(), mRobot.getMaxVelocity());
				vm.replaceValue(name, wrapper);
				mMainController.addOutputNeuron(wrapper);
			}
			
			mDistanceLeft = new DistanceScanner("DistanceScannerLeft", -10.0, 8.0, -180.0, -50.0, 50.0);
			mDistanceLeft.addObstacle(mLeftBox);
			mDistanceLeft.setReferenceObject(mRobotBody);
			addSimObject(mDistanceLeft);
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mDistanceLeft));
			
			
			mDistanceRight = new DistanceScanner("DistanceScannerRight", 10.0, 8.0, 0.0, -50.0, 50.0);
			mDistanceRight.addObstacle(mRightBox);
			mDistanceRight.setReferenceObject(mRobotBody);
			addSimObject(mDistanceRight);
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mDistanceRight));
			
			mBumpObstacles = new SimpleBumpDetector("BumpObstacles", 0, 0.0, 0.0);
			mBumpObstacles.setShape(mRobotBody.getShape());
			mBumpObstacles.setReferenceObject(mRobotBody);
			mBumpObstacles.addObstacle(mLeftBox);
			mBumpObstacles.addObstacle(mRightBox);
			addSimObject(mBumpObstacles);
			Simulator.getInstance().getScheduler().addTerminateTryValue(mBumpObstacles.getSensorValueObject());
			
			mLightSensor = new LightSensor("LightSensor", 0.0, 0.0, 0.0, 0.0, 1.0, true);
			mLightSensor.setReferenceObject(mRobot);
			addSimObject(mLightSensor);
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForLightSensor(mLightSensor));
			
			//Color beacons
			for(int i = 0; i < mMaxNumberOfSpots; ++i) {
				ColorBeacon beacon = new ColorBeacon("Beacon" + i, 0.0, 0.0, 0.0, 80.0, 8, 40.0);
				beacon.setColorTable(mColorTable);
				mLightSignals.add(beacon);
				addSimObject(beacon);
				mLightSensor.addLightSource(beacon);
			}
			
			
		}
		
	}


	public class LightSignalController extends SimObject {
		
		
		
		
		public LightSignalController() {
			
		}

		public boolean init() {
			super.init();
			
			return true;
		}

		public void reset() {
			super.reset();
		}

		public boolean update() {
			super.update();
			
			
			return true;
		}
		
		
		
	}
	
	
	
	public static void main(String[] args) {
		new Memory(args);
	}
	
}
