package de.fhg.iais.outdoor.simba.simulators;

import java.awt.Color;
import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.Random;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.SimObjectSet;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.DifferentialDriveUnit;
import de.fhg.iais.outdoor.simba.devices.DistanceScanner;
import de.fhg.iais.outdoor.simba.devices.DistanceToPointScanner;
import de.fhg.iais.outdoor.simba.devices.PathDrawer;
import de.fhg.iais.outdoor.simba.devices.SimpleBumpDetector;
import de.fhg.iais.outdoor.simba.fitnessfunctions.ForwardMoverFitness;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.SimbaDefaultFrame;

public class ObstacleBox extends SimbaDefaultFrame {
	
	private NeuralNetAgentInterface mNeuronInterface;
	
	public ObstacleBox(String[] args) {
		super("ObstacleBox", 700, 700);
		setDefaultOffset(-900.0, -900.0);
		setDefaultZoom(0.3);
		setVisualizationInterval(1);
		setSimulationDelay(15);
		
		mNeuronInterface = Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface();
		setFitnessFunction(new ForwardMoverFitness(), 10000);
		
		setMasterEnvironment(new BoxEnvironment(10));
		Simulator.getInstance().getNeuralNetManager().setCurrentNet(NeuralNet.createMinimalNet(mNeuronInterface));
		
		startSystem(mNeuronInterface, args);
	}
	
	public class BoxEnvironment extends EnvironmentModel {
		
		private double[] mObstacleXPositions;
		private double[] mObstacleYPositions;
		private double[] mObstacleAngles;
		private Point2D.Double mRobotPosition;
		private double mRobotAngle;
		private SimObjectSet mObstacleList;
		
		private IntValue mNumberOfObstacles;
		private IntValue mTargetZones;
		private DoubleValue mMinRobotAngle;
		private DoubleValue mMaxRobotAngleAmount;
		private PointValue mMinRobotPosition;
		private PointValue mMaxRobotPosition;
		private IntValue mBoxWidth;
		private IntValue mBoxHeight;
		private DoubleValue mHomeZoneSize;
		private int mMaxObstacles;
		private DoubleValue mScannerMinRange;
		private DoubleValue mScannerMaxRange;
		private DoubleValue mScannerNoise;
		private DoubleValue mMinObstAngle;
		private DoubleValue mMaxObstAngleAmount;
		private PointValue mMinObstPosition;
		private PointValue mMaxObstPosition;
		private DoubleValue mObstacleWidth;
		private BooleanValue mRandomizeObstacles;
		private BooleanValue mSwitchObstacleActivityAtZones;
		private BooleanValue mSwitchObstacleSeningAtZones;
		
		
		private DifferentialDriveUnit mRobot;
		private SimpleBumpDetector mTargetAreaBumper;
		private SimpleBumpDetector mObstacleBumper;
		private DistanceScanner mScanLeft;
		private DistanceScanner mScanRight;
		private DistanceScanner mScanMid;
		private DistanceToPointScanner mHomeZoneDistance;
		private PathDrawer mPathDrawer;
		private Box mTargetZone;
		private Box mHomeZone;
		private Box[] mObstacles;
		private Box mTopWall;
		private Box mBottomWall;
		private Box mLeftWall;
		private Box mRightWall;
		
		
		private Random mRandom = new Random();
		
		public BoxEnvironment(int maxObstacles) {
			super("BoxEnvironment");
			
			mMaxObstacles = maxObstacles;
			
			mObstacleXPositions = new double[mMaxObstacles];
			mObstacleYPositions = new double[mMaxObstacles];
			mObstacleAngles = new double[mMaxObstacles];
			mObstacles = new Box[mMaxObstacles];
			

			mNumberOfObstacles = new IntValue(2);
			mTargetZones = new IntValue(1);
			mMinRobotAngle = new DoubleValue(-45.0);
			mMaxRobotAngleAmount = new DoubleValue(45.0);
			mBoxWidth = new IntValue(400);
			mBoxHeight = new IntValue(700);
			mMinRobotPosition = new PointValue(50.0, 50.0);
			mMaxRobotPosition = new PointValue(350.0, 100.0);
			mHomeZoneSize = new DoubleValue(50.0);
			mScannerMinRange = new DoubleValue(-50.0);
			mScannerMaxRange = new DoubleValue(50.0);
			mScannerNoise = new DoubleValue(3.0);
			mMinObstAngle = new DoubleValue(45.0);
			mMaxObstAngleAmount = new DoubleValue(0.0);
			mMinObstPosition = new PointValue(50.0, 200.0);
			mMaxObstPosition = new PointValue(350.0, 500.0);
			mObstacleWidth = new DoubleValue(100.0);
			mRandomizeObstacles = new BooleanValue(true);
			mSwitchObstacleActivityAtZones = new BooleanValue(false);
			mSwitchObstacleSeningAtZones = new BooleanValue(false);
			
			addParameter(mNumberOfObstacles, "NumberOfObstacles (0 - " + maxObstacles + ")");
			addParameter(mTargetZones, "TargetZone");
			addParameter(mMinRobotAngle, "MinRobotAngle");
			addParameter(mMaxRobotAngleAmount, "MaxRobotAngleAmount");
			addParameter(mBoxWidth, "AreaWidth");
			addParameter(mBoxHeight, "AreaHeight");
			addParameter(mMinRobotPosition, "MinRobotInitPosition");
			addParameter(mMaxRobotPosition, "MaxRobotInitPosition");
			addParameter(mHomeZoneSize, "HomeZoneSize");
			addParameter(mScannerMaxRange, "ScannerMaxRange");
			addParameter(mScannerMinRange, "ScannerMinRange");
			addParameter(mMaxObstAngleAmount, "MaxObstAngleAmount");
			addParameter(mMinObstAngle, "MinObstAngle");
			addParameter(mMinObstPosition, "MinObstPosition");
			addParameter(mMaxObstPosition, "MaxObstPosition");
			addParameter(mObstacleWidth, "ObstWidth");
			addParameter(mRandomizeObstacles, "RandomizeObst");
			addParameter(mSwitchObstacleActivityAtZones, "SwitchObstActivity");
			addParameter(mSwitchObstacleSeningAtZones, "SwitchObstSensing");
		
		}
		
		

		protected synchronized Object clone() throws CloneNotSupportedException {
			BoxEnvironment box = (BoxEnvironment) super.clone();
			
			box.mObstacleXPositions = new double[mMaxObstacles];
			box.mObstacleYPositions = new double[mMaxObstacles];
			box.mObstacleAngles = new double[mMaxObstacles];
			box.mRobotPosition = new Point2D.Double();
			return box;
		}

		
		public void updateRandomizationParameters() {
			
			for(int i = 0; i < mMaxObstacles; ++i) {
				mObstacleAngles[i] = mMinObstAngle.get() + (mRandom.nextDouble() * mMaxObstAngleAmount.get());
				mObstacleXPositions[i] = mMinObstPosition.getX() + mRandom.nextInt((int) (mMaxObstPosition.getX() - mMinObstPosition.getX()));
				mObstacleYPositions[i] = mMinObstPosition.getY() + mRandom.nextInt((int) (mMaxObstPosition.getY() - mMinObstPosition.getY()));
			}
			
			mRobotAngle = mMinRobotAngle.get() + (mRandom.nextDouble() * mMaxRobotAngleAmount.get());
			double posX = mMinRobotPosition.get().x + (mRandom.nextDouble() * (mMaxRobotPosition.get().x - mMinRobotPosition.get().x));
			double posY = mMinRobotPosition.get().y + (mRandom.nextDouble() * (mMaxRobotPosition.get().y - mMinRobotPosition.get().y));
			mRobotPosition.setLocation(posX, posY);
			
		}

		

		public String createRandomizationParameterLog() {
			return super.createRandomizationParameterLog();
		}

		public void resetEnvironmentObjects() {
	
			mRobot.setPosition(mRobotPosition);
			mRobot.setAngle(mRobotAngle);
			
			mHomeZone.setPosition(mRobotPosition);
			mHomeZone.setAngle(0.0);
			mHomeZone.setSize(mHomeZoneSize.get(), mHomeZoneSize.get());
			
			mTopWall.setSize(mBoxWidth.get() + 60.0, 30.0);
			mTopWall.setAngle(0.0);
			mTopWall.setPosition(0.0, mBoxHeight.get() + 15.0);
			
			mBottomWall.setSize(mBoxWidth.get() + 60.0, 30.0);
			mBottomWall.setAngle(0.0);
			mBottomWall.setPosition(0.0, -15.0);
			
			mRightWall.setSize(30.0, mBoxHeight.get() + 60.0);
			mRightWall.setAngle(0.0);
			mRightWall.setPosition((mBoxWidth.get() / 2.0) + 15.0, mBoxHeight.get() / 2.0);
			
			mLeftWall.setSize(30.0, mBoxHeight.get() + 60.0);
			mLeftWall.setAngle(0.0);
			mLeftWall.setPosition((mBoxWidth.get() / -2.0) - 15.0, mBoxHeight.get() / 2.0);
			
			switch(mTargetZones.get()) {
			case 0:
				mTargetZone.setSize(mBoxWidth.get() + 60.0, 100.0);
				mTargetZone.setPosition(0.0, mBoxHeight.get() - 50.0);
				break;
			case 1:
				mTargetZone.setSize(100.0, mBoxHeight.get() + 60.0);
				mTargetZone.setPosition((mBoxWidth.get() / 2.0) - 50.0, mBoxHeight.get() / 2.0);
				break;
			case 2:
				mTargetZone.setSize(mBoxWidth.get() + 60.0, 100.0);
				mTargetZone.setPosition(0.0, 50.0);
				break;
			case 3:
				mTargetZone.setSize(100.0, mBoxHeight.get() + 60.0);
				mTargetZone.setPosition((mBoxWidth.get() / -2.0) + 50.0, mBoxHeight.get() / 2.0);
				break;
			default:
			}
			
			for(int i = 0; i < mMaxObstacles; ++i) {
				Box box = (Box) mObstacles[i];
				if(i < mNumberOfObstacles.get()) {
					box.setActive(true);
					box.setSize(mObstacleWidth.get(), 30.0);
					if(mRandomizeObstacles.get()) {
						box.setPosition(mObstacleXPositions[i], mObstacleYPositions[i]);
						box.setAngle(mObstacleAngles[i]);
					}
				}
				else {
					box.setActive(false);
				}
			}
			
			mScanLeft.setNoiseVariance(mScannerNoise.get());
			mScanLeft.setMinScanDistance(mScannerMinRange.get());
			mScanLeft.setMaxScanDistance(mScannerMaxRange.get());
			mScanLeft.addObstacleSet(mObstacleList);
			mScanRight.setNoiseVariance(mScannerNoise.get());
			mScanRight.setMinScanDistance(mScannerMinRange.get());
			mScanRight.setMaxScanDistance(mScannerMaxRange.get());
			mScanRight.addObstacleSet(mObstacleList);
			mScanMid.setNoiseVariance(mScannerNoise.get());
			mScanMid.setMinScanDistance(mScannerMinRange.get());
			mScanMid.setMaxScanDistance(mScannerMaxRange.get());
			mScanMid.addObstacleSet(mObstacleList);
		}

		
		
		public void createEnvironment() {
			mRobot = new DifferentialDriveUnit("Robot", 0.0, 10.0, 0.0, 20.0, 5.0, -1.5, 1.5);
			Box robotBody = new Box("Robot/Body", 0.0, 0.0, 90.0, 14.0, 20.0);
			robotBody.setReferenceObject(mRobot);
			robotBody.setShapeColor(Color.GREEN);
			addSimObject(mRobot);
			addSimObject(robotBody);
			
			//Add motor neurons
			ValueManager vm = Simulator.getInstance().getValueManager();
			String name = "/SimObjects/" + mRobot.getName() + "/VelocityLeft";
			NeuronWrapper motorLeft = new NeuronWrapper(name, 0.0, mRobot.getMinVelocity(), mRobot.getMaxVelocity());
			vm.replaceValue(name, motorLeft);
			mNeuronInterface.addOutputNeuron(motorLeft);
			
			name = "/SimObjects/" + mRobot.getName() + "/VelocityRight";
			NeuronWrapper motorRight = new NeuronWrapper(name, 0.0, mRobot.getMinVelocity(), mRobot.getMaxVelocity());
			vm.replaceValue(name, motorRight);
			mNeuronInterface.addOutputNeuron(motorRight);
			
			
			mPathDrawer = new PathDrawer("RobotTrajectory", 0.0, 0.0, 2, 100, 1000);
			mPathDrawer.setReferenceObject(mRobot);
			mPathDrawer.setLayer(1);
			mPathDrawer.setShapeColor(Color.blue);
			mPathDrawer.setSecondColor(Color.cyan);
			addSimObject(mPathDrawer);
			
			mTargetZone = new Box("TargetZone", 0.0, 0.0, 0.0, 0.0, 0.0);
			mTargetZone.setLayer(-5);
			mTargetZone.setShapeColor(Color.LIGHT_GRAY);
			addSimObject(mTargetZone);
			
			mHomeZone  = new Box("HomeZone", 0.0, 0.0, 0.0, 0.0, 0.0);
			mHomeZone.setLayer(-5);
			mHomeZone.setShapeColor(Color.LIGHT_GRAY);
			addSimObject(mHomeZone);
			
			//Obstacles
			mObstacleList = new SimObjectSet();
			for(int i = 0; i < mObstacles.length; ++i) {
				Box box = new Box("Obstacle" + i, 0.0, 0.0, 0.0, 0.0, 0.0);
				box.setLayer(0);
				box.setShapeColor(Color.BLACK);
				box.setFixedInitPosition(false);
				addSimObject(box);
				mObstacleList.add(box);
				mObstacles[i] = box;
				addParameter(box.getPositionValue(), "ObstPos" + i);
				addParameter(box.getAnglePosition(), "ObstAng" + i);
			}
			
			mTopWall = new Box("TopWall", 0.0, 0.0, 0.0, 0.0, 0.0);
			mTopWall.setShapeColor(Color.BLACK);
			addSimObject(mTopWall);
			mObstacleList.add(mTopWall);
			
			mBottomWall = new Box("BottomWall", 0.0, 0.0, 0.0, 0.0, 0.0);
			mBottomWall.setShapeColor(Color.BLACK);
			addSimObject(mBottomWall);
			mObstacleList.add(mBottomWall);
			
			mLeftWall = new Box("LeftWall", 0.0, 0.0, 0.0, 0.0, 0.0);
			mLeftWall.setShapeColor(Color.BLACK);
			addSimObject(mLeftWall);
			mObstacleList.add(mLeftWall);
			
			mRightWall = new Box("RightWall", 0.0, 0.0, 0.0, 0.0, 0.0);
			mRightWall.setShapeColor(Color.BLACK);
			addSimObject(mRightWall);
			mObstacleList.add(mRightWall);
			
			mScanLeft = new DistanceScanner("DistanceScannerLeft", 0.0, 10.0, 50.0, -100.0, 100.0);
			mScanLeft.setReferenceObject(robotBody);
			mScanLeft.addDependencies(mObstacleList);
			mScanLeft.addObstacleSet(mObstacleList);
			mScanLeft.setNoiseVariance(1.0);
			addSimObject(mScanLeft);
			mNeuronInterface.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mScanLeft));
			
			mScanRight = new DistanceScanner("DistanceScannerRight", 0.0, -10.0, -50.0, -100.0, 100.0);
			mScanRight.setReferenceObject(robotBody);
			mScanRight.addDependencies(mObstacleList);
			mScanRight.addObstacleSet(mObstacleList);
			mScanRight.setNoiseVariance(1.0);
			addSimObject(mScanRight);
			mNeuronInterface.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mScanRight));
			
			mScanMid = new DistanceScanner("DistanceScannerFront", 7.0, 0.0, 0.0, -80.0, 80.0);
			mScanMid.setReferenceObject(robotBody);
			mScanMid.addObstacleSet(mObstacleList);
			mScanMid.addObstacleSet(mObstacleList);
			mScanMid.setNoiseVariance(1.0);
			addSimObject(mScanMid);
			mNeuronInterface.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mScanMid));
			
			mObstacleBumper = new SimpleBumpDetector("BumpObstacles", 0.0, 0.0, 0.0);
			Polygon p = robotBody.getShape();
			mObstacleBumper.setShape(new Polygon(p.xpoints, p.ypoints, p.npoints));
			mObstacleBumper.setReferenceObject(robotBody);
			mObstacleBumper.addDependencies(mObstacleList);
			mObstacleBumper.addObstacleSet(mObstacleList);
			addSimObject(mObstacleBumper);
			Simulator.getInstance().getScheduler().addTerminateTryValue(mObstacleBumper.getSensorValueObject());
			
			mTargetAreaBumper = new SimpleBumpDetector("TargetAreaBumper", 0.0, 0.0, 0.0);
			mTargetAreaBumper.setReferenceObject(robotBody);
			mTargetAreaBumper.addObstacle(mTargetZone);
			addSimObject(mTargetAreaBumper);
			mNeuronInterface.addInputNeuron(NeuronWrapper.createWrapperForBumpDetector(mTargetAreaBumper));
			
			
			//Add SwitchObstacleActivityController
			SwitchObstacleActivityController actController = new SwitchObstacleActivityController();
			actController.addDependency(mTargetAreaBumper);
			addSimObject(actController);
			
			
			//Sensor only for fitness.
			mHomeZoneDistance = new DistanceToPointScanner("HomeZoneDistance", 0.0, 0.0, 0.0, 0.0, 1000.0);
			mHomeZoneDistance.addDependency(mHomeZone);
			mHomeZoneDistance.setReferenceObject(mRobot);
			mHomeZoneDistance.addObstacle(mHomeZone);
			addSimObject(mHomeZoneDistance);
		}
		
		public class SwitchObstacleActivityController extends SimObject implements ValueChangedListener {
			
			private boolean mLookingForTargetZone;
			
			public SwitchObstacleActivityController() {
				super("SwitchObstacleController");
				
				mTargetAreaBumper.getSensorValueObject().addValueChangedListener(this);
			}
			
			

			public void reset() {
				super.reset();
				mLookingForTargetZone = true;
			}



			public void valueChanged(Value value) {
				super.valueChanged(value);
				if(value == null) {
					return;
				}
				if(value == mTargetAreaBumper.getSensorValueObject()) {
					if(mTargetAreaBumper.getSensorValueObject().get()
							&& mLookingForTargetZone) 
					{
						if(mSwitchObstacleActivityAtZones.get()) {
							mLookingForTargetZone = false;
							for(int i = 0; i < mObstacles.length; ++i) {
								mObstacles[i].setActive(false);
							}
						}
						if(mSwitchObstacleSeningAtZones.get()) {
							mScanLeft.removeObstacleSet(mObstacleList);
							mScanRight.removeObstacleSet(mObstacleList);
							mScanMid.removeObstacleSet(mObstacleList);
						}
					}
				}
			}
			
			
			
		}
	}

	
	public static void main(String args[]) {
		new ObstacleBox(args);
	}

}
