package de.fhg.iais.outdoor.simba.simulators;

import java.awt.Color;
import java.util.Random;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.ParameterElement;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Beacon;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.DistanceScanner;
import de.fhg.iais.outdoor.simba.devices.LightSensor;
import de.fhg.iais.outdoor.simba.devices.OmniDrive;
import de.fhg.iais.outdoor.simba.devices.SimpleBumpDetector;
import de.fhg.iais.outdoor.simba.fitnessfunctions.SimpleBackAndForthScenarioFitness1;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.SimbaDefaultFrame;

public class SimpleBackAndForthMidLight extends SimbaDefaultFrame {

public NeuralNetAgentInterface mMainController;
	
	public SimpleBackAndForthMidLight(String args[]) {
		super("SimpleBackAndForthMidLight", 700, 500);

		setDefaultOffset(-235.0, -190.0);
		setDefaultZoom(1.0);
		setVisualizationInterval(1);
		setSimulationDelay(15);
		
		mMainController = Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface();
		
		setFitnessFunction(new SimpleBackAndForthScenarioFitness1(0.1), 10000);
		
		
		setMasterEnvironment(new MyEnvironment());
		
		Simulator.getInstance().getNeuralNetManager().setNeuralNetInterface(mMainController);
		Simulator.getInstance().getNeuralNetManager().setCurrentNet(NeuralNet.createMinimalNet(mMainController));
		
		startSystem(mMainController, args);
	}

	
	
	
	
	public class MyEnvironment extends EnvironmentModel {
		
		private Box mLeftBox;
		private Box mRightBox;
		private Box mFloor;
		private OmniDrive mRobot;
		private Box mRobotBody;
		private DistanceScanner mDistanceLeft;
		private DistanceScanner mDistanceRight;
		private SimpleBumpDetector mBumpLeft;
		private SimpleBumpDetector mBumpRight;
		private DoubleValue mSensorNoise;
		private DoubleValue mMinSensorValue;
		private DoubleValue mMaxSensorValue;
		private BooleanValue mInvertSensorReadings;
		private Beacon mMidLight;
		private LightSensor mLightSensor;
		private DoubleValue mMinMidLightXPosition;
		private DoubleValue mMaxMidLightXPosition;
		private double mMidLightXPosition;
		
		private double mStartPositionOffset;
		
		public MyEnvironment() {
			super("BackAndForthScenarioShortScanner");
			
			mSensorNoise = new DoubleValue(0.0);
			addParameter(mSensorNoise, "SensorNoise");
			
			mMinSensorValue = new DoubleValue(-1);
			addParameter(mMinSensorValue, "MinSensorValue");
			
			mMaxSensorValue = new DoubleValue(1);
			addParameter(mMaxSensorValue, "MaxSensorValue");
			
			mInvertSensorReadings = new BooleanValue(false);
			addParameter(mInvertSensorReadings, "InvertSensorReadings");
			
			mMinMidLightXPosition = new DoubleValue(0.0);
			addParameter(mMinMidLightXPosition, "MinBeaconXPosition");
			
			mMaxMidLightXPosition = new DoubleValue(0.0);
			addParameter(mMaxMidLightXPosition, "MaxBeaconXPosition");
			
		}

		public void resetEnvironmentObjects() {
			mRobot.setPosition(mStartPositionOffset, mRobot.getPosition().y);
			mDistanceLeft.setNoiseVariance(mSensorNoise.get());
			mDistanceRight.setNoiseVariance(mSensorNoise.get());
			
			mDistanceLeft.getDistanceValue().setNormalizedValueRange(mMinSensorValue.get(), mMaxSensorValue.get());
			mDistanceRight.getDistanceValue().setNormalizedValueRange(mMinSensorValue.get(), mMaxSensorValue.get());
			
			mDistanceLeft.getDistanceValue().setInverted(mInvertSensorReadings.get());
			mDistanceRight.getDistanceValue().setInverted(mInvertSensorReadings.get());
			
			mMidLight.setPosition(mMidLightXPosition, 80.0);
			mMidLight.getLightArea().setAngle(250.0);
		}
		
		public void updateRandomizationParameters() {
			Random rand = new Random();
			mStartPositionOffset = -50.0 + (rand.nextDouble() * 100.0);
			
			double range = Math.max(0.0, mMaxMidLightXPosition.get() - mMinMidLightXPosition.get());
			mMidLightXPosition = mMinMidLightXPosition.get() + (rand.nextDouble() * range);
			
		}
		
		public String createRandomizationParameterLog() {
			StringBuffer entry = new StringBuffer();
			entry.append(ParameterElement.getParamEntry("StartPosition", Double.toString(mStartPositionOffset)));
			entry.append(ParameterElement.getParamEntry("BeaconXPosition", Double.toString(mMidLightXPosition)));
			
			return entry.toString();
		}

		public void createEnvironment() {

			ValueManager vm = Simulator.getInstance().getValueManager();
			
			//Set up scenario.
			mLeftBox = new Box("LeftBound", -200.0, 15.0, 0.0, 10.0, 30.0);
			mLeftBox.setShapeColor(Color.DARK_GRAY);
			addSimObject(mLeftBox);
			
			mRightBox = new Box("RightBound", 200.0, 15.0, 0.0, 10.0, 30.0);
			mRightBox.setShapeColor(Color.DARK_GRAY);
			addSimObject(mRightBox);
			
			mFloor = new Box("Floor", 0.0, -3.0, 0.0, 410.0, 6.0);
			mFloor.setShapeColor(Color.DARK_GRAY);
			addSimObject(mFloor);
			
			mMidLight = new Beacon("Beacon", 0.0, 80.0, 0.0, 90.0, 3, 40.0, 1.0, 1.0);
			mMidLight.setShapeColor(Color.RED);
			addSimObject(mMidLight);
			
			mRobot = new OmniDrive("Robot", 0.0, 10.0, 0.0, -3.0, 3.0, 0.0, 0.0);
			mRobotBody = new Box("Robot/Body", 0.0, 0.0, 0.0, 20.0, 20.0);
			mRobotBody.setReferenceObject(mRobot);
			mRobotBody.setShapeColor(Color.GREEN);
			addSimObject(mRobot);
			addSimObject(mRobotBody);
			//output neuron for motor.
			{
				String name = "/SimObjects/" + mRobot.getName() + "/VelocityX";
				NeuronWrapper wrapper = new NeuronWrapper(name, mRobot.getVelocityX(), 
						mRobot.getMinVelocity(), mRobot.getMaxVelocity());
				vm.replaceValue(name, wrapper);
				mMainController.addOutputNeuron(wrapper);
			}
			
			mLightSensor = new LightSensor("LightSensor", 0.0, 0.0, 0.0, 0.0, 1.0);
			mLightSensor.setReferenceObject(mRobot);
			mLightSensor.addLightSource(mMidLight);
			mLightSensor.addDependency(mMidLight);
			addSimObject(mLightSensor);
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForLightSensor(mLightSensor));
			
			mDistanceLeft = new DistanceScanner("DistanceScannerLeft", -10.0, 8.0, -180.0, 0.0, 100.0);
			mDistanceLeft.addObstacle(mLeftBox);
			mDistanceLeft.setReferenceObject(mRobotBody);
			mDistanceRight = new DistanceScanner("DistanceScannerRight", 10.0, 8.0, 0.0, 0.0, 100.0);
			mDistanceRight.addObstacle(mRightBox);
			mDistanceRight.setReferenceObject(mRobotBody);
			addSimObject(mDistanceLeft);
			addSimObject(mDistanceRight);
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mDistanceLeft));
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mDistanceRight));
			
			mBumpLeft = new SimpleBumpDetector("BumpLeft", -10, 0.0, 180.0);
			mBumpLeft.setReferenceObject(mRobotBody);
			mBumpLeft.addObstacle(mLeftBox);
			addSimObject(mBumpLeft);
			Simulator.getInstance().getScheduler().addTerminateTryValue(mBumpLeft.getSensorValueObject());
			
			mBumpRight = new SimpleBumpDetector("BumpRight", 10, 0.0, 180.0);
			mBumpRight.setReferenceObject(mRobotBody);
			mBumpRight.addObstacle(mRightBox);
			addSimObject(mBumpRight);
			Simulator.getInstance().getScheduler().addTerminateTryValue(mBumpRight.getSensorValueObject());
			
			mMainController.setNeuronRange(-1.0, 1.0); //tanh
			
		}
		
	}

	
	public static void main(String[] args) {
		new SimpleBackAndForthMidLight(args);
	}
	
}
