package de.fhg.iais.outdoor.simba.test;

import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;


import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.Sensor;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.DistanceScanner;
import junit.framework.TestCase;

public class TestEnvironmentModel extends TestCase {

	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	
	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.EnvironmentModel.reset()'
	 */
	public void testReset() {
		TestEnvironment env = new TestEnvironment();
		
		MyTestSimObject obj1 = new MyTestSimObject();
		MyTestSimObject obj2 = new MyTestSimObject();
		obj1.addDependency(obj2);
		
		MasterCounter counter = new MasterCounter();
		env.mMasterCounter = counter;
		obj1.mMasterCounter = counter;
		obj2.mMasterCounter = counter;
		
		assertEquals(counter.mCounter, 0);
		
		env.addSimObject(obj1);
		env.initEnvironment();
		assertEquals(env.getSimObjects().size(), 1);
		
		env.addSimObject(obj2); //still in addQueue (not added to SimObjects.
		assertEquals(env.getSimObjects().size(), 1);
		
		assertEquals(env.mResetEnvironmentObjectsCounter, 0);
		assertEquals(env.mUpdateRandomizationCounter, 0);
		assertEquals(obj1.mResetCounter, 0);
		assertEquals(obj2.mResetCounter, 0);
		
		env.reset(); //updates the add / remove list before the objects are reseted.
		
		assertEquals(counter.mCounter, 3);
		//Following order, each method called exactly once: obj2, obj1, env.
		assertEquals(obj2.mResetCounter, 1);
		assertEquals(obj1.mResetCounter, 2);
		assertEquals(env.mResetEnvironmentObjectsCounter, 3);
		assertEquals(env.mUpdateRandomizationCounter, 0);
	}

	
	public void testLayeredObjects() {
		EnvironmentModel model = new TestEnvironment();
		
		 SimObject o1 = new SimObject("o1", 10.0, 20.0, 0.0, new Polygon());
		 SimObject o2 = new SimObject("o2", 10.0, 20.0, 0.0, new Polygon());
		 SimObject o3 = new SimObject("o3", 10.0, 20.0, 0.0, new Polygon());
		 SimObject o4 = new SimObject("o4", 10.0, 20.0, 0.0, new Polygon());
		 Sensor o5 = new Sensor("sensor", 10.0, 20.0, 0.0);
		 
		 o1.setLayer(10);
		 o2.setLayer(2);
		 o3.setLayer(15);
		 o5.setLayer(8);
		 
		 model.addSimObject(o1);
		 model.addSimObject(o2);
		 model.addSimObject(o3);
		 model.addSimObject(o4);
		 model.addSimObject(o5);
		 
		 Vector objects = model.getSimObjectsSortedByLayers();
		 assertEquals(objects.size(), 5); 
		 
		 model.updateContent(); //this moves the added objects from the add queue to the model.
		 objects = model.getSimObjectsSortedByLayers();
		 assertEquals(objects.size(), 5);
		 assertSame(objects.elementAt(0), o4);
		 assertSame(objects.elementAt(1), o2);
		 assertSame(objects.elementAt(2), o5);
		 assertSame(objects.elementAt(3), o1);
		 assertSame(objects.elementAt(4), o3);
		 
		 //test changed layers.
		 o2.setLayer(20);
		 o5.setLayer(-10);
		 
		 objects = model.getSimObjectsSortedByLayers();
		 assertEquals(objects.size(), 5);
		 assertSame(objects.elementAt(0), o5);
		 assertSame(objects.elementAt(1), o4);
		 assertSame(objects.elementAt(2), o1);
		 assertSame(objects.elementAt(3), o3);
		 assertSame(objects.elementAt(4), o2);
		 
	}

	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.EnvironmentModel.executeSimulationStep()'
	 */
	public void testExecuteSimulationStep() {

		//Create mini simulation environment
		TestActuator robot = new TestActuator("Actuator", 0.0, 0.0, 0.0, 2.0, 2.0, 1.0);
		TestActuator robot2 = new TestActuator("MovingObstacle", 53.0, 0.0, 0.0, 2.0, 10.0, -2.0);
		DistanceScanner distScanner = new DistanceScanner("LeftScanner", 1.0, 1.0, 0.0, 0.0, 100.0);
		Box obstacle = new Box("Obstacle1", 50, 0.0, 0.0, 2.0, 10.0);
		distScanner.setReferenceObject(robot); //automatically added to the dependency list
		distScanner.addObstacle(obstacle);
		distScanner.addObstacle(robot2); //automatically added to the dependency list.
		
		assertEquals(distScanner.getDependencies().size(), 3);
		
		EnvironmentModel model = new TestEnvironment();
		model.addSimObject(distScanner);
		model.addSimObject(robot);
		model.addSimObject(robot2);
		model.addSimObject(obstacle);
		
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(0.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(53.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(1.0, 1.0));
		assertEquals(obstacle.getGlobalPosition(), new Point2D.Double(50.0, 0.0));
		
		model.executeSimulationStep();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(1.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(51.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(2.0, 1.0));
		assertEquals(obstacle.getGlobalPosition(), new Point2D.Double(50.0, 0.0));
		assertEquals(distScanner.getSensorReading(), 47.0, 0.001);
		
		model.executeSimulationStep();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(2.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(49.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(3.0, 1.0));
		assertEquals(distScanner.getSensorReading(), 45.0, 0.001);
		
		model.executeSimulationStep();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(3.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(47.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(4.0, 1.0));
		assertEquals(distScanner.getSensorReading(), 42.0, 0.001);
		
		model.executeSimulationStep();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(2.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(49.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(3.0, 1.0));
		assertEquals(distScanner.getSensorReading(), 45.0, 0.001);
		
		//reset all SimObjects.
		model.reset();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(0.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(53.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(1.0, 1.0));
		assertEquals(obstacle.getGlobalPosition(), new Point2D.Double(50.0, 0.0));
		
		//set robot1 nactive
		robot.setActive(false);
		model.executeSimulationStep();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(0.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(55.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(1.0, 1.0));
		assertEquals(obstacle.getGlobalPosition(), new Point2D.Double(50.0, 0.0));
		
		//activate robot and deactivate robot2
		robot.setActive(true);
		robot2.setActive(false);
		model.executeSimulationStep();
		assertEquals(robot.getGlobalPosition(), new Point2D.Double(-1.0, 0.0));
		assertEquals(robot2.getGlobalPosition(), new Point2D.Double(55.0, 0.0));
		assertEquals(distScanner.getGlobalPosition(), new Point2D.Double(0.0, 1.0));
		assertEquals(obstacle.getGlobalPosition(), new Point2D.Double(50.0, 0.0));
		
		
		//test: if objects are added during step execution, then these objects are integrated after the step automatically
		MyTestSimObject obj1 = new MyTestSimObject();
		SimObject obj2 = new SimObject();
		
		obj1.mModel = model;
		obj1.mObjectToAdd = obj2;
		
		model.addSimObject(obj1);
		
		assertEquals(model.getSimObjects().size(), 4);
		assertFalse(model.getSimObjects().contains(obj1));
		assertFalse(model.getSimObjects().contains(obj2));
		
		model.executeSimulationStep();
		//obj1 was integrated BEFORE the objects are updated, during its update it adds obj2, which is then integrated AFTER the update.
		assertEquals(model.getSimObjects().size(), 6);
		assertTrue(model.getSimObjects().contains(obj1));
		assertTrue(model.getSimObjects().contains(obj2));
		
	}

	

	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.EnvironmentModel.addSimObject(SimObject)'
	 */
	public void testAddSimObject() {
		
		SimObject o1 = new SimObject();
		SimObject o2 = new SimObject();
		SimObject o3 = new SimObject();
		SimObject o4 = new SimObject();
		SimObject o5 = new SimObject();
		SimObject o6 = new SimObject();
		SimObject o7 = new SimObject();
		
		//create dependencies
		o1.addDependency(o2);
		o1.addDependency(o4);
		o1.addDependency(o6);
		
		
		
		EnvironmentModel model = new TestEnvironment();
		assertEquals(model.getSimObjects().size(), 0);
		
		
		assertTrue(model.addSimObject(o1));
		assertEquals(model.getSimObjects().size(), 0); //object in add-queue (not in model).
		
		model.updateContent(); //moves the object from the add-queue to the model.
		assertEquals(model.getSimObjects().size(), 1);
		
		//can not add the same object twice.
		assertFalse(model.addSimObject(o1));
		assertEquals(model.getSimObjects().size(), 1);
		
		o2.setReferenceObject(o3);
		o2.addDependency(o4);
		o2.addDependency(o5);
		
		//add o2. o1 is dependent on o2, so it should be moved behind o2.
		assertTrue(model.addSimObject(o2));
		model.updateContent();
		
		Vector obj = model.getSimObjects();
		assertEquals(obj.size(), 2);
		assertSame(obj.elementAt(0), o2);
		assertSame(obj.elementAt(1), o1);
		
		//add o3. o2 is dependent on o3, so the new order is 03 02 01
		assertTrue(model.addSimObject(o3));
		model.updateContent();
		
		obj = model.getSimObjects();
		assertEquals(obj.size(), 3);
		assertSame(obj.elementAt(0), o3);
		assertSame(obj.elementAt(1), o2);
		assertSame(obj.elementAt(2), o1);
		
		o4.addDependency(o3);
		
		//add o4. 
		assertTrue(model.addSimObject(o4));
		model.updateContent();
		
		obj = model.getSimObjects();
		assertEquals(obj.size(), 4);
		assertSame(obj.elementAt(0), o3);
		assertSame(obj.elementAt(1), o4);
		assertSame(obj.elementAt(2), o2);
		assertSame(obj.elementAt(3), o1);
		
		o5.setReferenceObject(o4);
		
		//add 05
		assertTrue(model.addSimObject(o5));
		model.updateContent();
		
		obj = model.getSimObjects();
		assertEquals(obj.size(), 5);
		assertSame(obj.elementAt(0), o3);
		assertSame(obj.elementAt(1), o4); 
		assertSame(obj.elementAt(2), o5); //is dependent on o4 (because of reference object)
		assertSame(obj.elementAt(3), o2);
		assertSame(obj.elementAt(4), o1);
		
		o6.addDependency(o5);
		o6.addDependency(o3);
		
		//add o6
		assertTrue(model.addSimObject(o6));
		model.updateContent();
		
		obj = model.getSimObjects();
		assertEquals(obj.size(), 6);
		assertSame(obj.elementAt(0), o3);
		assertSame(obj.elementAt(1), o4);
		assertSame(obj.elementAt(2), o5);
		assertSame(obj.elementAt(3), o6);
		assertSame(obj.elementAt(4), o2);
		assertSame(obj.elementAt(5), o1);
		
		o7.addDependency(o2);
		o7.setReferenceObject(o3);
		o7.addDependency(o4);
		o7.addDependency(o5);
		o7.addDependency(o6);
	
		//add o7
		assertTrue(model.addSimObject(o7));
		model.updateContent();
		
		obj = model.getSimObjects();
		assertEquals(obj.size(), 7);
		assertSame(obj.elementAt(0), o3);
		assertSame(obj.elementAt(1), o4);
		assertSame(obj.elementAt(2), o5);
		assertSame(obj.elementAt(3), o6);
		assertSame(obj.elementAt(4), o2);
		assertSame(obj.elementAt(5), o7);
		assertSame(obj.elementAt(6), o1);
	}
	
	
	/**
	 * There might be cycles in the dependencies of SimObjects, like 
	 * o1 is dependent on 02, 02 is dependent on 01. Such cycles have to 
	 * be prevented by the user to get a stable, working system. 
	 * Nevertheless a violation of this rule should not lead to an infinite loop.
	 * In such a situation the user should be warned with a logger message, 
	 * but the simulator should go on processing. Because the dependencies might
	 * be wrong, the behavior of the simulator might be wrong!
	 *
	 */
	public void testDependencyCycleDetection() {
		SimObject o1 = new SimObject();
		SimObject o2 = new SimObject();
		SimObject o3 = new SimObject();
		
		o1.addDependency(o2);
		o2.addDependency(o3);
		o3.setReferenceObject(o1);
		
		EnvironmentModel model = new TestEnvironment();
		assertEquals(model.getSimObjects().size(), 0);
		
		assertTrue(model.addSimObject(o1));
		model.updateContent();
		assertEquals(model.getSimObjects().size(), 1);
		
		//can not add the same object twice.
		assertFalse(model.addSimObject(o1));
		model.updateContent();
		assertEquals(model.getSimObjects().size(), 1);
		
		assertTrue(model.addSimObject(o2));
		model.updateContent();
		assertEquals(model.getSimObjects().size(), 2);
		
		//To see the warning message, comment out the next line.
		Logger.getLogger("Simulator").setLevel(Level.OFF);
		//here a cycle should be detected.
		assertTrue(model.addSimObject(o3));
		model.updateContent();
		Vector obj = model.getSimObjects();
		assertEquals(obj.size(), 3);
		
	}
	
	public void testInit() {

		MyTestEnvironment env = new MyTestEnvironment();
		SimObject obj1 = new SimObject();
		obj1.setLayer(5);
		SimObject obj2 = new SimObject();
		obj2.setLayer(10);
		env.mObj = obj2;
		obj1.addDependency(obj2);
		
		assertFalse(env.isInitialized());
		assertEquals(env.getSimObjects().size(), 0);
		
		assertTrue(env.addSimObject(obj1));
		assertEquals(env.getSimObjects().size(), 0); //SimObjects only in queue.
		
		assertEquals(env.mCreateEnvironmentCounter, 0);
		
		env.initEnvironment();
		//Both objects are integrated into the system, layers and dependencies matching.
		assertTrue(env.isInitialized());
		assertEquals(env.getSimObjects().size(), 2);
		assertSame(env.getSimObjects().elementAt(0), obj2);
		assertSame(env.getSimObjects().elementAt(1), obj1);
		assertEquals(env.getSimObjectsSortedByLayers().size(), 2);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(0), obj1);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(1), obj2);
		assertEquals(env.mCreateEnvironmentCounter, 1);

	}
	
	
	public void testCreateEnvironments() { //tests implicitly method createEnvironmentVariation()
		MyTestEnvironment env = new MyTestEnvironment();
		SimObject obj1 = new SimObject();
		obj1.setLayer(5);
		SimObject obj2 = new SimObject();
		obj2.setLayer(10);
		env.mObj = obj2;
		obj1.addDependency(obj2);
		
		assertFalse(env.isInitialized());
		assertEquals(env.getSimObjects().size(), 0);
		
		assertTrue(env.addSimObject(obj1));
		assertEquals(env.getSimObjects().size(), 0); //SimObjects only in queue.
		
		assertEquals(env.mCreateEnvironmentCounter, 0);
		
		Vector environments = env.createEnvironments(3);
		assertTrue(env.isInitialized()); //initialization was done implicit
		assertEquals(env.getSimObjects().size(), 2);
		assertEquals(env.mCreateEnvironmentCounter, 1);
		assertEquals(env.mUpdateRandomizationCounter, 0); //this environment is the master environment and not randomized.
		assertEquals(env.mResetEnvironmentObjectsCounter, 0); //this environment is the master environment and not reseted.
		
		for(int i = 0; i < environments.size(); ++i) {
			MyTestEnvironment mte = (MyTestEnvironment) environments.elementAt(i);
			assertTrue(mte.isInitialized());
			assertEquals(mte.getSimObjects().size(), 2);
			assertEquals(mte.mCreateEnvironmentCounter, 1); //was cloned, but method createEnv. was not executed.
			assertEquals(mte.mUpdateRandomizationCounter, 1);
			assertEquals(mte.mResetEnvironmentObjectsCounter, 0); //not called, only called when the environment is used for a simulation duing setup.
			assertNotSame(mte, env); //All variation environments are NOT the master environment itself but copies.
			for(int j = 0; j < environments.size(); ++j) {
				if(j == i) {
					continue;
				}
				assertNotSame(environments.elementAt(j), mte); //all EnvironmentModel objects are separate instances.
			}
		}

		Vector environments2 = env.createEnvironments(3);
		assertTrue(env.isInitialized());
		assertEquals(env.mCreateEnvironmentCounter, 1); //not initialized a second time.
		
		for(int i = 0; i < environments2.size(); ++i) {
			assertFalse(environments.contains(environments2.elementAt(i))); //the new environments are separate, new objects.
		}
		
	}
	
	
	public void testClone() {
		TestEnvironment env = new TestEnvironment();
		SimObject obj1 = new SimObject();
		SimObject obj2 = new SimObject();
		obj2.setLayer(-2);
		SimObject obj3 = new SimObject();
		obj3.setLayer(5);
		
		obj1.addDependency(obj3);
		obj2.addDependency(obj1);
		
		env.addSimObject(obj1);
		env.addSimObject(obj2);
		env.addSimObject(obj3);
		
		assertFalse(env.isInitialized());
		env.initEnvironment();
		assertTrue(env.isInitialized());
		
		assertEquals(env.getSimObjects().size(), 3);
		assertEquals(env.getSimObjects().elementAt(0), obj3);
		assertEquals(env.getSimObjects().elementAt(1), obj1);
		assertEquals(env.getSimObjects().elementAt(2), obj2);
		
		assertEquals(env.getSimObjectsSortedByLayers().size(), 3);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(0), obj2);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(1), obj1);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(2), obj3);
		
		//start tests
		TestEnvironment clone = null;
		try {
			 clone = (TestEnvironment) env.cloneEnvironment();
		}
		catch(CloneNotSupportedException e) {
			fail("CloneNotSupportedException");
		}
		assertNotNull(clone);
		assertNotSame(clone, env);
		
		assertTrue(clone.isInitialized());
		
		assertEquals(clone.getSimObjects().size(), 3);
		assertEquals(clone.getSimObjects().elementAt(0), obj3);
		assertEquals(clone.getSimObjects().elementAt(1), obj1);
		assertEquals(clone.getSimObjects().elementAt(2), obj2);
		
		assertEquals(clone.getSimObjectsSortedByLayers().size(), 3);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(0), obj2);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(1), obj1);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(2), obj3);
		
		//internal data structures are independent now.
		SimObject obj4 = new SimObject();
		assertTrue(clone.addSimObject(obj4));
		
		env.updateContent();
		assertEquals(env.getSimObjects().size(), 3);
		assertEquals(clone.getSimObjects().size(), 3);
		
		clone.updateContent();
		assertEquals(env.getSimObjects().size(), 3);
		assertEquals(clone.getSimObjects().size(), 4);
		
		//TODO test remove when supported.
		
		obj1.setLayer(-10);
		obj1.removeDependency(obj3);
		obj3.addDependency(obj2);
		obj4.addDependency(obj3);
		obj4.setLayer(1);
		
		clone.resolveDependencies();
		
		assertEquals(env.getSimObjects().size(), 3);
		assertEquals(env.getSimObjects().elementAt(0), obj3);
		assertEquals(env.getSimObjects().elementAt(1), obj1);
		assertEquals(env.getSimObjects().elementAt(2), obj2);
		
		assertEquals(env.getSimObjectsSortedByLayers().size(), 3);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(0), obj1);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(1), obj2);
		assertSame(env.getSimObjectsSortedByLayers().elementAt(2), obj3);
		
		assertEquals(clone.getSimObjects().size(), 4);
		assertEquals(clone.getSimObjects().elementAt(0), obj1);
		assertEquals(clone.getSimObjects().elementAt(1), obj2);
		assertEquals(clone.getSimObjects().elementAt(2), obj3);
		assertEquals(clone.getSimObjects().elementAt(3), obj4);
		
		assertEquals(clone.getSimObjectsSortedByLayers().size(), 4);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(0), obj1);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(1), obj2);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(2), obj4);
		assertSame(clone.getSimObjectsSortedByLayers().elementAt(3), obj3);
		
		
	}
	
	
	
	
	private class MyTestEnvironment extends TestEnvironment {
		
		public SimObject mObj;

		public void createEnvironment() {
			mCreateEnvironmentCounter++;
			assertEquals(getSimObjects().size(), 1); //the other object is already integrated and can be used  in this method.
			assertTrue(addSimObject(mObj));
		}

	};
	
	public class TestEnvironment extends EnvironmentModel {

		public int mUpdateRandomizationCounter;
		public int mCreateEnvironmentCounter;
		public int mResetEnvironmentObjectsCounter;
		public MasterCounter mMasterCounter = new MasterCounter();
		

		public void createEnvironment() {
			mCreateEnvironmentCounter++;
		}

		public void updateRandomizationParameters() {
			mUpdateRandomizationCounter++;
		}

		public void resetEnvironmentObjects() {
			mResetEnvironmentObjectsCounter = ++mMasterCounter.mCounter;
		}
		
		public Object cloneEnvironment() throws CloneNotSupportedException {
			return this.clone();
		}
		
	}
	
	public class MasterCounter {
		public int mCounter;
	}
	
	public class MyTestSimObject extends SimObject {
		
		public SimObject mObjectToAdd;
		public EnvironmentModel mModel;
		
		public int mResetCounter;
		public MasterCounter mMasterCounter = new MasterCounter();

		public void reset() {
			super.reset();
			mResetCounter = ++mMasterCounter.mCounter;
		}
		
		public boolean update() {
			super.update();
			if(mObjectToAdd != null && mModel != null) {
				mModel.addSimObject(mObjectToAdd);
			}
			return true;
		}
		
	}

}
