package de.fhg.iais.outdoor.simba.test;

import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunction;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunctionRamp;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunctionSigmoid;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunctionTanh;
import junit.framework.TestCase;

public class TestNeuralNet extends TestCase {

	
	protected void setUp() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	public void testConstructor() {
		NeuralNet net = new NeuralNet();
		assertNotNull(net.getDefaultTransferFunction());
		assertTrue(net.getDefaultTransferFunction() instanceof TransferFunctionTanh);
		assertEquals(net.getNeurons().size(), 0);
		
		TransferFunctionSigmoid tfs = new TransferFunctionSigmoid();
		net.setDefaultTransferFunction(tfs);
		assertSame(tfs, net.getDefaultTransferFunction());
		
		assertEquals(net.getSynapses().size(), 0);
	}
	
	public void addRemoveNeurons() {
		NeuralNet net = new NeuralNet();
		
		Neuron n1 = new Neuron("N1");
		Neuron n2 = new Neuron("N2");
		Neuron n3 = new Neuron("N3");
		
		Synapse s1 = new Synapse(n1);
		Synapse s2 = new Synapse(n1);
		Synapse s3 = new Synapse(n2);
		Synapse s4 = new Synapse(n3);
		
		n1.addSynapse(s1);
		n1.addSynapse(s3);
		n2.addSynapse(s4);
		n3.addSynapse(s2);
		
		assertEquals(net.getSynapses().size(), 0);
		assertEquals(net.getNeurons().size(), 0);
		
		assertNull(n1.getTransferFunction());
		assertNull(n2.getTransferFunction());
		assertNull(n3.getTransferFunction());
		
		TransferFunctionSigmoid tfs = new TransferFunctionSigmoid();
		net.setDefaultTransferFunction(tfs);
		assertSame(tfs, net.getDefaultTransferFunction());
		
		//add neurons
		assertFalse(net.addNeuron(null));
		assertTrue(net.addNeuron(n1));
		assertEquals(net.getNeurons().size(), 1);
		assertFalse(net.addNeuron(n1));
		assertEquals(net.getNeurons().size(), 1);
		assertEquals(net.getSynapses().size(), 2);
		
		assertTrue(net.addNeuron(n2));
		assertEquals(net.getNeurons().size(), 2);
		assertEquals(net.getSynapses().size(), 3);
		assertTrue(net.addNeuron(n3));
		assertEquals(net.getNeurons().size(), 3);
		assertEquals(net.getSynapses().size(), 4);
		
		assertEquals(n1.getTransferFunction(), tfs);
		assertEquals(n2.getTransferFunction(), tfs);
		assertEquals(n3.getTransferFunction(), tfs);
		
		net.validateNetworkConnections(); //no change
		assertEquals(net.getNeurons().size(), 3);
		assertEquals(net.getSynapses().size(), 4);
		
		//remove neurons
		assertFalse(net.removeNeuron(null));
		assertTrue(net.removeNeuron(n1));
		assertEquals(net.getNeurons().size(), 2);
		assertEquals(net.getSynapses().size(), 2);
		assertFalse(net.removeNeuron(n1)); //already removed
		assertEquals(net.getNeurons().size(), 2);
		assertEquals(net.getSynapses().size(), 2);
		
		net.validateNetworkConnections(); //removes one invalid synapse (n1 -> n3)
		assertEquals(net.getNeurons().size(), 2);
		assertEquals(net.getSynapses().size(), 1);
		assertSame(((Synapse) net.getSynapses().elementAt(0)).getSourceNeuron(), n3); //(n3 -> n2);
		
		assertTrue(net.removeNeuron(n3));
		assertEquals(net.getNeurons().size(), 1);
		assertEquals(net.getSynapses().size(), 1);
		
		net.validateNetworkConnections(); //removes last synapse (n3 -> n2)
		assertEquals(net.getNeurons().size(), 1);
		assertEquals(net.getSynapses().size(), 0);
		
		assertTrue(net.removeNeuron(n2));
		assertEquals(net.getNeurons().size(), 0);
		
		assertSame(n1.getTransferFunction(), tfs); //did not change during remove
		
		
	}
	
	
	public void testMinimalNetwork() {
		
		NeuralNetAgentInterface ai = new NeuralNetAgentInterface("AI");
		NeuronWrapper nw1 = new NeuronWrapper("NW1", 0.0, 0.0, 100.0);
		NeuronWrapper nw2 = new NeuronWrapper("NW2", 0.0, 0.0, 50.0);
		NeuronWrapper nw3 = new NeuronWrapper("NW3", 0.0, 0.0, 100.0);
		NeuronWrapper nw4 = new NeuronWrapper("NW4", 0.0, 0.0, 50.0);
		NeuronWrapper nw5 = new NeuronWrapper("NW5", 0.0, 0.0, 100.0);
		
		ai.addInputNeuron(nw1);
		ai.addInputNeuron(nw2);
		ai.addOutputNeuron(nw3);
		ai.addOutputNeuron(nw4);
		ai.addInfoNeuron(nw5);
		
		NeuralNet net = NeuralNet.createMinimalNet(null);
		assertNull(null);
		
		NeuralNetAgentInterface ai2 = new NeuralNetAgentInterface("AIEmpty");
		net = NeuralNet.createMinimalNet(ai2);
		assertNotNull(net);
		assertEquals(net.getNeurons().size(), 0);
		assertEquals(net.getSynapses().size(), 0);
		
		
		net = NeuralNet.createMinimalNet(ai);
		assertNotNull(net);
		assertEquals(net.getNeurons().size(), 4);
		assertEquals(net.getSynapses().size(), 0);
		assertTrue(ai.getInputNeurons().contains(((Neuron) net.getNeurons().elementAt(0)).getActivationValue()));
		assertTrue(ai.getInputNeurons().contains(((Neuron) net.getNeurons().elementAt(1)).getActivationValue()));
		assertTrue(ai.getOutputNeurons().contains(((Neuron) net.getNeurons().elementAt(2)).getActivationValue()));
		assertTrue(ai.getOutputNeurons().contains(((Neuron) net.getNeurons().elementAt(3)).getActivationValue()));
		
		//info-neurons are ignored
		
	}
	
	public void testUpdate() {
		NeuralNet net = new NeuralNet();
		net.setDefaultTransferFunction(new TransferFunctionRamp()); //easier to verify
		
		AdapterNeuron n1 = new AdapterNeuron("N1");
		AdapterNeuron n2 = new AdapterNeuron("N2");
		AdapterNeuron n3 = new AdapterNeuron("N3");
		
		NeuronWrapper nw1 = new NeuronWrapper("", 0.0, 0.0, 100.0);
		NeuronWrapper nw2 = new NeuronWrapper("", 0.0, 0.0, 50.0);
		NeuronWrapper nw3 = new NeuronWrapper("", 0.0, 0.0, 10.0);
		
		Synapse s1 = new Synapse(n1);
		Synapse s2 = new Synapse(n1);
		Synapse s3 = new Synapse(n2);
		Synapse s4 = new Synapse(n3);
		
		n1.addSynapse(s1);
		n1.addSynapse(s3);
		n2.addSynapse(s4);
		n3.addSynapse(s2);
		
		n1.setActivationValue(nw1);
		n2.setActivationValue(nw2);
		n3.setActivationValue(nw3);
		
		net.addNeuron(n1);
		net.addNeuron(n2);
		net.addNeuron(n3);
		
		assertEquals(net.getNeurons().size(), 3);
		assertEquals(net.getSynapses().size(), 4);
		
		assertEquals(n1.mCountReset, 0);
		assertEquals(n1.mCountUpdateActivation, 0);
		assertEquals(n3.mCountReset, 0);
		assertEquals(n3.mCountUpdateActivation, 0);
		
		net.reset();
		assertEquals(n1.mCountReset, 1);
		assertEquals(n1.mCountUpdateActivation, 0);
		assertEquals(n3.mCountReset, 1);
		assertEquals(n3.mCountUpdateActivation, 0);
		
		assertEquals(n1.getActivation(), 0.0, 0.001);
		assertEquals(n2.getActivation(), 0.0, 0.001);
		assertEquals(n3.getActivation(), 0.0, 0.001);
		
		net.updateNetwork();
		assertEquals(n1.mCountReset, 1);
		assertEquals(n1.mCountUpdateActivation, 1);
		assertEquals(n3.mCountReset, 1);
		assertEquals(n3.mCountUpdateActivation, 1);
		
		//Strength of all synapses was 0
		assertEquals(n1.getActivation(), 0.0, 0.001);
		assertEquals(n2.getActivation(), 0.0, 0.001);
		assertEquals(n3.getActivation(), 0.0, 0.001);
		
		nw1.setNormalized(0.5);
		nw2.setNormalized(1.0);
		nw3.setNormalized(0.25);
		assertEquals(nw1.getNormalized(), 0.5, 0.001);
		
		s1.setStrength(0.5);
		s2.setStrength(1.0);
		s3.setStrength(-0.1);
		s4.setStrength(-0.75);
		
		net.updateNetwork();
		assertEquals(n1.mCountUpdateActivation, 2);
		
		assertEquals(n1.getActivation(), 0.15, 0.001);
		assertEquals(n2.getActivation(), 0.0, 0.001); //below 0.0
		assertEquals(n3.getActivation(), 0.5, 0.001);
		
		s1.setStrength(2.0);
		s2.setStrength(100.0);
		s3.setStrength(1.0);
		s4.setStrength(0.3);
		
		net.updateNetwork();
		assertEquals(n1.mCountUpdateActivation, 3);
		
		assertEquals(n1.getActivation(), 0.3, 0.001);
		assertEquals(n2.getActivation(), 0.3 * 0.5, 0.001); 
		assertEquals(n3.getActivation(), 1.0, 0.001); //more than 1.0 (capped by transfer function)
		
		
	}
	
	public void testClone() {
		
		NeuralNet net = new NeuralNet();
		TransferFunction tf = new TransferFunctionRamp();
		net.setDefaultTransferFunction(tf); 
		
		AdapterNeuron n1 = new AdapterNeuron("N1");
		AdapterNeuron n2 = new AdapterNeuron("N2");
		AdapterNeuron n3 = new AdapterNeuron("N3");
		
		NeuronWrapper nw1 = new NeuronWrapper("", 0.0, 0.0, 100.0);
		NeuronWrapper nw2 = new NeuronWrapper("", 0.0, 0.0, 50.0);
		NeuronWrapper nw3 = new NeuronWrapper("", 0.0, 0.0, 10.0);
		
		Synapse s1 = new Synapse(n1);
		Synapse s2 = new Synapse(n1);
		Synapse s3 = new Synapse(n2);
		Synapse s4 = new Synapse(n3);
		
		n1.addSynapse(s1);
		n1.addSynapse(s3);
		n2.addSynapse(s4);
		n3.addSynapse(s2);
		
		n1.setActivationValue(nw1);
		n2.setActivationValue(nw2);
		n3.setActivationValue(nw3);
		
		net.addNeuron(n1);
		net.addNeuron(n2);
		net.addNeuron(n3);
		
		assertEquals(net.getNeurons().size(), 3);
		assertEquals(net.getSynapses().size(), 4);
		assertSame(n1.getTransferFunction(), tf);
		
		NeuralNet clone = null;
		try {
			clone = (NeuralNet) net.clone();
		}
		catch(CloneNotSupportedException e) {
			fail("Clone not supported");
		}
		
		assertNotSame(clone, net);
		assertNotNull(clone);
		assertSame(clone.getDefaultTransferFunction(), net.getDefaultTransferFunction());
		assertSame(clone.getDefaultTransferFunction(), tf);
		
		assertEquals(clone.getNeurons().size(), 3);
		assertEquals(clone.getSynapses().size(), 4);
		Vector neurons = net.getNeurons();
		Vector synapses = net.getSynapses();
		Vector clonedNeurons = clone.getNeurons();
		Vector clonedSynapses = clone.getSynapses();
		
		for(Iterator i = neurons.iterator(); i.hasNext();) {
			if(clonedNeurons.contains(i.next())) {
				fail("Neuron is not allowed to be contained in both networks!");
			}
		}
		for(Iterator i = synapses.iterator(); i.hasNext();) {
			Synapse s = (Synapse) i.next();
			if(clonedSynapses.contains(s)) {
				fail("Synapse is not allowed to be contained in both networks!");
			}
			if(clonedNeurons.contains(s.getSourceNeuron())) {
				fail("Synapses of the one network are not allowed to have a source in the other");
			}
			if(s.getSourceNeuron() == null || !neurons.contains(s.getSourceNeuron())) {
				fail("Synapses have to have a source in the own network");
			}
		}
		for(Iterator i = clonedNeurons.iterator(); i.hasNext();) {
			Neuron n = (Neuron) i.next();
			if(neurons.contains(n)) {
				fail("Neuron is not allowed to be contained in both networks!");
			}
		}
		for(Iterator i = clonedSynapses.iterator(); i.hasNext();) {
			Synapse s = (Synapse) i.next();
			if(synapses.contains(s)) {
				fail("Synapse is not allowed to be contained in both networks!");
			}
			if(neurons.contains(s.getSourceNeuron())) {
				fail("Synapses of the one network are not allowed to have a source in the other");
			}
			if(s.getSourceNeuron() == null || !clonedNeurons.contains(s.getSourceNeuron())) {
				fail("Synapses have to have a source in the own network");
			}
		}

	}
	
	
	public class AdapterNeuron extends Neuron {

		public int mCountClone;
		public int mCountReset;
		public int mCountUpdateActivation;
		
		public AdapterNeuron(String name) {
			super(name);
		}
		
		public Object clone() throws CloneNotSupportedException {
			mCountClone++;
			return super.clone();
		}

		public void reset() {
			mCountReset++;
			super.reset();
		}

		public void updateActivation() {
			mCountUpdateActivation++;
			super.updateActivation();
		}
		
		
		
	}
	
	
	
	
	
}
