package de.fhg.iais.outdoor.simba.test;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunction;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunctionSigmoid;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunctionTanh;
import junit.framework.TestCase;

public class TestNeuronAndSynapse extends TestCase {

	protected void setUp() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	//TODO test bias
	
	public void testNeuronConstructors() {
		Neuron n1 = new Neuron("Neuron1");
		
		TransferFunctionTanh tanh = new TransferFunctionTanh();
		
		Neuron n2 = new Neuron("Neuron2", tanh);
		
		assertEquals(n1.getName(), "Neuron1");
		assertNull(n1.getTransferFunction());
		
		assertSame(n2.getTransferFunction(), tanh);

		assertEquals(n1.getActivation(), 0.0, 0.0);
		assertNotNull(n1.getActivationValue());
		assertEquals(n1.getSynapses().size(), 0);
		
		try {
			n1.updateActivation();
			fail("Activation did not throw an exception although the transfer functin is null.");
		}
		catch(IllegalStateException e) {	
		}
		
		n1.setTransferFunction(new TransferFunctionTanh());
		
		n1.setName("AnotherName");
		assertEquals(n1.getName(), "AnotherName");
		
		NeuronWrapper wrapper = new NeuronWrapper("MyNW", 1.0, 0.0, 1.0);
		wrapper.setNormalizedValueRange(10.0, 20.0); //will be overridden during addActivationValue.
		n1.setActivationValue(wrapper);
		assertSame(n1.getActivationValue(), wrapper);
		n1.setActivationValue(null);
		assertSame(n1.getActivationValue(), wrapper);
		
		assertEquals(n1.getActivation(), 1.0, 0.001);
		wrapper.set(0.0);
		assertEquals(n1.getActivation(), -1.0, 0.001);
		wrapper.set(0.5);
		assertEquals(n1.getActivation(), 0.0, 0.001);
		
		TransferFunctionSigmoid sig = new TransferFunctionSigmoid();
		n1.setTransferFunction(sig);
		assertSame(n1.getTransferFunction(), sig);
		n1.setTransferFunction(null);
		assertSame(n1.getTransferFunction(), sig); // was not replaced.
		
		assertEquals(n1.getActivation(), 0.5, 0.00001); //range of wrapper is not 0.0 - 1.0 (sigmoid)
		
		n1.reset();
		assertEquals(n1.getActivation(), 0.0, 0.0001); //activation reset to 0.0;

	}
	
	public void testCloneNeuron() {
		Neuron n1 = new Neuron("N1");
		Neuron n2 = new Neuron("N2");
		Neuron n3 = new Neuron("N3");
		
		Synapse s1 = new Synapse(n1);
		Synapse s2 = new Synapse(n2);
		Synapse s3 = new Synapse(n3);
		
		n1.addSynapse(s1);
		n1.addSynapse(s3);
		n1.addSynapse(s2);
		
		assertSame(s1.getTargetNeuron(), n1);
		assertSame(s2.getTargetNeuron(), n1);
		assertSame(s3.getTargetNeuron(), n1);
		
		assertEquals(n1.getSynapses().size(), 3);
		n1.setType("HelloWorldType");
		n1.setProtectionMode("HelloWorldMode");
		
		Neuron clone = null;
		try {
			clone = (Neuron) n1.clone();
		}
		catch(CloneNotSupportedException e) {
			fail("Clone not supported");
		}
		assertNotNull(clone);
		assertNotSame(clone, n1);
		assertSame(n1.getActivationValue(), clone.getActivationValue()); //use the same activation value
		assertNotNull(clone.getActivationValue());
		assertNotSame(n1.getBiasValue(), clone.getBiasValue());
		assertNotNull(clone.getBiasValue());
		assertEquals(n1.getName(), clone.getName());
		assertNotSame(n1.getNameValue(), clone.getNameValue());
		assertNotNull(clone.getNameValue());
		assertSame(n1.getTransferFunction(), clone.getTransferFunction());
		assertNotSame(n1.getHistory(), clone.getHistory());
		assertEquals(n1.getType(), clone.getType());
		assertEquals(clone.getType(), "HelloWorldType");
		assertEquals(n1.getProtectionMode(), clone.getProtectionMode());
		assertEquals(clone.getProtectionMode(), "HelloWorldMode");
		assertEquals(n1.getId(), clone.getId());
		
		assertEquals(clone.getSynapses().size(), 3);
		assertFalse(clone.getSynapses().contains(s1));
		assertFalse(clone.getSynapses().contains(s2));
		assertFalse(clone.getSynapses().contains(s3));
		
		assertSame(((Synapse) clone.getSynapses().elementAt(0)).getSourceNeuron(), n1);
		assertSame(((Synapse) clone.getSynapses().elementAt(1)).getSourceNeuron(), n3);
		assertSame(((Synapse) clone.getSynapses().elementAt(2)).getSourceNeuron(), n2);
		
		assertSame(((Synapse) clone.getSynapses().elementAt(0)).getTargetNeuron(), clone);
		assertSame(((Synapse) clone.getSynapses().elementAt(1)).getTargetNeuron(), clone);
		assertSame(((Synapse) clone.getSynapses().elementAt(2)).getTargetNeuron(), clone);
	}
	
	public void testAddRemoveSynapses() {
		Neuron n1 = new Neuron("N1");
		Neuron n2 = new Neuron("N2");
		Neuron n3 = new Neuron("N3");
		
		DummyTransferFunction tf = new DummyTransferFunction();
		n1.setTransferFunction(tf);
		
		AdapterSynapse as1 = new AdapterSynapse(n2);
		AdapterSynapse as2 = new AdapterSynapse(n2); //same source!
		AdapterSynapse as3 = new AdapterSynapse(n3);
		AdapterSynapse as4 = new AdapterSynapse(n1); //self loop N1 -> N1
		AdapterSynapse as5 = new AdapterSynapse(null); //no source neuron.
		
		assertSame(n1.getTransferFunction(), tf);
	
		assertEquals(n1.getSynapses().size(), 0);
		assertFalse(n1.addSynapse(null));
		assertEquals(n1.getSynapses().size(), 0);
		assertNull(as1.getTargetNeuron());
		assertTrue(n1.addSynapse(as1));
		assertEquals(n1.getSynapses().size(), 1);
		assertSame(n1.getSynapses().elementAt(0), as1);
		assertSame(as1.getTargetNeuron(), n1);
		assertFalse(n1.addSynapse(as1)); //can not add the synapse a second time.
		assertEquals(n1.getSynapses().size(), 1);
		
		assertFalse(n1.addSynapse(as2)); //because the synapses source is already available.
		assertEquals(n1.getSynapses().size(), 1);
		
		assertTrue(n1.addSynapse(as3)); //comming from a different synapse.
		assertEquals(n1.getSynapses().size(), 2);
		
		assertTrue(n1.addSynapse(as4)); //comming from N1 itself (self coupling)
		assertEquals(n1.getSynapses().size(), 3);
		assertTrue(n1.getSynapses().contains(as1));
		assertTrue(n1.getSynapses().contains(as3));
		assertTrue(n1.getSynapses().contains(as4));
		
		//remove
		assertFalse(n1.removeSynapse(null));
		assertEquals(n1.getSynapses().size(), 3);
		assertSame(as3.getTargetNeuron(), n1);
		assertTrue(n1.removeSynapse(as3));
		assertEquals(n1.getSynapses().size(), 2);
		assertNull(as3.getTargetNeuron());
		assertTrue(n1.getSynapses().contains(as1));
		assertTrue(n1.getSynapses().contains(as4));
		
		assertFalse(n1.removeSynapse(as3)); //not contained any more
		assertEquals(n1.getSynapses().size(), 2);
		
		assertFalse(n1.addSynapse(as5)); //because it has no source neuron.
		assertEquals(n1.getSynapses().size(), 2);
		
	}
	
	public void testActivationCalculation() {
		Neuron n1 = new Neuron("N1");
		Neuron n2 = new Neuron("N2", new TransferFunctionTanh());
		Neuron n3 = new Neuron("N3", new TransferFunctionTanh());
		
		DummyTransferFunction tf = new DummyTransferFunction();
		n1.setTransferFunction(tf);
		
		AdapterSynapse as1 = new AdapterSynapse(n2);
		AdapterSynapse as3 = new AdapterSynapse(n3);
		AdapterSynapse as4 = new AdapterSynapse(n1); //self loop N1 -> N1
		NeuronWrapper av2 = new NeuronWrapper("AV2", 0.0, 0.0, 0.0);
		NeuronWrapper av3 = new NeuronWrapper("AV3", 0.0, 0.0, 0.0);
		n2.setActivationValue(av2);
		n3.setActivationValue(av3);
		n2.storeLastActivity();
		
		assertSame(n1.getTransferFunction(), tf);
		assertSame(n2.getActivationValue(), av2);
		assertSame(n3.getActivationValue(), av3);
		
		
		//Calculation without synapses
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		assertEquals(tf.mCountTransfers, 0);
		n1.storeLastActivity();
		n1.updateActivation();
		assertEquals(tf.mCountTransfers, 1);
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		
		//add one synapse (from N2)
		assertTrue(n1.addSynapse(as1));
		as1.setStrength(0.5);
		av2.setNormalized(0.5);
		n2.storeLastActivity();
		assertEquals(n2.getActivation(), 0.5, 0.0001);
		
		assertEquals(as1.mCountGetActivation, 0);
		assertEquals(tf.mCountTransfers, 1);
		
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		n1.storeLastActivity();
		assertEquals(as1.mCountGetActivation, 0);
		assertEquals(tf.mCountTransfers, 1);
		n1.updateActivation();
		assertEquals(as1.mCountGetActivation, 1);
		assertEquals(tf.mCountTransfers, 2);
		assertEquals(n1.getActivation(), 0.25, 0.0001);
		
		as1.setStrength(8.0);
		av2.setNormalized(-0.5);
		n1.storeLastActivity();
		n2.storeLastActivity();
		assertEquals(as1.mCountGetActivation, 1);
		n1.updateActivation();
		assertEquals(as1.mCountGetActivation, 2);
		assertEquals(tf.mCountTransfers, 3);
		assertEquals(n1.getActivation(), -1.0, 0.0001); //bounded by NeuronWrapper
		
		
		//Another synapse
		assertTrue(n1.addSynapse(as3));
		as3.setStrength(2.0);
		av3.setNormalized(0.1);
		n1.reset();
		assertEquals(n3.getActivation(), 0.1, 0.0001);
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		
		//calculate
		as1.mCountGetActivation = 0;
		tf.mCountTransfers = 0;
		av2.setNormalized(0.5);
		av3.setNormalized(-0.75);
		as1.setStrength(5.0);
		as3.setStrength(4.0);
		assertEquals(as1.mCountGetActivation, 0);
		assertEquals(as3.mCountGetActivation, 0);
		assertEquals(tf.mCountTransfers, 0);
		
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		n1.storeLastActivity();
		n2.storeLastActivity();
		n3.storeLastActivity();
		n1.updateActivation();
		assertEquals(as1.mCountGetActivation, 1);
		assertEquals(as3.mCountGetActivation, 1);
		assertEquals(tf.mCountTransfers, 1);
		assertEquals(n1.getActivation(), -0.5, 0.0001);
		
		//self coupling
		assertTrue(n1.addSynapse(as4));
		n1.reset();
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		
		//calculate
		as1.mCountGetActivation = 0;
		as3.mCountGetActivation = 0;
		tf.mCountTransfers = 0;
		av2.setNormalized(0.5);
		av3.setNormalized(-0.75);
		as1.setStrength(5.0);
		as3.setStrength(4.0);
		as4.setStrength(-2.5);
		
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		n1.storeLastActivity();
		n1.updateActivation();
		assertEquals(n1.getActivation(), -0.5, 0.0001);
		
		n1.storeLastActivity();
		n1.updateActivation();
		assertEquals(n1.getActivation(), 0.75, 0.0001); //-0.5 + (-2.5 * -0.5)
		
		//disabled synapse
		n1.reset();
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		
		as1.mCountGetActivation = 0;
		as3.mCountGetActivation = 0;
		tf.mCountTransfers = 0;
		av2.setNormalized(0.5);
		av3.setNormalized(-0.75);
		as1.setStrength(5.0);
		as3.setStrength(4.0);
		as4.setStrength(-2.5);
		as1.getEnabledValue().set(false);
		as4.getEnabledValue().set(false); //only as3 is active
		
		assertEquals(n1.getActivation(), 0.0, 0.0001);
		n1.storeLastActivity();
		n1.updateActivation();
		assertEquals(n1.getActivation(), -1.0, 0.0001);
		
	}
	
	
	
	public class DummyTransferFunction extends TransferFunction {
		
		public double mLowerBound;
		public double mUpperBound;
		public int mCountTransfers;
		
		public DummyTransferFunction() {
			super("Dummy");
			mLowerBound = -1.0;
			mUpperBound = 1.0;
		}

		public double getLowerBound() {
			return mLowerBound;
		}

		public double getUpperBound() {
			return mUpperBound;
		}

		public double transferActivation(double activation) {
			mCountTransfers++;
			return activation;
		} 
	}
	
	public class AdapterSynapse extends Synapse {
		
		public int mCountGetActivation;
		
		public AdapterSynapse(Neuron source) {
			super(source);
		}

		public double getActivation() {
			mCountGetActivation++;
			return super.getActivation();
		}
		
		
	}
}
