package de.fhg.iais.outdoor.simba.test;

import java.awt.Polygon;
import java.awt.geom.Point2D;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.OrthogonalDistanceScanner;
import junit.framework.TestCase;

public class TestOrthogonalDistanceScanner extends TestCase {

	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}

	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.devices.OrthogonalDistanceScanner.OrthogonalDistanceScanner(String, double, double, double, double, double, double, double)'
	 */
	public void testOrthogonalDistanceScanner() {
		OrthogonalDistanceScanner obj = new OrthogonalDistanceScanner(
				"Scanner1", 1.0, 2.0, 0.1, 0.5, 6.5);
		
		assertEquals(obj.getName(), "Scanner1");
		assertEquals(obj.getPosition(), new Point2D.Double(1.0, 2.0));
		assertEquals(obj.getAngle(), 0.1, 0.001);
		assertEquals(obj.getMinScanDistance(), 0.5, 0.001);
		assertEquals(obj.getMaxScanDistance(), 6.5, 0.001);
		assertEquals(obj.getObstacles().size(), 0);
		assertEquals(obj.getDependencies().size(), 0);
		assertTrue(obj.isActive());
		assertTrue(obj.isVisible());
	}
	
	public void testGlobalStuffCalculation() {
		
		OrthogonalDistanceScanner obj = new OrthogonalDistanceScanner(
				"Scanner1", 1.0, 2.0, 0.1, 0.5, 6.5);
		
		assertEquals(obj.getGlobalAngle(), 0.1, 0.001);
		assertEquals(obj.getGlobalPosition().x, 1.0, 0.001);
		assertEquals(obj.getGlobalPosition().y, 2.0, 0.001);
		
		obj.setPosition(10.0, 12.0);
		obj.setAngle(43.21);
		
		assertEquals(obj.getGlobalAngle(), 0.1, 0.001);
		assertEquals(obj.getGlobalPosition().x, 1.0, 0.001);
		assertEquals(obj.getGlobalPosition().y, 2.0, 0.001);
		
		obj.update();
		assertEquals(obj.getGlobalAngle(), 43.21, 0.001);
		assertEquals(obj.getGlobalPosition().x, 10.0, 0.001);
		assertEquals(obj.getGlobalPosition().y, 12.0, 0.001);
	}
	
	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.devices.OrthogonalDistanceScanner.update()'
	 */
	public void testUpdate() {
		OrthogonalDistanceScanner scanner = new OrthogonalDistanceScanner(
				"Scanner1", 3.0, 2.0, 0.0, 0.0, 10.0);
		
		SimObject obstacle = new SimObject("Obstacle", 11.0, 2.0, 0.0, new Polygon());
		scanner.addObstacle(obstacle);
		
		assertEquals(scanner.getSensorReading(), 0.0, 0.001);
		
		assertTrue(scanner.update());
		//System.out.println("dist: " + scanner.getSensorReading() + " Raw: " + scanner.getRawDistance());
		assertEquals(scanner.getSensorReading(), 8.0, 0.001);
		
		obstacle.setActive(false); //inactive obstacles are ignored.
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 10.0, 0.001);

		obstacle.setActive(true);
		scanner.setMinScanDistance(2.0);
		scanner.setMaxScanDistance(12.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 8.0, 0.001);
		
		scanner.setMinScanDistance(2.0);
		scanner.setMaxScanDistance(4.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 4.0, 0.001);
		
		scanner.setMinScanDistance(9.0);
		scanner.setMaxScanDistance(10.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 9.0, 0.001);
		
		
		//move the obstacle
		
		obstacle.setPosition(11.0, 4.0); //change only in y axis, so the orthogonal distance is unchanged.
		scanner.setMinScanDistance(0.0);
		scanner.setMaxScanDistance(10.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 8.0, 0.001);
		
		obstacle.setPosition(3.0, 100.0); 
		obstacle.update();
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 0.0, 0.001);
		
		obstacle.setPosition(3.0, -100.0); //no difference for symmetric settings.
		obstacle.update();
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 0.0, 0.001);
		
		obstacle.setPosition(2.9, 100.0); //too far to the left, so not in the scanner range any more.
		obstacle.update();
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 10.0, 0.001);
		
		obstacle.setPosition(7.0, 5.0);
		obstacle.update();
		
		//Change the angle of the scanner
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 4.0, 0.001);
		
		scanner.setAngle(90.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 3.0, 0.001); //now the coordinate system was rotated by 90 degree.
		
		scanner.setAngle(180.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 10.0, 0.001);
		
		scanner.setAngle(270.0);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 10.0, 0.001);
		
		obstacle.setPosition(6.0, 5.0);
		obstacle.update();
		scanner.setAngle(135.0); //the obstacle is exactly on the line through the origin of the scanner.
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 0.0, 0.001);
		
		//some arbitrary values (these values are reference material calculated by the object itself.
		//these values are only as control to detect changes. They are not verified mathematically.
		obstacle.setPosition(8.7, 3.1);
		obstacle.update();
		scanner.setAngle(40.0); 
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 5.0735, 0.001);
		
		obstacle.setPosition(-14.0, -9.0);
		obstacle.update();
		scanner.setAngle(143.5); 
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 7.1225, 0.001);
		
		obstacle.setPosition(0.0, 0.0);
		obstacle.update();
		scanner.setAngle(199.2);
		assertTrue(scanner.update());
		assertEquals(scanner.getSensorReading(), 3.4908, 0.001);
	}

	

}
