package de.fhg.iais.outdoor.simba.test;

import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.GenerationListener;
import de.fhg.iais.outdoor.simba.core.IndividualListener;
import de.fhg.iais.outdoor.simba.core.Scheduler;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.core.StepListener;
import de.fhg.iais.outdoor.simba.core.TryListener;
import de.fhg.iais.outdoor.simba.fitness.FitnessFunction;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import junit.framework.TestCase;

public class TestScheduler extends TestCase {

	protected void setUp() throws Exception {
		super.setUp();
		Simulator.getInstance().clearInstance();
	}

	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}

	public void testScheduler() {
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.removeAllValues();
		
		assertNull(vm.getIntValue("/Simulation/NumberOfSteps"));
		assertNull(vm.getIntValue("/Simulation/CurrentStep"));
		assertNull(vm.getIntValue("/Simulation/NumberOfTries"));
		assertNull(vm.getIntValue("/Simulation/CurrentTry"));
		assertNull(vm.getIntValue("/Simulation/CurrentGeneration"));
		assertNull(vm.getIntValue("/Simulation/CurrentIndividual"));
		assertNull(vm.getBooleanValue("/Simulation/TerminateCurrentTry"));
		
		Scheduler scheduler = new Scheduler(Simulator.getInstance());
		
		assertNotNull(vm.getIntValue("/Simulation/NumberOfSteps"));
		assertNotNull(vm.getIntValue("/Simulation/CurrentStep"));
		assertNotNull(vm.getIntValue("/Simulation/NumberOfTries"));
		assertNotNull(vm.getIntValue("/Simulation/CurrentTry"));
		assertNotNull(vm.getIntValue("/Simulation/CurrentGeneration"));
		assertNotNull(vm.getIntValue("/Simulation/CurrentIndividual"));
		assertNotNull(vm.getBooleanValue("/Simulation/TerminateCurrentTry"));
		
		assertTrue(Simulator.getInstance().getSystemObjects().contains(scheduler));
		
		assertNull(scheduler.getCurrentEnvironment());
		assertNull(scheduler.getMasterEnvironment());
		
		assertEquals(scheduler.getNumberOfSteps(), 0);
		assertEquals(scheduler.getNumberOfTries(), 0);
		assertEquals(scheduler.getCurrentGeneration(), 0);
		assertEquals(scheduler.getCurrentStep(), 0);
		assertEquals(scheduler.getCurrentTry(), 0);
		assertEquals(scheduler.getCurrentIndividual(), 0);
		
		assertFalse(scheduler.shouldCurrentTryBeTerminated());
	}

	public void testSetMasterEnvironment() {
		Scheduler s = Simulator.getInstance().getScheduler();
		
		MasterCounter counter = new MasterCounter();
		AdapterEnvironment env = new AdapterEnvironment();
		env.mMasterCounter = counter;
		
		assertNull(s.getMasterEnvironment());
		assertEquals(env.mInitCounter, 0);
		s.setMasterEnvironment(env);
		//executed init, because this was not done yet.
		assertEquals(env.mInitCounter, 1);
		assertSame(s.getMasterEnvironment(), env);
		assertEquals(env.mUpdateContentCounter, 2);
		assertEquals(env.mResolveDependenciesCounter, 3);
		assertEquals(counter.mCounter, 3);
		
		assertEquals(s.getEnvironmentModels().size(), 0);
		
		//add and remove a single EnvironmentModel
		AdapterEnvironment env2 = new AdapterEnvironment();
		assertFalse(s.addEnvironmentModel(null));
		assertTrue(s.addEnvironmentModel(env2));
		assertEquals(s.getEnvironmentModels().size(), 1);
		assertSame(s.getEnvironmentModels().elementAt(0), env2);
		assertTrue(s.addEnvironmentModel(env2)); //here duplicates ARE ALLOWED!
		assertEquals(s.getEnvironmentModels().size(), 2);
		assertSame(s.getEnvironmentModels().elementAt(0), env2);
		assertSame(s.getEnvironmentModels().elementAt(1), env2);
		
		s.removeAllEnvironmentModels();
		assertEquals(s.getEnvironmentModels().size(), 0);
		
		assertNull(s.getCurrentEnvironment());
		s.setCurrentEnvironment(env2);
		assertSame(s.getCurrentEnvironment(), env2);
		assertEquals(s.getEnvironmentModels().size(), 1);
		assertSame(s.getEnvironmentModels().elementAt(0), env2);
		
		//autogeneration only works if a master environment is set.
		s.setMasterEnvironment(null);
		assertNull(s.getMasterEnvironment());
		s.createEnvironments(2);
		assertEquals(s.getEnvironmentModels().size(), 1);
		assertSame(s.getEnvironmentModels().elementAt(0), env2);
		assertSame(s.getCurrentEnvironment(), env2);
		
		//add 2 autogenerated EnvironmentModels to the scheduler
		s.setMasterEnvironment(env);
		assertSame(s.getMasterEnvironment(), env);
		s.createEnvironments(2);
		assertEquals(s.getEnvironmentModels().size(), 2); //created 2 new EnvironmentModels
		assertFalse(s.getEnvironmentModels().contains(env2)); //env2 (and all other previous environments) was removed.
		assertNotNull(s.getCurrentEnvironment()); //one of the new environments is the current one.
		assertNotSame(s.getCurrentEnvironment(), env2);
		assertNotSame(s.getCurrentEnvironment(), env);
		assertFalse(s.getEnvironmentModels().contains(env));
		
		assertEquals(env.mCreateEnvironmentsCounter, 4);
		assertEquals(env.mResolveDependenciesCounter, 5);
		assertEquals(counter.mCounter, 7); //called updateContent at env for each of the new environments
		assertEquals(env.mUpdateContentCounter, 7);
		
		s.removeAllEnvironmentModels();
		assertNull(s.getCurrentEnvironment());
		assertEquals(s.getEnvironmentModels().size(), 0);
		assertSame(s.getMasterEnvironment(), env);
	}


	public void testExecutePhases() {
		Scheduler s = Simulator.getInstance().getScheduler();
		Simulator.getInstance().setSimulationDelay(0);
		
		AdapterSimulationListener listener = new AdapterSimulationListener();
		MasterCounter counter = new MasterCounter();
		listener.mMasterCounter = counter;
		s.addGenerationListener(listener);
		s.addIndividualListener(listener);
		s.addTryListener(listener);
		s.addStepListener(listener);
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		IntValue numberOfSteps = vm.getIntValue("/Simulation/NumberOfSteps");
		IntValue currentStep = vm.getIntValue("/Simulation/CurrentStep");
		IntValue numberOfTries = vm.getIntValue("/Simulation/NumberOfTries");
		IntValue currentTry = vm.getIntValue("/Simulation/CurrentTry");
		IntValue currentGen = vm.getIntValue("/Simulation/CurrentGeneration");
		IntValue currentInd = vm.getIntValue("/Simulation/CurrentIndividual");
		BooleanValue terminateTry = vm.getBooleanValue("/Simulation/TerminateCurrentTry");
		BooleanValue killTry = new BooleanValue(false);
		
		assertNotNull(numberOfSteps);
		assertNotNull(currentStep);
		assertNotNull(numberOfTries);
		assertNotNull(currentTry);
		assertNotNull(currentGen);
		assertNotNull(currentInd);
		assertNotNull(terminateTry);
		assertNotNull(killTry);
		
		s.addTerminateTryValue(killTry);
		
		AdapterEnvironment env = new AdapterEnvironment();
		env.mMasterCounter = counter;
		
		AdapterEnvironment env2 = new AdapterEnvironment();
		env2.mMasterCounter = counter;
		
		AdapterFitness fitness = new AdapterFitness();
		fitness.mCounter = counter;
		Simulator.getInstance().getFitnessManager().addFitnessFunction(fitness);
		
		s.setMasterEnvironment(env);
		
		s.prepareNextGeneration(1, 2, 3);
		assertEquals(s.getEnvironmentModels().size(), 2);
		assertNotNull(s.getCurrentEnvironment());
		assertEquals(env.mUpdateContentCounter, 7); //took 7 calls at environment (see testSetMasterEnvironment)
		assertEquals(listener.mGenCounter, 1);
		assertEquals(listener.mGenStartCounter, 8);
		assertEquals(listener.mGenCompCounter, 0);
		assertEquals(listener.mIndCounter, 0);
		assertEquals(listener.mTryCounter, 0);
		assertEquals(listener.mStepCounter, 0);
		assertEquals(counter.mCounter, 8);
		assertEquals(s.getCurrentGeneration(), 1);
		assertEquals(s.getNumberOfSteps(), 3);
		assertEquals(s.getNumberOfTries(), 2);
		assertEquals(currentGen.get(), 1);
		assertEquals(numberOfSteps.get(), 3);
		assertEquals(numberOfTries.get(), 2);
		
		s.removeAllEnvironmentModels();
		s.addEnvironmentModel(env);
		s.addEnvironmentModel(env2);
		s.setCurrentEnvironment(env); //to be able to check to calls to this object.
		
		
		s.prepareNextIndividual(8);
		assertEquals(counter.mCounter, 11);
		assertEquals(env.mResetCounter, 9);
		assertEquals(env.mUpdateContentCounter, 10);
		assertEquals(listener.mIndStartCounter, 11);
		assertEquals(s.getCurrentIndividual(), 8);
		assertEquals(currentInd.get(), 8);
		
		//prepare the next try
		assertFalse(terminateTry.get());
		terminateTry.set(true);
		assertTrue(s.shouldCurrentTryBeTerminated());
		
		s.prepareNextTry(98);
		assertFalse(terminateTry.get()); //was reset to false;
		assertFalse(s.shouldCurrentTryBeTerminated());
		assertEquals(s.getCurrentTry(), 98);
		assertEquals(currentTry.get(), 98);
		assertSame(s.getCurrentEnvironment(), env); //because the try was positive: env.size() % tries.
		assertEquals(counter.mCounter, 15);
		assertEquals(env.mResetCounter, 13);
		assertEquals(env.mUpdateContentCounter, 14);
		assertEquals(fitness.mResetCounter, 12);
		assertEquals(listener.mTryStartCounter, 15);
		
		
		s.executeSimulationStep(58);
		assertEquals(counter.mCounter, 19);
		assertEquals(s.getCurrentStep(), 58);
		assertEquals(currentStep.get(), 58);
		assertEquals(listener.mStepStartCounter, 16);
		assertEquals(env.mExecuteSimulationStepCounter, 17);
		assertEquals(fitness.mExecuteCounter, 18);
		assertEquals(listener.mStepCompCounter, 19);
		assertEquals(terminateTry.get(), false);
		
		killTry.set(true);
		
		s.executeSimulationStep(11);
		assertEquals(counter.mCounter, 23);
		assertEquals(s.getCurrentStep(), 11);
		assertEquals(listener.mStepStartCounter, 20);
		assertEquals(env.mExecuteSimulationStepCounter, 21);
		assertEquals(fitness.mExecuteCounter, 22);
		assertEquals(listener.mStepCompCounter, 23);
		assertEquals(terminateTry.get(), true);
		assertTrue(s.shouldCurrentTryBeTerminated()); //was switched to true.
		
		s.prepareNextTry(45);
		assertFalse(terminateTry.get()); //was reset to false;
		assertTrue(s.shouldCurrentTryBeTerminated()); //since killTry still is true.
		killTry.set(false);
		assertFalse(s.shouldCurrentTryBeTerminated());
		assertEquals(s.getCurrentTry(), 45);
		assertSame(s.getCurrentEnvironment(), env2); //because the try was negative: env.size() % tries.
		assertEquals(counter.mCounter, 28);
		assertEquals(listener.mTryCompCounter, 24);
		assertEquals(env2.mResetCounter, 26);
		assertEquals(env2.mUpdateContentCounter, 27);
		assertEquals(fitness.mResetCounter, 25);
		assertEquals(listener.mTryStartCounter, 28);
		
		
		s.prepareNextIndividual(4);
		assertEquals(counter.mCounter, 33);
		assertEquals(listener.mTryCompCounter, 29);
		assertEquals(listener.mIndCompCounter, 30);
		assertEquals(env2.mResetCounter, 31);
		assertEquals(env2.mUpdateContentCounter, 32);
		assertEquals(listener.mIndStartCounter, 33);
		assertEquals(s.getCurrentIndividual(), 4);
		
		s.prepareNextGeneration(3, 1, 5);
		assertEquals(s.getEnvironmentModels().size(), 1);
		assertNotNull(s.getCurrentEnvironment());
		assertEquals(listener.mGenCounter, 3);
		assertEquals(listener.mIndCompCounter, 34);
		assertEquals(listener.mGenCompCounter, 35);
		assertEquals(env.mUpdateContentCounter, 38); //took 3 calls at environment (see testSetMasterEnvironment)
		assertEquals(listener.mGenStartCounter, 39);
		assertEquals(counter.mCounter, 39);
		assertEquals(s.getCurrentGeneration(), 3);
		assertEquals(s.getNumberOfSteps(), 5);
		assertEquals(s.getNumberOfTries(), 1);

		//TODO evt. check arbitrary combinations (step, gen, try, ...)
		
		
	}

	public void testEmulateEvolutionRun() {
		
	}

	public void testAddListeners() {
		AdapterSimulationListener listener = new AdapterSimulationListener();
		MasterCounter counter = new MasterCounter();
		listener.mMasterCounter = counter;
		
		Scheduler s = Simulator.getInstance().getScheduler();
		
		assertFalse(s.addStepListener(null));
		assertTrue(s.addStepListener(listener));
		assertFalse(s.addStepListener(listener)); //already contained
		
		assertFalse(s.addGenerationListener(null));
		assertTrue(s.addGenerationListener(listener));
		assertFalse(s.addGenerationListener(listener)); //already contained
		
		assertFalse(s.addIndividualListener(null));
		assertTrue(s.addIndividualListener(listener));
		assertFalse(s.addIndividualListener(listener)); //already contained
		
		assertFalse(s.addTryListener(null));
		assertTrue(s.addTryListener(listener));
		assertFalse(s.addTryListener(listener)); //already contained
		
		counter.mCounter = 0;
		s.notifyGenerationListeners(4, false);
		assertEquals(listener.mGenCounter, 0); //there was no generation started, so there is no completition to report.
		
		s.notifyGenerationListeners(4, true);
		assertEquals(listener.mGenCounter, 1);
		assertEquals(listener.mLastReceivedId, 4);
		assertEquals(listener.mGenCompCounter, 0);
		assertEquals(listener.mGenStartCounter, 1);
		
		s.notifyIndividualListeners(3, false);
		assertEquals(listener.mIndCounter, 0); //no individual was started before, so there is nothing to complete
		
		s.notifyIndividualListeners(3, true);
		assertEquals(listener.mIndCounter, 1);
		assertEquals(listener.mLastReceivedId, 3);
		assertEquals(listener.mIndStartCounter, 2);
		assertEquals(listener.mIndCompCounter, 0);
		
		s.notifyTryListeners(7, false);
		assertEquals(listener.mTryCounter, 0); //no Try was started before, so there is nothing to complete
		
		s.notifyTryListeners(7, true);
		assertEquals(listener.mTryCounter, 1);
		assertEquals(listener.mLastReceivedId, 7);
		assertEquals(listener.mTryStartCounter, 3);
		assertEquals(listener.mTryCompCounter, 0);
		
		s.notifyStepListeners(55, false);
		assertEquals(listener.mStepCounter, 0); //no Step was started before, so there is nothing to complete
		
		s.notifyStepListeners(55, true);
		assertEquals(listener.mStepCounter, 1);
		assertEquals(listener.mLastReceivedId, 55);
		assertEquals(listener.mStepStartCounter, 4);
		assertEquals(listener.mStepCompCounter, 0);
		
		//notify termination
		s.notifyGenerationListeners(8, false);
		assertEquals(listener.mGenCounter, 2);
		assertEquals(listener.mLastReceivedId, 8);
		assertEquals(listener.mGenCompCounter, 5);
		assertEquals(listener.mGenStartCounter, 1);
		
		s.notifyIndividualListeners(30, false);
		assertEquals(listener.mIndCounter, 2);
		assertEquals(listener.mLastReceivedId, 30);
		assertEquals(listener.mIndStartCounter, 2);
		assertEquals(listener.mIndCompCounter, 6);
		
		s.notifyTryListeners(77, false);
		assertEquals(listener.mTryCounter, 2);
		assertEquals(listener.mLastReceivedId, 77);
		assertEquals(listener.mTryStartCounter, 3);
		assertEquals(listener.mTryCompCounter, 7);
		
		s.notifyStepListeners(555, false);
		assertEquals(listener.mStepCounter, 2);
		assertEquals(listener.mLastReceivedId, 555);
		assertEquals(listener.mStepStartCounter, 4);
		assertEquals(listener.mStepCompCounter, 8);
		
		//remove listeners
		assertFalse(s.removeGenerationListener(null));
		assertTrue(s.removeGenerationListener(listener));
		assertFalse(s.removeGenerationListener(listener)); //not contained any more
		s.notifyGenerationListeners(99, true);
		assertEquals(listener.mGenCounter, 2); //not changed.
		
		assertFalse(s.removeIndividualListener(null));
		assertTrue(s.removeIndividualListener(listener));
		assertFalse(s.removeIndividualListener(listener)); //not contained any more
		s.notifyIndividualListeners(99, true);
		assertEquals(listener.mIndCounter, 2); //not changed.
		
		assertFalse(s.removeStepListener(null));
		assertTrue(s.removeStepListener(listener));
		assertFalse(s.removeStepListener(listener)); //not contained any more
		s.notifyStepListeners(99, true);
		assertEquals(listener.mStepCounter, 2); //not changed.
		
		assertFalse(s.removeTryListener(null));
		assertTrue(s.removeTryListener(listener));
		assertFalse(s.removeTryListener(listener)); //not contained any more
		s.notifyTryListeners(99, true);
		assertEquals(listener.mTryCounter, 2); //not changed.
		
	}

	public void testAddTerminateTryValue() {
		BooleanValue b1 = new BooleanValue(false);
		BooleanValue b2 = new BooleanValue(false);
		
		Scheduler s = new Scheduler(Simulator.getInstance());
		
		assertFalse(s.shouldCurrentTryBeTerminated());
		
		s.addTerminateTryValue(b1);
		assertFalse(s.shouldCurrentTryBeTerminated());
		
		s.addTerminateTryValue(b2);
		assertFalse(s.shouldCurrentTryBeTerminated());
		
		b2.set(true);
		assertTrue(s.shouldCurrentTryBeTerminated());
		
		b2.set(false);
		assertFalse(s.shouldCurrentTryBeTerminated());
		
		b1.set(true);
		assertTrue(s.shouldCurrentTryBeTerminated());
		
		s.removeTerminateTryValue(b1);
		assertFalse(s.shouldCurrentTryBeTerminated());
		
	}
	
	public class MasterCounter {
		public int mCounter;
	}
	
	public class AdapterEnvironment extends EnvironmentModel {
		
		public int mCreateEnvironmentsCounter;
		public int mExecuteSimulationStepCounter;
		public int mInitCounter;
		public int mResetCounter;
		public int mUpdateContentCounter;
		public int mResolveDependenciesCounter;
		public MasterCounter mMasterCounter;

		public void createEnvironment() {
		}

		public void resetEnvironmentObjects() {
		}

		public void updateRandomizationParameters() {
		}

		public Vector createEnvironments(int numberOfEnvironments) {
			mCreateEnvironmentsCounter = ++mMasterCounter.mCounter;
			return super.createEnvironments(numberOfEnvironments);
		}

		public void executeSimulationStep() {
			mExecuteSimulationStepCounter = ++mMasterCounter.mCounter;
		}

		public boolean initEnvironment() {
			mInitCounter = ++mMasterCounter.mCounter;
			return super.initEnvironment();
		}

		public void reset() {
			mResetCounter = ++mMasterCounter.mCounter;
			super.reset();
		}

		public void resolveDependencies() {
			mResolveDependenciesCounter  = ++mMasterCounter.mCounter;
			super.resolveDependencies();
		}

		public synchronized void updateContent() {
			mUpdateContentCounter = ++mMasterCounter.mCounter;
			super.updateContent();
		}
	}
	
	public class AdapterSimulationListener implements StepListener, TryListener, GenerationListener, IndividualListener {

		public MasterCounter mMasterCounter;
		public int mStepCompCounter;
		public int mStepStartCounter;
		public int mStepCounter;
		public int mIndCompCounter;
		public int mIndStartCounter;
		public int mIndCounter;
		public int mTryCompCounter;
		public int mTryStartCounter;
		public int mTryCounter;
		public int mGenCompCounter;
		public int mGenStartCounter;
		public int mGenCounter;
		public int mLastReceivedId;
		
		
		public void stepCompleted(int stepNumber) {
			mStepCounter++;
			mStepCompCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = stepNumber;
		}

		public void stepStarted(int stepNumber) {
			mStepCounter++;
			mStepStartCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = stepNumber;
		}

		public void individualCompleted(int individualId) {
			mIndCounter++;
			mIndCompCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = individualId;
		}

		public void individualStarted(int individualId) {
			mIndCounter++;
			mIndStartCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = individualId;
		}

		public void tryCompleted(int tryId) {
			mTryCounter++;
			mTryCompCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = tryId;
		}

		public void tryStarted(int tryId) {
			mTryCounter++;
			mTryStartCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = tryId;
		}

		public void generationCompleted(int generationId) {
			mGenCounter++;
			mGenCompCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = generationId;
		}

		public void generationStarted(int generationId) {
			mGenCounter++;
			mGenStartCounter = ++mMasterCounter.mCounter;
			mLastReceivedId = generationId;
		}
	}
	
	public class AdapterFitness extends FitnessFunction {
		
		public MasterCounter mCounter;
		public int mResetCounter;
		public int mExecuteCounter;

		public AdapterFitness() {
			super("");
		}
		
		public void reset() {
			mResetCounter = ++mCounter.mCounter;
		}

		public double calculateFitness() {
			mExecuteCounter = ++mCounter.mCounter;
			return 0.0;
		}	
	}

}
