
package de.fhg.iais.outdoor.simba.visualization;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.Properties;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import java.util.logging.Logger;

import javax.swing.JButton;
import javax.swing.JPanel;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.StepListener;
import de.fhg.iais.outdoor.simba.core.Scheduler;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.SimObjectValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class SimulationVisualization implements Visualization, 
		StepListener, MouseListener, MouseMotionListener, ValueChangedListener, 
		MouseWheelListener, OptimizePerformanceListener, KeyListener
{
	

	protected DisplayPanel mVisuArea;
	protected JPanel mPanel;
	protected Point2D.Double mDefaultOffset;
	
	protected PointValue mOffset;
	protected SimObjectValue mDraggedObject;
	protected SimObjectValue mLastSelectedObject;
	protected BooleanValue mShowSimulation;
	protected BooleanValue mPauseSystem;
	protected IntValue mVisualizationInterval;
	protected DoubleValue mScale;
	protected PointValue mLastLeftClickPosition;
	protected PointValue mLastRightClickPosition;
	
	protected boolean mShowSimulationFlag;
	protected Point2D.Double mDragStartingPoint;
	protected Point2D.Double mDragParentStartingPoint;
	protected int mPressedButton;
	protected int mModifierButton;
	protected JButton mZoomInButton;
	protected JButton mZoomDefaultButton;
	protected JButton mZoomOutButton;
	protected JButton mShowButton;
	protected JButton mPauseButton;
	protected JButton mSetDefaultButton;
	protected double mScaleInterval = 0.9;
	protected double mDefaultScale = 1.0;
	protected double mLastScaling = 1.0;
	protected int mVisualizationIntervalCounter;
	protected SimObjectValue mObjectToFollow;
	protected EnvironmentModel mModelToVisualize;
	protected Timer mTimer;
	protected String mValuePath;
	protected Vector mSelectedObjects = new Vector();
	protected Rectangle mSelectRect;
	protected IntValue mLastPressedControlKey;
	private BooleanValue mSelectedObjectsChanged;
	
	
	public SimulationVisualization() {
		this("/Visualization/Control", null);
	}
	/**
	 * The SimulationVisualization is the main visualization to view the simulated objects.
	 * The provided JPanel supports zooming and moving of the viewed perspective by mouse movents
	 * and the manipulation of objects with the right mouse button. The zoom can be controlled
	 * with the mouse wheel of the zoom buttons on the visualization.
	 * 
	 * It is also possible to set the camera to a follow mode. In this mode the camera follows
	 * an object in the simulation so that this object is always centered in the screen. To
	 * choose an object to follow, press SHIFT + CTRL and left click on the desired object.
	 * Any other click of any mouse button deselects the object.
	 * 
	 * To manipulate an object, you have to click on that object with the right mouse button.
	 * If the object has no reference object, then this object can be moved to another place in the
	 * simulation. If the object has a reference object, then the root object of the reference chain 
	 * is selected and moved, so the entire "component" is moved. 
	 * To move an object relative to its reference object, hold down the CTRL key when right clicking.
	 * Objects can also be rotated around their center point. To achieve this the object has to 
	 * be right clicked while pressing the ALT key. WIth drag movements to the left and right the
	 * rotation of the object can be controlled.
	 * 
	 * The default button focuses the view on the default location with the default zoom. 
	 * These default settings can be set by calling the method setDefaultView() or by
	 * clicking the "Make Default" button in the visualization.
	 *
	 */
	public SimulationVisualization(String valuePath, EnvironmentModel modelToVisualize) {
		mValuePath = valuePath;
		mShowSimulationFlag = true;
		mModelToVisualize = modelToVisualize;
		mVisuArea = new DisplayPanel();
		mVisuArea.addMouseListener(this);
		mVisuArea.addMouseMotionListener(this);
		mVisuArea.addMouseWheelListener(this);
		mVisualizationIntervalCounter = 0;
		mDraggedObject = new SimObjectValue();
		mDraggedObject.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/DraggedObject", mDraggedObject);
		mLastSelectedObject = new SimObjectValue();
		mLastSelectedObject.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/LastSelectedObject", mLastSelectedObject);
		mPressedButton = -1;
		mShowSimulation = new BooleanValue(true);
		mShowSimulation.addValueChangedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/ShowSimulation", mShowSimulation);
		mScale = new DoubleValue(mDefaultScale);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/Scale", mScale);
		mScale.addValueChangedListener(this);
		mOffset = new PointValue(0.0, 0.0);
		mOffset.addValueChangedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/Offset", mOffset);
		mVisualizationInterval = new IntValue(1);
		mVisualizationInterval.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/VisualizationInterval", mVisualizationInterval);
		mObjectToFollow = new SimObjectValue();
		mObjectToFollow.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/ObjectToFollow", mObjectToFollow);
		mLastLeftClickPosition = new PointValue();
		mLastLeftClickPosition.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/LastLeftClickPosition", mLastLeftClickPosition);
		mLastRightClickPosition = new PointValue();
		mLastRightClickPosition.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/LastRightClickPosition", mLastRightClickPosition);
		mLastPressedControlKey = new IntValue();
		mLastPressedControlKey.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/LastControlKey", mLastPressedControlKey);
		mSelectedObjectsChanged = new BooleanValue(false);
		mSelectedObjectsChanged.addValueReplacedListener(this);
		Simulator.getInstance().getValueManager().addValue(valuePath + "/SelectedObjectsChanged", mSelectedObjectsChanged);
		
		mDefaultOffset = new Point2D.Double(0.0, 0.0);
		mDefaultScale = 1.0;
		
		mZoomInButton = new JButton("Z+");
		mZoomDefaultButton = new JButton("Default");
		mZoomOutButton = new JButton("Z-");
		mShowButton = new JButton("Hide");
		mPauseButton = new JButton("Pause");
		mSetDefaultButton = new JButton("Make Default");
		
		
		mZoomInButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				mScale.set(mScale.get() / mScaleInterval);
			}
		});
		
		mZoomDefaultButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				restoreDefaultView();
			}
		});
		
		mZoomOutButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				mScale.set(mScale.get() * mScaleInterval);
			}
		});
		
		mShowButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				mShowSimulation.set(!mShowSimulation.get());
				mShowSimulationFlag = mShowSimulation.get();
			}
		});
		
		mPauseButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(mPauseButton.getText().equals("Pause")) {
					Simulator.getInstance().pauseSimulation(true);
				}
				else {
					Simulator.getInstance().pauseSimulation(false);
				}
			}
		});
		
		mSetDefaultButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				setDefaultView();
			}
		});
		
		JPanel controlPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		controlPanel.add(mSetDefaultButton);
		controlPanel.add(mZoomDefaultButton);
		controlPanel.add(mZoomInButton);
		controlPanel.add(mZoomOutButton);
		controlPanel.add(mPauseButton);
		controlPanel.add(mShowButton);
		
		
		mPanel = new JPanel(new BorderLayout());
		mPanel.add(BorderLayout.CENTER, mVisuArea);
		mPanel.add(BorderLayout.SOUTH, controlPanel);
		
		
		Simulator.getInstance().addOptimizePerformanceListener(this);
		Simulator.getInstance().getScheduler().addStepListener(this);
		
		mSelectRect = new Rectangle(0, 0, -1, -1);
		
		mTimer = new Timer();
		mVisuArea.setFocusable(true);
		mVisuArea.addKeyListener(this);
	}
	
	public void restoreDefaultView() {
		mScale.set(mDefaultScale);
		setOffset(mDefaultOffset.x, mDefaultOffset.y);
	}
	
	public void setDefaultView() {
		mDefaultScale = mScale.get();
		mDefaultOffset.x = mOffset.get().x;
		mDefaultOffset.y = mOffset.get().y;
		Simulator.getInstance().reportStatus("Default: Offset: [" 
				+ ((int) mDefaultOffset.x) + "/"
				+ ((int) mDefaultOffset.y) + "] Zoom: " + mDefaultScale); 
	}
	
	public boolean init() {
		boolean ok = true;
		
		acquireValueObjects();
		
		//Check if there is a previuosly saved view point available.
		Properties props = Simulator.getInstance().getSystemProperties();
		String initView = props.getProperty(mValuePath + "/InitView");
		if(initView != null) {
			String[] view = initView.split(",");
			if(view.length == 3) {
				try {
					double x = Double.parseDouble(view[0]);
					double y = Double.parseDouble(view[1]);
					double s = Double.parseDouble(view[2]);
					mScale.set(s);
					mOffset.set(x, y);
				} catch (NumberFormatException e) {
				}
			}
		}
		mSelectedObjects.clear();
		mSelectedObjectsChanged.set(true);
		
		mTimer.cancel();
		mTimer = new Timer();
		mTimer.schedule(new RepaintByTimerTask(), 500, 500);
		
		return ok;
	}
	
	public boolean cleanUp() {
		
		//Save the current viewpoint in the properties.
		Properties props = Simulator.getInstance().getSystemProperties();
		props.put(mValuePath + "/InitView", 
				Double.toString(mOffset.get().x) + "," +
				Double.toString(mOffset.get().y) + "," +
				Double.toString(mScale.get()));

		mSelectedObjects.clear();
		mSelectedObjectsChanged.set(true);
		return true;
	}
	
	
	protected void acquireValueObjects() {
		mPauseSystem = Simulator.getInstance().getValueManager().getBooleanValue("/Simulation/Pause");
		if(mPauseSystem == null) {
			Logger.getLogger("Simulator")
				.warning("SimulationVisualization: Could not find BooleanValue\n[/Simulation/Pause]");
		}
		else {
			mPauseSystem.addValueChangedListener(this);
		}
	}
	
	
	
	public void stepStarted(int stepNumber) {
		
	}
	
	/**
	 * Repaints the visualization after each step.
	 */
	public void stepCompleted(int stepNumber) {
		mVisualizationIntervalCounter++;
		if(mShowSimulation.get()) {
			if(mVisualizationIntervalCounter >= mVisualizationInterval.get()) {
				mVisualizationIntervalCounter = 0;
				mVisuArea.repaint();
			}
		}
	}
	
	/**
	 * Returns the JPanel with the user interface of this visualization.
	 */
	public JPanel getPanel() {
		return mPanel;
	}
	
	/**
	 * Sets the current scaling factor.
	 * @param scale the new scaling factor.
	 */
	public void setScale(double scale) {
		mScale.set(scale);
	}

	/**
	 * Returns the current scale factor.
	 * 
	 * @return the scaling factor.
	 */
	public double getScale() {
		return mScale.get();
	}
	
	/**
	 * Sets the current offset of the visualization area.
	 * @param x the x offset.
	 * @param y the y offset.
	 */
	public void setOffset(double x, double y) {
		mOffset.set(x, y);
	}
	
	/**
	 * Returns the current offset of the visualization area.
	 * @return the offset.
	 */
	public Point2D.Double getOffset() {
		return mOffset.get();
	}
	
	public DisplayPanel getDisplayPanel() {
		return mVisuArea;
	}
	
	public Vector getSelectedObjects() {
		return new Vector(mSelectedObjects);
	}
	
	public void setSelectedObjects(Vector selectedObjects) {
		mSelectedObjects.clear();
		mSelectedObjects.addAll(selectedObjects);
		mSelectedObjectsChanged.set(true);
		mVisuArea.repaint();
	}
	
	/**
	 * Sets the interval in which the visualization is painted. The interval is 
	 * specified in steps. 
	 * Example: If the interval is 5, then the visualizatino is repainted only every 
	 * 5th step or if the screen has to be repained anyhow (e.g. because the window was
	 * hidden and has to be refreshed now).
	 * 
	 * @param interval the new interval for the visualization update.
	 */
	public void setVisualizationInterval(int interval) {
		if(mVisualizationInterval.get() != interval) {
			mVisualizationInterval.set(interval);
		}
	}
	
	/**
	 * Returns the visualization interval that determines the frequency at which
	 * the visualization is refreshed. The interval is given in steps.
	 * 
	 * @return the visualization interval.
	 */
	public int getVisualizationInterval() {
		return mVisualizationInterval.get();
	}

	/**
	 * If the mouse was double clicked, then the offset of the visualization 
	 * is changed so that the clicked position is in the center of the 
	 * visualization panel.
	 */
	public void mouseClicked(MouseEvent e) {
		if(e.getButton() == MouseEvent.BUTTON1) {
			if(e.getClickCount() >= 2) {
				Point p = e.getPoint();
				double width = mVisuArea.getSize().getWidth();
				double height = mVisuArea.getSize().getHeight();
				double scale = mScale.get();
				
				Point2D.Double offset = mOffset.get();
				mOffset.set(((((double) p.x) / scale) + offset.x) - ((width / 2.0) / scale),
							((((double) p.y) / scale) + offset.y)  - ((height / 2.0) / scale));
				
				mVisuArea.repaint();
			}
		}
	}
    
	public void mouseEntered(MouseEvent e) {	
	}
    
	public void mouseExited(MouseEvent e) {
	}
  
	/**
	 * If the left or right mouse button is pressed, then check whether
	 * there is an object at that position. If so, then start dragging 
	 * around that object.
	 * If the right button was pressend AND the shift key was down, then
	 * add the object to the selectio list. All objects in the selection 
	 * list can be moved together. Clicking on an object, that is already
	 * contained in that list, will remove the object.
	 */
	public void mousePressed(MouseEvent e) {
		
		mVisuArea.requestFocusInWindow();
		
		mDragStartingPoint = new Point2D.Double((double)  e.getX(),(double) e.getY());
		Point2D.Double globalPos = new Point2D.Double();
		AffineTransform transformation = new AffineTransform();
		transformation.scale(mScale.get(), -1 * mScale.get());
		transformation.translate(mOffset.getX() * -1, mOffset.getY());
		try {
			transformation = transformation.createInverse();
		}
		catch(NoninvertibleTransformException ex) {
			ex.printStackTrace();
		}
		transformation.transform(mDragStartingPoint, globalPos);
		
		
		//Left mouse button
		if(e.getButton() == MouseEvent.BUTTON1 && e.getClickCount() == 1) {
			
			//+ CTRL + SHIFT
			if(((e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) == MouseEvent.SHIFT_DOWN_MASK)
					&& ((e.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) == MouseEvent.CTRL_DOWN_MASK))
			{
				grabObject(e.getPoint());
				mObjectToFollow.set(mDraggedObject.get());
				mDraggedObject.set(null);
				mDragStartingPoint = null;
				mDragParentStartingPoint = null;
			}
			//Other modifiers
			else {
				if(mObjectToFollow.get() != null) {
					mObjectToFollow.set(null);
				}
				mPressedButton = e.getButton();
				mModifierButton = e.getModifiersEx();
				//grabObject(e.getPoint());
			}
			mLastLeftClickPosition.set(globalPos);
		}
		//right mouse button
		else if(e.getButton() == MouseEvent.BUTTON3 && e.getClickCount() == 1) {
			if(mObjectToFollow.get() != null) {
				mObjectToFollow.set(null);
			}
			mPressedButton = e.getButton();
			mModifierButton = e.getModifiersEx();
			grabObject(e.getPoint());
			mLastRightClickPosition.set(globalPos);
			
			//+ SHIFT: Select several objects to move.
			if(((e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) == MouseEvent.SHIFT_DOWN_MASK)) {
				if(mDraggedObject.get() != null) {
					SimObject obj = (SimObject) mDraggedObject.get();
					
					if(mSelectedObjects.contains(obj)) {
						obj.setMarked(false);
						mSelectedObjects.remove(obj);
						mSelectedObjectsChanged.set(true);
					}
					else {
						if(obj.isActive()) {
							obj.setMarked(true);
							mSelectedObjects.add(obj);
							mSelectedObjectsChanged.set(true);
						}
					}
				}
			}
			else { //no modifiers pressed.
				if(!(mDraggedObject.get() != null && mSelectedObjects.contains(mDraggedObject.get()))
						|| mDraggedObject.get() == null) 
				{
					//if the currently dragged object is one of the marked ones, then do nothing.
					//Otherwise clear the select list.
					for(Iterator i = mSelectedObjects.iterator(); i.hasNext();) {
						((SimObject) i.next()).setMarked(false);
					}
					mSelectedObjects.clear();
					mSelectedObjectsChanged.set(true);
				}
				if(mDraggedObject.get() != null && !mSelectedObjects.contains(mDraggedObject.get())) {
					if(mDraggedObject.get().isActive()) {
						mSelectedObjects.add(mDraggedObject.get());
						mDraggedObject.get().setMarked(true);
						mSelectedObjectsChanged.set(true);
					}
				}
				mLastSelectedObject.set(mDraggedObject.get());
			}
		}
		mVisuArea.repaint();
	}
   
	/**
	 * Release all grabbed objects.
	 */
	public void mouseReleased(MouseEvent e) {
		mDraggedObject.set(null);
		mDragStartingPoint = null;
		mDragParentStartingPoint = null;
		mSelectRect.width = -1;
		mVisuArea.repaint();
		mVisuArea.requestFocusInWindow();
	}
	
	
	/**
	 * If an object was grabbed with a left click, then that object is moved
	 * to the new coordinates specified by the current mouse position.
	 * If an object was grabbed with a right click, then that object's orientation
	 * is changed according to the horizontal mouse movement.
	 * 
	 */
	public void mouseDragged(MouseEvent e) {
		SimObject draggedObject = mDraggedObject.get();
		Point p = e.getPoint();
		
		
		if(draggedObject == null && mPressedButton == MouseEvent.BUTTON3) {
			//Select neurons with select rectangle.
			Point2D.Double globalPos = new Point2D.Double();
			AffineTransform transformation = new AffineTransform();
			transformation.scale(mScale.get(), -1 * mScale.get());
			transformation.translate(mOffset.getX() * -1, mOffset.getY());
			try {
				transformation = transformation.createInverse();
			}
			catch(NoninvertibleTransformException ex) {
				ex.printStackTrace();
			}
			transformation.transform(p, globalPos);
			
			if(globalPos.x > mLastRightClickPosition.getX()) {
				if(globalPos.y < mLastRightClickPosition.getY()) {
					mSelectRect.x = (int) mLastRightClickPosition.getX();
					mSelectRect.y = (int) globalPos.y;
					mSelectRect.width = (int) (globalPos.x - mLastRightClickPosition.getX());
					mSelectRect.height = (int) (mLastRightClickPosition.getY() - globalPos.y);
				}
				else {
					mSelectRect.x = (int) mLastRightClickPosition.getX();
					mSelectRect.y = (int) mLastRightClickPosition.getY();
					mSelectRect.width = (int) (globalPos.x - mLastRightClickPosition.getX());
					mSelectRect.height = (int) (globalPos.y - mLastRightClickPosition.getY());
				}
			}
			else {
				if(globalPos.y < mLastRightClickPosition.getY()) {
					mSelectRect.x = (int) globalPos.x;
					mSelectRect.y = (int) globalPos.y;
					mSelectRect.width = (int) (mLastRightClickPosition.getX() - globalPos.x);
					mSelectRect.height = (int) (mLastRightClickPosition.getY() - globalPos.y);
				}
				else {
					mSelectRect.x = (int) globalPos.x;
					mSelectRect.y = (int) mLastRightClickPosition.getY();
					mSelectRect.width = (int) (mLastRightClickPosition.getX() - globalPos.x);
					mSelectRect.height = (int) (globalPos.y - mLastRightClickPosition.getY());
				
				}
			}
			
			Vector oldSelectedObjects = new Vector(mSelectedObjects);
			
			if((e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != MouseEvent.SHIFT_DOWN_MASK) {
				for(Iterator i = mSelectedObjects.iterator(); i.hasNext();) {
					((SimObject) i.next()).setMarked(false);
				}
				mSelectedObjects.clear();
				mSelectedObjectsChanged.set(true);
			}
			
			
			Vector objects = null;
			if(mModelToVisualize == null) {
				if(Simulator.getInstance().getScheduler().getCurrentEnvironment() == null) {
					return;
				}
				objects = Simulator.getInstance().getScheduler().getCurrentEnvironment().getSimObjectsSortedByLayers();
			}
			else {
				objects = mModelToVisualize.getSimObjectsSortedByLayers();
			}
			
			SimObject currentSelectedObject = null;
			
			boolean selectedObjectsChanged = false;
			for(Iterator i = objects.iterator(); i.hasNext();) {
				SimObject obj = (SimObject) i.next();
				if(obj.getReferenceObject() == null && obj.isActive()) {
					if(mSelectRect.contains(obj.getGlobalPosition())) {
						if(!mSelectedObjects.contains(obj)) {
							mSelectedObjects.add(obj);
							selectedObjectsChanged = true;
							obj.setMarked(true);
							if(!oldSelectedObjects.contains(obj)) {
								currentSelectedObject = obj;
							}
						}
					}
				}
			}
			if(selectedObjectsChanged) {
				mSelectedObjectsChanged.set(true);
			}
			
			if(currentSelectedObject != null) {
				mLastSelectedObject.set(currentSelectedObject);
			}
			else if(!mSelectedObjects.isEmpty() && mLastSelectedObject.get() != null) {
				if(!mSelectedObjects.contains(mLastSelectedObject.get())) {
					mLastSelectedObject.set((SimObject) mSelectedObjects.firstElement());
				}
			}
			else {
				mLastSelectedObject.set(null);
			}
			
		}
		else if(mDragStartingPoint != null) {
			
			
			if(draggedObject != null) {
				Point2D.Double objPos = draggedObject.getPosition();
				
				if(draggedObject != null && mPressedButton == MouseEvent.BUTTON3) {
					if((mModifierButton & MouseEvent.ALT_DOWN_MASK) != MouseEvent.ALT_DOWN_MASK) 
					{
						double scale = mScale.get();
						
						double x = (((double) p.x) - mDragStartingPoint.x) / scale;
						double y = (((double) p.y) - mDragStartingPoint.y) / (-1 * scale);
						
						if(mSelectedObjects.isEmpty()) {
							draggedObject.setPosition(objPos.x + x,
									objPos.y + y);
						}
						else {
							for(Iterator i = mSelectedObjects.iterator(); i.hasNext();) {
								SimObject obj = (SimObject) i.next();
								objPos = obj.getPosition();
								obj.setPosition(objPos.x + x,
															objPos.y + y);
							}
						}
					}
					else if((mModifierButton & MouseEvent.ALT_DOWN_MASK) == MouseEvent.ALT_DOWN_MASK
							&& (mModifierButton & MouseEvent.CTRL_DOWN_MASK) != MouseEvent.CTRL_DOWN_MASK)
					{
						double changeX = (((double) p.x) - mDragStartingPoint.x);
						double angle = draggedObject.getAngle() + changeX;
						while(angle > 360.0) {
							angle -= 360.0;
						}
						while(angle < 0.0) {
							angle += 360.0;
						}
						draggedObject.setAngle(angle);
					}
				}
			}
			else if(draggedObject == null && mPressedButton == MouseEvent.BUTTON1) {
				if((mModifierButton & MouseEvent.CTRL_DOWN_MASK) != MouseEvent.CTRL_DOWN_MASK
					&& (mModifierButton & MouseEvent.ALT_DOWN_MASK) != MouseEvent.ALT_DOWN_MASK) 
				{
					double scale = mScale.get();
					
					double x = (((double) p.x) - mDragStartingPoint.x) / scale;
					double y = (((double) p.y) - mDragStartingPoint.y) / (-1 * scale);
					
					Point2D.Double offset = mOffset.get();
					mOffset.set(offset.x - x, offset.y + y);
				}
			}
			mDragStartingPoint.x = (double) p.x;
			mDragStartingPoint.y = (double) p.y;
			if(mDraggedObject.get() != draggedObject) {
				mDraggedObject.set(draggedObject);
			}
		}
		mVisuArea.repaint();
	 }
	 
	 public void mouseMoved(MouseEvent e) {
	 }
	 
	 /**
	  * The mouse wheel can be used to zoom in and out.
	  */
	 public void mouseWheelMoved(MouseWheelEvent e) {
		 int rotations = e.getWheelRotation();
		 if(rotations != 0) {
			 if(rotations > 0) {
				 mScale.set(Math.pow(mScaleInterval, (double) rotations) * mScale.get());
			 }
			 else {
				 mScale.set(Math.pow(mScaleInterval, 1.0 / (double) rotations) * mScale.get());
			 }
		 }
		 mVisuArea.repaint();
	 }
	 
	 
	 
	 
	public void keyPressed(KeyEvent e) {
	}
	public void keyReleased(KeyEvent e) {
	}
	public void keyTyped(KeyEvent e) {
		if((e.getModifiersEx() & KeyEvent.CTRL_DOWN_MASK) == KeyEvent.CTRL_DOWN_MASK) {
			mLastPressedControlKey.set(e.getKeyChar());
		}
	}
	
	/**
	 * Returns the Value object that is notified if the selected group of objects
	 * changed. 
	 * Important note: the content of this Value object is not expressive and is 
	 * always true (or undefined). This object is only used as a notification system.
	 * Whenever the group of selected objects is modified, then this object is set,
	 * so all registered ValueChangedListeners are notified. This is the only correct 
	 * use of this BooleanValue!
	 * 
	 * @return
	 */
	public BooleanValue getSelectedObjectsChangedValue() {
		return mSelectedObjectsChanged;
	}
	
	/**
	  * Called whenever a Value changes where this Visualization was
	  * set as ValueListener.
	  * 
	  * @param value the value that changed.
	  */
	 public void valueChanged(Value value) {
		 if(value == null) {
			 return;
		 }
		 if(value == mShowSimulation) {
			 if(mShowSimulation.get()) {
				 mShowButton.setText("Hide");
				 mZoomDefaultButton.setEnabled(true);
				 mZoomInButton.setEnabled(true);
				 mZoomOutButton.setEnabled(true);
				 mShowSimulationFlag = true;
				 mVisuArea.repaint();
				 
			 }
			 else {
				 mShowButton.setText("Show");
				 mZoomDefaultButton.setEnabled(false);
				 mZoomInButton.setEnabled(false);
				 mZoomOutButton.setEnabled(false);
				 mShowSimulationFlag = false;
				 mVisuArea.repaint();
			 }
		 }
		 else if(value == mScale) {
			 scaleVisualization();
		 }
		 else if(value == mPauseSystem) {
			 if(!mPauseSystem.get()) {
				 mTimer.cancel();
				 mPauseButton.setText("Pause");
				 mVisuArea.repaint();
			 }
			 else {
				 mTimer.cancel();
				 mTimer = new Timer();
				 mTimer.schedule(new RepaintByTimerTask(), 500, 500);
				 mPauseButton.setText("Resume");
				 mVisuArea.repaint();
			 }
		 }
	 }
	 
	 /**
	 * Called if the one of the registered Value objects were replaced in the ValueManager.
	 */
	 public void valueReplaced(Value oldValue, Value newValue) {
		 mOffset = (PointValue) ValueManager.synchronizeValue(mOffset, oldValue, newValue, this);
		 mDraggedObject = (SimObjectValue) ValueManager.synchronizeValue(mDraggedObject, oldValue, newValue, this);
		 mLastSelectedObject = (SimObjectValue) ValueManager.synchronizeValue(mLastSelectedObject, oldValue, newValue, this);
		 mShowSimulation = (BooleanValue) ValueManager.synchronizeValue(mShowSimulation, oldValue, newValue, this);
		 mPauseSystem = (BooleanValue) ValueManager.synchronizeValue(mPauseSystem, oldValue, newValue, this);
		 mVisualizationInterval = (IntValue) ValueManager.synchronizeValue(mVisualizationInterval, oldValue, newValue, this);
		 mScale = (DoubleValue) ValueManager.synchronizeValue(mScale, oldValue, newValue, this);
		 mObjectToFollow = (SimObjectValue) ValueManager.synchronizeValue(mObjectToFollow, oldValue, newValue, this);
		 mSelectedObjectsChanged = (BooleanValue) ValueManager.synchronizeValue(mSelectedObjectsChanged, oldValue, newValue, this);
	 }
	
	 /**
	  * Checks whether there is a SimObject at position p. If so, then the SimObject is
	  * memorized to be the dragged object if the mouse is moved.
	  * If there are several objects at position p, then the object with the highest layer
	  * will be choosen.
	  * @param p
	  */
	 private void grabObject(Point p) {
		Point2D.Double startPoint = new Point2D.Double((double) p.x, (double) p.y);
		
		Vector objects = null;
		if(mModelToVisualize == null) {
			if(Simulator.getInstance().getScheduler().getCurrentEnvironment() == null) {
				return;
			}
			objects = Simulator.getInstance().getScheduler().getCurrentEnvironment().getSimObjectsSortedByLayers();
		}
		else {
			objects = mModelToVisualize.getSimObjectsSortedByLayers();
		}
				
		
		for(int i = objects.size() - 1; i >= 0; i--) {
			SimObject obj = (SimObject) objects.elementAt(i);
			if(!obj.isActive() || !obj.isVisible()) {
				continue;
				
			}
			AffineTransform transformation = new AffineTransform();
			double scale = mScale.get();
			transformation.scale(scale, -1 * scale);
			transformation.translate(mOffset.getX() * -1, mOffset.getY());
			obj.addAffineTransform(transformation);
			try {
				transformation = transformation.createInverse();
			}
			catch(NoninvertibleTransformException ex) {
				ex.printStackTrace();
			}
			Point2D.Double pos = new Point2D.Double();
			transformation.transform(startPoint, pos);
			if(obj.getShape().contains(pos)) {
				if((mModifierButton & MouseEvent.CTRL_DOWN_MASK) == MouseEvent.CTRL_DOWN_MASK) {
					mDraggedObject.set(obj);
				}
				else {
					int safetyCount = 10000;
					while(obj.getReferenceObject() != null && --safetyCount > 0) {
						obj = obj.getReferenceObject();
					}
					if(safetyCount <= 0) {
						Logger.getLogger("Simulator")
							.warning("SimulationVisualization: Detected a potential loop in the object reference relations!");
					}
					mDraggedObject.set(obj);
				}
				return;
			}
		}
		mDraggedObject.set(null);
//		mLastSelectedObject.set(null);
	}
	 
	 /** 
	 * Ensures that after a change of the scaling factor the offset is adapted 
	 * so that the mid point of the previous screen remains in the middle.
	 *
	 */
	private void scaleVisualization() {
		//make sure that the screen still is centered on the old midpoint	
		double width = mVisuArea.getSize().getWidth();
		double height = mVisuArea.getSize().getHeight();
		
		Point2D.Double offset = mOffset.get();
		mOffset.set((((width / 2.0) / mLastScaling) + offset.x) - ((width / 2.0) / mScale.get()),
					(((height / 2.0) / mLastScaling) + offset.y) - ((height / 2.0) / mScale.get()));
		
		mLastScaling = mScale.get();
		mVisuArea.repaint();
	}
	 
	/**
	 * Shows of hides the simulation display. Hiding the display increases the preformance.
	 * @param show
	 */
	public void showSimulation(boolean show) {
		mShowSimulationFlag = show;
		mShowSimulation.set(show);
	}

	public void optimizePerformance(boolean optimize) {
		boolean show = mShowSimulationFlag;
		if(mShowSimulation.get() == optimize) {
			mShowSimulation.set(!optimize);
		}
		mShowSimulationFlag = show;
	}


	
	public class DisplayPanel extends JPanel {
		
		public static final long serialVersionUID = 100001;
		private boolean mTriggeredByRepaint;
		
		public DisplayPanel() {
			setDoubleBuffered(true);
			mTriggeredByRepaint = false;
		}
		
		

		public void repaint() {
			mTriggeredByRepaint = true;
			super.repaint();
		}



		protected void paintComponent(Graphics g) {
			super.paintComponent(g);
			
			if(!mTriggeredByRepaint) {
				repaint();
			}
			else {
				mTriggeredByRepaint = false;
			}
			
			Graphics2D g2d = (Graphics2D) g;
			//g2d.setClip(0, 0, getWidth(), getHeight());
			Color oldColor = g2d.getColor();
			
			if(mShowSimulation.get()) {
				
				
				//Check if the offset has to be changed to follow an object.
				if(mObjectToFollow.get() != null) {
					Point2D.Double globalObjectPosition = mObjectToFollow.get().getGlobalPosition();
					double offsetx = globalObjectPosition.x - (getSize().width / 2.0 / mScale.get());
					double offsety = -1 * (globalObjectPosition.y + (getSize().height / 2.0 / mScale.get()));
					setOffset(offsetx, offsety);
				}
				
				Rectangle clip = getBounds();
				g2d.clearRect(clip.x, clip.y, clip.width, clip.height);
				g2d.setColor(oldColor);
				
				Scheduler scheduler = Simulator.getInstance().getScheduler();
				if(scheduler == null) {
					return;
				}
				EnvironmentModel model = mModelToVisualize;
				if(model == null) {
					model = scheduler.getCurrentEnvironment();
				}
				if(model == null) {
					return;
				}
				
				
				Vector objects = model.getSimObjectsSortedByLayers();
				for(Iterator i = objects.iterator(); i.hasNext();) {
					SimObject obj = (SimObject) i.next();
					if(obj.isActive() && obj.isVisible()) {
						AffineTransform transformation = new AffineTransform();
						double scale = mScale.get();
						transformation.scale(scale, -1 * scale);
						transformation.translate(mOffset.getX() * -1, mOffset.getY());
						obj.addAffineTransform(transformation);
						
						g2d.setTransform(transformation);
						obj.paintSelf(g2d);
					}
					
				}
				
				//Select Rect
				if(!mSelectRect.isEmpty()) {
					AffineTransform transformation = new AffineTransform();
					double scale = mScale.get();
					transformation.scale(scale, -1 * scale);
					transformation.translate(mOffset.getX() * -1, mOffset.getY());
					g2d.setTransform(transformation);
					Stroke stroke = g2d.getStroke();
					float[] dash = {8f, 8f};
					g2d.setStroke(new BasicStroke(0.01f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1.0f, dash, 0));
					g2d.drawRect(mSelectRect.x, mSelectRect.y, mSelectRect.width, mSelectRect.height);
					g2d.setStroke(stroke);
				}
				
			}
			else {
				g2d.setColor(Color.lightGray);
				Rectangle clip = g2d.getClipBounds();
				g2d.fillRect(clip.x, clip.y, clip.width, clip.height);
				g2d.setColor(oldColor);
			}
			if(mPauseSystem.get()) {
				AffineTransform transform = new AffineTransform();
				transform.scale(3.0, 3.0);
				g2d.setTransform(transform);
				Color c = g2d.getColor();
				g2d.setColor(Color.WHITE);
				g2d.fillRoundRect(17, 8, 100, 15, 8, 10);
				g2d.setColor(Color.BLACK);
				g2d.drawString("Simulator Paused", 20, 20);
				g2d.setColor(c);
			}
			Thread.yield();
		}
	}
	
	protected class RepaintByTimerTask extends TimerTask {
		
		public synchronized void run() {
			if(mPauseSystem.get() || !Simulator.getInstance().getScheduler().isGenerationRunning()) {
				if(mModelToVisualize != null) {
					mModelToVisualize.invalidateCalculationBuffers();
				}
				else {
					EnvironmentModel model = Simulator.getInstance().getScheduler().getCurrentEnvironment();
					if(model != null) {
						model.invalidateCalculationBuffers();
					}
				}
			}
		}
	
	}

}
