package de.fhg.iais.outdoor.simba.visualization.neuron;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Logger;

import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToggleButton;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetListener;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.SimObjectValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.Visualization;


/**
 * This Visualization allows the modification of the currently displayed NeuralNetwork
 * 
 * It can be used to insert or remove neurons and synapses.
 * To use the functionality the check box has to be selected previously. This is
 * to prevent changes by accident. 
 * 
 * To add a neuron, press the corresponding toggle button and afterwards right click 
 * at the position where the neurons should be inserted. The toggle button will 
 * toggle back after each insertion to prevent multiple insertions.
 * 
 * To add a synapse, press the corresponding toggle button and afterwards right click
 * at first on the source neuron, followed by a right click on the target neuron.
 * The toggle button will toggle back after each insertion to prevent multiple insertions.
 * 
 * To remove a neuron or synapse, press the delet toggle button and afterwards right click
 * on the neuron or the synapse (the strength label) to remove.
 * The delete button will toggle back after each deletion to prevent multiple deletions. 
 * 
 * @author Christian Rempis
 *
 */
public class NetworkModificationPanel implements Visualization,
		ValueChangedListener, NeuralNetListener 
{
	
	private JPanel mDisplay;
	private SimObjectValue mDraggedObject;
	private PointValue mLastRightClickPosition;
	private SimObject mLastSelectedObject; //not required
	private NeuronObject mSourceNeuron;
	private JCheckBox mActivateModificationsCheckBox;
	private JToggleButton mDeleteButton;
	private JToggleButton mInsertNeuronButton;
	private JToggleButton mInsertSynapseButton;
	private NeuronVisualizationEnvironment mVisualizationEnvironment;
	
	/**
	 * Constructor.
	 * 
	 * @param environment the neuronEnvironment that underlays the neural net visualization.
	 */
	public NetworkModificationPanel(NeuronVisualizationEnvironment environment) {
		mVisualizationEnvironment = environment;
		mDisplay = new JPanel(new BorderLayout());
		mDisplay.setBorder(BorderFactory.createTitledBorder("Structure Modification"));
		
		mActivateModificationsCheckBox = new JCheckBox("");
		mActivateModificationsCheckBox.setSelected(false);
		

		mDeleteButton = new JToggleButton("Delete Mode");
		mInsertNeuronButton = new JToggleButton("Neuron");
		mInsertSynapseButton = new JToggleButton("Sypapse");
		
		mDeleteButton.setSelected(false);
		mInsertNeuronButton.setSelected(false);
		mInsertSynapseButton.setSelected(false);
		
		JPanel addPanel = new JPanel(new BorderLayout());
		addPanel.setBorder(BorderFactory.createTitledBorder("Insert"));
		addPanel.add(mInsertNeuronButton, BorderLayout.WEST);
		addPanel.add(mInsertSynapseButton, BorderLayout.EAST);
		
		JPanel buttonPanel = new JPanel(new BorderLayout());
		buttonPanel.add(addPanel, BorderLayout.NORTH);
		JPanel delButtonPanel = new JPanel(new BorderLayout());
		delButtonPanel.setBorder(BorderFactory.createTitledBorder(""));
		delButtonPanel.add(mDeleteButton, BorderLayout.NORTH);
		buttonPanel.add(delButtonPanel, BorderLayout.SOUTH);
		
		JPanel area = new JPanel(new BorderLayout());
		area.add(buttonPanel, BorderLayout.CENTER);
		area.add(mActivateModificationsCheckBox, BorderLayout.WEST);
		
		mDisplay.add(area, BorderLayout.NORTH);
		
		mDeleteButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(mDeleteButton.isSelected()) {
					mInsertNeuronButton.setSelected(false);
					mInsertSynapseButton.setSelected(false);
				}
			}
		});
		
		mInsertNeuronButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(mInsertNeuronButton.isSelected()) {
					mDeleteButton.setSelected(false);
					mInsertSynapseButton.setSelected(false);
				}
				mSourceNeuron = null;
			}
		});
		
		mInsertSynapseButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(mInsertSynapseButton.isSelected()) {
					mInsertNeuronButton.setSelected(false);
					mDeleteButton.setSelected(false);
				}
			}
		});
		
		mActivateModificationsCheckBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				synchronizeComponents();
			}
		});
		
		synchronizeComponents();
	}
	

	public JPanel getPanel() {
		return mDisplay;
	}

	public boolean cleanUp() {
		return true;
	}

	public boolean init() {
		ValueManager vm = Simulator.getInstance().getValueManager();
		
		//This Value is used to get notified when an object was right clicked.
		mDraggedObject = vm.getSimObjectValue("/Visualization/Neurons/Control/DraggedObject");
		if(mDraggedObject != null) {
			mDraggedObject.addValueReplacedListener(this);
			mDraggedObject.addValueChangedListener(this);
		}
		//This Value is used to get the position of the last right click.
		mLastRightClickPosition = vm.getPointValue("/Visualization/Neurons/Control/LastRightClickPosition");
		if(mLastRightClickPosition != null) {
			mLastRightClickPosition.addValueChangedListener(this);
			mLastRightClickPosition.addValueChangedListener(this);
		}
		
		if(mDraggedObject == null || mLastRightClickPosition == null) {
			Logger.getLogger("Simulator").warning("NetworkModificationPanel: Could not find all required value objects.");
		}
		
		Simulator.getInstance().getNeuralNetManager().addNeuralNetListener(this);
		
		return true;
	}

	
	/**
	 * Reacts on changed of the Value DraggedObject (from the NeuralNetVisualization)
	 * and Value LastRightClickPosition. Both are used to determin whether an object
	 * and if so which object was selected. Dependent on the current state of the
	 * selection buttons the desired action is performed, e.g. deletion of a neuron
	 * or synapse or the insertion of one of them.
	 */
	public void valueChanged(Value value) {
		if(value == mDraggedObject) {
			if(mDraggedObject != null) {
				SimObject obj = mDraggedObject.get();

				if(mDeleteButton.isSelected()) {
					//Handle delete if selected.
					if(obj instanceof NeuronObject) {
						deleteNeuron((NeuronObject) obj);
						mDeleteButton.setSelected(false);
					}
					else if(obj instanceof SynapseObject) {
						deleteSynapse((SynapseObject) obj); 
						mDeleteButton.setSelected(false);
					}
				}
				else if(mInsertSynapseButton.isSelected()) {
					//handle synapse isertion if selected.
					if(mLastSelectedObject == null && obj == null) {
						mSourceNeuron = null;
					}
					else if(mSourceNeuron == null && obj instanceof NeuronObject) {
						mSourceNeuron = (NeuronObject) obj;
					}
					else if(mSourceNeuron != null && obj instanceof NeuronObject) {
						addSynapse(mSourceNeuron, (NeuronObject) obj);
						mInsertSynapseButton.setSelected(false);
						mSourceNeuron = null;
					}
				}
				mLastSelectedObject = obj;
			}
		}
		if(value == mLastRightClickPosition) {
			if(mLastRightClickPosition.get() != null && mDraggedObject.get() == null) {
				//handle neuron insertion if selected.
				if(mInsertNeuronButton.isSelected()) {
					addNeuron(mLastRightClickPosition.get());
					mInsertNeuronButton.setSelected(false);
				}
			}
		}
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		mDraggedObject = (SimObjectValue) ValueManager.synchronizeValue(mDraggedObject, oldValue, newValue, this);
		mLastRightClickPosition = (PointValue) ValueManager.synchronizeValue(mLastRightClickPosition, oldValue, newValue, this);
	}

	/**
	 * Called if the current neural network in the NeuralNetwork manager changed.
	 */
	public void currentNeuralNetChanged(NeuralNet currentNet) {
		mSourceNeuron = null;
		mDeleteButton.setSelected(false);
		mInsertNeuronButton.setSelected(false);
	}

	public void networkActivationsUpdated(NeuralNet currentNet) {
	}

	/**
	 * Called if the netowork structure changed. 
	 * If the delete button was active, then it is deselected to prevent
	 * accidental successive deletions of neurons. The same holds true for the
	 * insertion buttons.
	 */
	public void networkStructureChanged(NeuralNet currentNet) {
		mSourceNeuron = null;
		mDeleteButton.setSelected(false);
		mInsertNeuronButton.setSelected(false);
	}
	
	
	/**
	 * Updates the state of each of the buttons according to the selection
	 * of the activate check box.
	 *
	 */
	public void synchronizeComponents() {
		if(mActivateModificationsCheckBox.isSelected()) {
			mDeleteButton.setEnabled(true);
			mInsertNeuronButton.setEnabled(true);
			mInsertSynapseButton.setEnabled(true);
		}
		else {
			mDeleteButton.setSelected(false);
			mInsertNeuronButton.setSelected(false);
			mInsertSynapseButton.setSelected(false);
			mDeleteButton.setEnabled(false);
			mInsertNeuronButton.setEnabled(false);
			mInsertSynapseButton.setEnabled(false);
			mSourceNeuron = null;
		}
	}
	
	
	/**
	 * Removes the Neuron from the current neural net.
	 * If the selected neuron is an input or output neuron, then the neuron 
	 * is not removed and a warning message is displayed.
	 * 
	 * @param obj the neuron object to remove.
	 */
	private void deleteNeuron(NeuronObject obj) {
		
		boolean netChanged = false;
		NeuralNet net = null;
		synchronized (Simulator.getInstance().getNeuralNetManager()) {
			net = Simulator.getInstance().getNeuralNetManager().getCurrentNet();
			if(net == null) {
				return;
			}
			synchronized(net) {
				Vector neurons = net.getNeurons();
				for(Iterator i = neurons.iterator(); i.hasNext();) {
					Neuron n = (Neuron) i.next();
					if(n == obj.getNeuron()) {
						if(n.getType().toLowerCase().equals(Neuron.INPUT) 
								|| n.getType().toLowerCase().equals(Neuron.OUTPUT))
						{
							JOptionPane.showMessageDialog(null, "Input and output neurons can not be deleted!", "Warning", JOptionPane.WARNING_MESSAGE);
							return;
						}
						net.removeNeuron(n);
						netChanged = true;
						//remove all dangling synapses from the net.
						net.validateNetworkConnections();
						break;
					}
				}
			}
		}
		if(netChanged && net != null) {
			Simulator.getInstance().getNeuralNetManager().notifyNetworkStructureChanged(net);
		}
	}
	
	
	/**
	 * Removes a synapse from the neural net.
	 * @param obj the synapse object that represents the synapse to remove.
	 */
	private void deleteSynapse(SynapseObject obj) {
		
		if(obj.getSynapse() == null) {
			return;
		}
		
		boolean netChanged = false;
		NeuralNet net = null;
		
		synchronized (Simulator.getInstance().getNeuralNetManager()) {
			net = Simulator.getInstance().getNeuralNetManager().getCurrentNet();
			if(net == null) {
				return;
			}
			synchronized(net) {
				Vector neurons = net.getNeurons();
				NeuronObject targetObject = obj.getTargetNeuron();
				if(targetObject == null) {
					return;
				}
				Neuron target = targetObject.getNeuron();
				if(target == null) {
					return;
				}
				for(Iterator i = neurons.iterator(); i.hasNext();) {
					Neuron n = (Neuron) i.next();
					if(n == target) {
						netChanged = n.removeSynapse(obj.getSynapse());
						break;
					}
				}
			}
		}
		if(netChanged && net != null) {
			Simulator.getInstance().getNeuralNetManager().notifyNetworkStructureChanged(net);
		}
	}
	
	/**
	 * Adds a synapse.
	 * @param source
	 * @param target
	 */
	public void addSynapse(NeuronObject source, NeuronObject target) {
		
		if(source == null || target == null || source.getNeuron() == null || target.getNeuron() == null) {
			return;
		}
		
		boolean netChanged = false;
		NeuralNet net = null;
		
		synchronized (Simulator.getInstance().getNeuralNetManager()) {
			net = Simulator.getInstance().getNeuralNetManager().getCurrentNet();
			if(net == null) {
				return;
			}
			
			
			synchronized (net) {
				Synapse synapse = new Synapse(source.getNeuron());
				netChanged = target.getNeuron().addSynapse(synapse);
			}
		}
		if(netChanged && net != null) {
			Simulator.getInstance().getNeuralNetManager().notifyNetworkStructureChanged(net);
		}
		
	}
	
	/**
	 * Adds a neuron at a specific position.
	 * @param position
	 */
	public void addNeuron(Point2D.Double position) {
		
		if(position == null) {
			return;
		}
		
		boolean netChanged = false;
		NeuralNet net = null;
		
		synchronized (Simulator.getInstance().getNeuralNetManager()) {
			net = Simulator.getInstance().getNeuralNetManager().getCurrentNet();
			if(net == null) {
				return;
			}
			
			
			synchronized (net) {
				Neuron neuron = new Neuron("HiddenNeuron", net.getDefaultTransferFunction(), Neuron.HIDDEN);
				netChanged = net.addNeuron(neuron);
				if(netChanged) {
					Vector neuronObjects = mVisualizationEnvironment.getNeuronObjects();
					NeuronObject neuronObject = null;
					for(Iterator i = neuronObjects.iterator(); i.hasNext();) {
						NeuronObject no = (NeuronObject) i.next();
						if(!no.isActive()) {
							neuronObject = no;
							break;
						}
						if(neuronObject == null) {
							neuronObject = new NeuronObject(position.x, position.y, 0.0);
							mVisualizationEnvironment.addSimObject(neuronObject);
						}
						else {
							neuronObject.setPosition(position);
						}
						neuronObject.setNeuron(neuron);
						
					}
					
				}
			}
		}
		if(netChanged && net != null) {
			Simulator.getInstance().getNeuralNetManager().notifyNetworkStructureChanged(net);
		}
	}
	

}
