package de.fhg.iais.outdoor.simba.visualization.neuron;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.TextLabel;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunction;
import de.fhg.iais.outdoor.simba.value.Value;

public class NeuronObject extends SimObject {
	
	private Neuron mRepresentedNeuron;
	private TextLabel mNameLabel;
	private NeuronActivationChangeBar mActivationChangeBar;
	private Box mSelectedBox;
	private String mTransferFunctionAbbreviation;
	private TransferFunction mCurrentTransferFunction;
	
	
	public NeuronObject(double x, double y, double angle) {
		super("Neuron", x, y, angle, new Polygon(), false);
		
		Polygon shape = new Polygon();
		shape.addPoint(-20, -20);
		shape.addPoint(-20, 20);
		shape.addPoint(20, 20);
		shape.addPoint(20, -20);
		
		setShape(shape);
		setLayer(3);
		setShapeColor(Color.BLACK);
		setFillShape(false);
		
		mNameLabel = new TextLabel("Name", 0.0, -25.0, 0.0, "", 15, false);
		mNameLabel.setReferenceObject(this);
		mNameLabel.setLayer(5);
		
		mActivationChangeBar = new NeuronActivationChangeBar("ActivationBar", -9.0, 0.0, 0.0, 16.0, 34.0, false);
		mActivationChangeBar.setReferenceObject(this);
		mActivationChangeBar.setLayer(5);
		
		mSelectedBox = new Box("MarkedBox", 0.0, 0.0, 0.0, 46.0, 46.0, false);
		mSelectedBox.setReferenceObject(this);
		mSelectedBox.setLayer(50);
		mSelectedBox.setActive(false);
		mSelectedBox.setShapeTransparency(0.3);
		mSelectedBox.setShapeColor(Color.RED);
		mMarked.addValueChangedListener(this);

		setActive(false);
	}

	public void addChildren(EnvironmentModel model) {
		super.addChildren(model);
		model.addSimObject(mNameLabel);
		model.addSimObject(mActivationChangeBar);
		model.addSimObject(mSelectedBox);
	}

	public boolean init() {
		boolean ok = super.init();
		
		return ok;
	}

	public void reset() {
		super.reset();
	}

	public boolean update() {
		boolean ok = super.update();
		
		if(mRepresentedNeuron != null) {
			double activity = mRepresentedNeuron.getActivation();
			if(activity > 0.0) {
				setShapeColor(new Color(1.0f, (float) (1.0 - activity), (float) (1.0 - activity)));
			}
			else if(activity < 0.0) {
				setShapeColor(new Color((float) (1.0 + activity), (float) (1.0 + activity), 1.0f));
			}
			else {
				setShapeColor(Color.WHITE);
			}
			if(mCurrentTransferFunction != mRepresentedNeuron.getTransferFunction()) {
				synchronizeTransferFunctionAbbreviation();
			}
		}
		
		return ok;
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
	}
	
	public void setNeuron(Neuron neuron) {
		if(mRepresentedNeuron != null && mRepresentedNeuron != neuron) {
			mRepresentedNeuron.getNameValue().removeValueChangedListener(this);
		}
		mRepresentedNeuron = neuron;
		setActive(mRepresentedNeuron != null);
		if(mRepresentedNeuron != null) {
			mName = neuron.getName();
			mActivationChangeBar.setNeuron(neuron);
			mRepresentedNeuron.getNameValue().addValueChangedListener(this);
			mCurrentTransferFunction = mRepresentedNeuron.getTransferFunction();
			synchronizeTransferFunctionAbbreviation();
		}
		else {
			mActivationChangeBar.setNeuron(null);
		}
		mNameLabel.setText(mName);
	}
	
	public void synchronizeTransferFunctionAbbreviation() {
		TransferFunction tf = mRepresentedNeuron.getTransferFunction();
		mTransferFunctionAbbreviation = "";
		if(tf != null) {
			Vector transferFunctions = Simulator.getInstance().getNeuralNetManager().getTransferFunctionPrototypes();
			for(int i = 0; i < transferFunctions.size(); ++i) {
				TransferFunction ltf = (TransferFunction) transferFunctions.elementAt(i);
				if(ltf.getName().equals(tf.getName())) {
					mTransferFunctionAbbreviation = tf.getName().substring(0, Math.min(1, tf.getName().length())) + (i + 1);
				}
			}
		}
	}
	

	public void paintSelf(Graphics2D g) {
		Color c = g.getColor();
		Composite composite = g.getComposite();
		
		g.setColor(mShapeColor);
		g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float) mTransparency.get()));
		g.fill(mShape);
		
		g.setColor(Color.BLACK);
		g.drawPolygon(mShape);
		
		if(mRepresentedNeuron != null) {
			g.scale(1., -1.);
			g.setColor(Color.BLACK);
			g.drawString(mTransferFunctionAbbreviation, 2.0f, -7.0f);
			g.scale(1., -1.);
		}
		
		g.setComposite(composite);
		g.setColor(c);
	}

	public Neuron getNeuron() {
		return mRepresentedNeuron;
	}

	public void setActive(boolean active) {
		super.setActive(active);
		mNameLabel.setActive(active);
		mActivationChangeBar.setActive(active);
		if(!active) {
			setMarked(false);
		}
	}

	public void valueChanged(Value value) {
		super.valueChanged(value);
		if(value == null) {
			return;
		}
		else if(mRepresentedNeuron != null && mRepresentedNeuron.getNameValue() == value) {
			mName = mRepresentedNeuron.getName();
			mNameLabel.setText(mName);
		}
		else if(value == mMarked) {
			mSelectedBox.setActive(mMarked.get());
		}
	}

	public void setShapeTransparency(double transparency) {
		super.setShapeTransparency(transparency);
		mActivationChangeBar.setShapeTransparency(transparency);
	}
	
	
	
	
	

}
